/*
    File:               argonArray.cpp
    Description:        Array and hashtable handling classes.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         23.11.2003
    last updated:       04.02.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonArray_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonArray.h"

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4706 )    // assignment within conditional expression
#endif

namespace ArgonLibrary
{




///<param name="cMemUnit"> desired size of one memory unit in bytes. </param>
///<param name="cBlock"> number of memory units to hold in one block
///  and each block will be allocated at once on the heap. </param>
MemList::MemList( _size_t_ cMemUnit, _size_t_ cBlock )
{
    m_cMemUnit = ( cMemUnit > 0 )? cMemUnit : 1;
    m_cBlock = ( cBlock >= 1 )? cBlock : 1;

    m_cUsedUnits = 0;

    m_pBlock1 = m_pBlockL = NULL;

    m_pPartiallyUsedBlock = NULL;
    m_cPartiallyUsedBlocks = 0;
}



void MemList::Empty()
{
    Block *p = m_pBlockL;

    while( p )
    {
        Block *pp = p->m_prev;
        delete[] (BYTE*) p;
        p = pp;
    }

    m_cUsedUnits = 0;
    m_pBlock1 = m_pBlockL = NULL;
    m_pPartiallyUsedBlock = NULL;
    m_cPartiallyUsedBlocks = 0;
}



///<summary> Creates a new unit but somewhere into the list
///  (the insertion position is random). </summary>
///<param name="bClear"> if TRUE then the memory
///  unit will be cleared as well </param>
///<returns> pointer to new memory unit. <br/>
///    Note: NULL can *never* occur. </returns>
void *MemList::Create( bool bClear )
{
    Unit *unit;
    Block *block = FindAPartiallyUsedBlock();

    if( !block )
    {
        block = CreateBlock();
        unit = (Unit*) &block[1];
        if( ++(block->m_cUsedUnits) < m_cBlock ) {
            m_pPartiallyUsedBlock = block;
            ++m_cPartiallyUsedBlocks;
            block->m_pUnusedUnit = (Unit*) &((BYTE*) unit)[ CUnit() ];
        }
    }
    else
    {
        unit = FindAnUnusedUnit( block );
        if( ++(block->m_cUsedUnits) == m_cBlock ) {
            m_pPartiallyUsedBlock = NULL;
            --m_cPartiallyUsedBlocks;
            block->m_pUnusedUnit = NULL;
        } else {
            _size_t_ cunit = CUnit();
            if( unit == (Unit*) &((BYTE*) &block[1])[ (m_cBlock - 1) * cunit ] ) {
                block->m_pUnusedUnit = NULL;
            } else {
                Unit *nextunit = (Unit*) &((BYTE*) unit)[ cunit ];
                block->m_pUnusedUnit = ( nextunit->m_bUsed )? NULL : nextunit;
            }
        }
    }

    unit->m_pBlock = block;
    unit->m_bUsed  = TRUE;
    ++m_cUsedUnits;

    void *ptr = (VOID*) &unit[1];
    if( bClear ) memset( ptr, 0, m_cMemUnit );

    return ptr;
}

///<summary> Deletes a memory unit. </summary>
///<remarks> IMPORTANT: The <b>pUnit</b> *must* be a valid
///  pointer to a *valid* memory unit of *<b>this</b>* MemList.
///  So it will *not* be checked, if <b>pUnit</b> really is
///  a memory unit of <b>this</b> MemList. </remarks>
void MemList::Delete( void *pUnit )
{
    if( !pUnit )
        return;

    Unit *unit = &((Unit*) pUnit)[-1];
    Block *block = unit->m_pBlock;
    BOOL block_was_partially_used = ( block->m_cUsedUnits < m_cBlock );

    --m_cUsedUnits;
    if( --block->m_cUsedUnits == 0 )
    {
        if( m_pPartiallyUsedBlock == block ) m_pPartiallyUsedBlock = NULL;
        if( block_was_partially_used ) --m_cPartiallyUsedBlocks;
        DeleteBlock( block );
    }
    else
    {
        unit->m_bUsed = FALSE;
        block->m_pUnusedUnit = unit;
        if( !block_was_partially_used ) ++m_cPartiallyUsedBlocks;
        if( !m_pPartiallyUsedBlock ) m_pPartiallyUsedBlock = block;
    }
}



///<summary> Creates a new block at the end of the linked list of blocks. </summary>
MemList::Block *MemList::CreateBlock()
{
    _size_t_ cblock = CBlock();
    Block *block = (Block*) new BYTE [cblock];
    // we expect always a valid pointer. We could throw an exception here if not.

    // Now set the header of the block
    block->m_next = NULL;
    if( block->m_prev = m_pBlockL ) block->m_prev->m_next = block;
    m_pBlockL = block;
    if( !m_pBlock1 ) m_pBlock1 = block;
    block->m_cUsedUnits = 0;
    block->m_pUnusedUnit = NULL; // *must* be set to NULL in here
    block->m_pMemList = this;

    // Now remark each of the unused units (and all of
    // the new units are unused of course) as unused.
    _size_t_ cunit = CUnit();
    BYTE *unit = (BYTE*) &block[1];
    for( _idx_ i = m_cBlock - 1 ; i >= 0 ; --i ) {
        ((Unit*) unit)->m_bUsed = FALSE;
        unit += cunit;
    }

    return block;
}

///<remarks> IMPORTANT: It is imagined always when this
///  method is called, that the <b>block</b> does *not* contain any
///  used memory units any more (regardless of the values
///  set in the block header and its unit headers). </remarks>
void MemList::DeleteBlock( Block *block )
{
    if( block->m_prev ) block->m_prev->m_next = block->m_next;
    else m_pBlock1 = block->m_next;
    if( block->m_next ) block->m_next->m_prev = block->m_prev;
    else m_pBlockL = block->m_prev;

    delete[] (BYTE*) block;
}

MemList::Block *MemList::FindAPartiallyUsedBlock()
{
    if( m_pPartiallyUsedBlock ) return m_pPartiallyUsedBlock;
    if( m_cPartiallyUsedBlocks == 0 ) return NULL;

    Block *block = m_pBlock1;
    while( block )
    {
        if( block->m_cUsedUnits < m_cBlock ) return ( m_pPartiallyUsedBlock = block );
        block = block->m_next;
    }

    return NULL;
}

MemList::Unit *MemList::FindAnUnusedUnit( Block *block )
{
    if( block->m_pUnusedUnit ) return block->m_pUnusedUnit;
    _size_t_ cunit = CUnit();
    BYTE *unit = (BYTE*) &block[1];
    for( _idx_ i = m_cBlock - 1 ; i >= 0 ; --i ) {
        if( !((Unit*) unit)->m_bUsed ) return ( block->m_pUnusedUnit = (Unit*) unit );
        unit += cunit;
    }
    return NULL;
}




}
