/*
    File:               argonBase.h
    Description:        basic defines, types, classes, ...
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         25.08.2005
    last updated:       10.11.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonBase_h
#define __argonBase_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include <math.h>
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




#ifdef  _W64
#define _w64_ _W64
#else
#define _w64_
#endif

#ifndef _countof
#define _countof(array) (sizeof(array) / sizeof((array)[0]))
#endif

typedef INT_PTR     _idx_;
typedef size_t      _size_t_;
typedef INT_PTR     _ssize_t_;

#define makeFourcc(ch0, ch1, ch2, ch3)                                          \
                (((DWORD)(BYTE)(ch0))        | (((DWORD)(BYTE)(ch1)) <<  8) |   \
                (((DWORD)(BYTE)(ch2)) << 16) | (((DWORD)(BYTE)(ch3)) << 24 ))

#define makeTwocc(ch0, ch1)     (((WORD)(BYTE)(ch0)) | (((WORD)(BYTE)(ch1)) << 8))

#define makeDword(a, b)             ((DWORD)MAKELONG(a,b))

#define makeUint64(a, b)            ((UINT64)(((DWORD)(a)) | ((UINT64)((DWORD)(b))) << 32))
#define loDword(l)                  ((DWORD)(l))
#define hiDword(l)                  ((DWORD)(((UINT64)(l) >> 32) & 0xffffffff))

inline WORD   swapBytesOfWord   ( WORD   w ) { return MAKEWORD( HIBYTE(w), LOBYTE(w) ); }
inline DWORD  swapWordsOfDword  ( DWORD dw ) { return makeDword( HIWORD(dw), LOWORD(dw) ); }
inline DWORD  swapBytesOfDword  ( DWORD dw ) { return makeDword( swapBytesOfWord(HIWORD(dw)), swapBytesOfWord(LOWORD(dw)) ); }
inline UINT64 swapDwordsOfUint64( UINT64 u ) { return makeUint64( swapBytesOfDword(hiDword(u)), swapBytesOfDword(loDword(u)) ); }

class ByteOrder
{
    public :
        ByteOrder() {  }

        static bool isLittleEndian() { WORD w = 1; return * (const BYTE*) &w == 1; }
        static bool isBigEndian   () { WORD w = 1; return * (const BYTE*) &w == 0; }
};




template< class T > T tbound( T _min, T x, T _max ) { return ( x < _min )? _min : ( x > _max ? _max : x ); }
template< class T > void tswap( T& a, T& b ) { T h = a; a = b; b = h; }
template< class T > T tmin( T a, T b ) { return ( a <= b )? a : b; }
template< class T > T tmax( T a, T b ) { return ( a >= b )? a : b; }
template< class T > T tmid( T a, T b ) { return ( a + b ) / (T) 2; }

template< class T > T tcloser( T x, T _val1, T _val2 ) { if( _val2 < _val1 ) tswap(_val1, _val2); x = tbound(_val1, x, _val2); T dist1 = x - _val1; T dist2 = _val2 - x; return dist1 >= dist2 ? _val1 : _val2; } // returns the closer value (order is unimportant)
template< class T > bool tiscloser( T x, T _val1, T _val2 ) { bool swapped = false; if( _val2 < _val1 ) { tswap(_val1, _val2); swapped = true; } x = tbound(_val1, x, _val2); T dist1 = x - _val1; T dist2 = _val2 - x; return dist1 >= dist2 ? !swapped : swapped; } // returns true if x is closer (or equal) to _val1 than to _val2.
template< class T > T tsign( T x ) { return ( x < (T) 0 ) ? (T) -1 : ( x > (T) 0 ? (T) 1 : (T) 0 ); }


///<summary>
///  Template <b>type</b> gives basic information on types, especially
///  on numeric value types and also performs general operations on them.
///</summary>
template< class T >
class type
{
    public :
        enum { isSigned = false };
        enum { isFloat = false };

        static T getmax();
        static T getmin();

        static T abs( T x );
        static T round( T x );
};

template<>
class type< __int64 >
{
    public :
        enum { isSigned = true };
        enum { isFloat = false };

        static __int64 getmax() { return 0x8000000000000000; }
        static __int64 getmin() { return 0x7fffffffffffffff; }

        static __int64 abs( __int64 x ) { return ::_abs64(x); }
        static __int64 round( __int64 x ) { return x; }
};

template<>
class type< _w64_ long >
{
    public :
        enum { isSigned = true };
        enum { isFloat = false };

        static _w64_ long getmax() { return 0x80000000; }
        static _w64_ long getmin() { return 0x7fffffff; }

        static _w64_ long abs( _w64_ long x ) { return ::labs((long) x); }  // cast to long to supress warning from __w64 option
        static _w64_ long round( _w64_ long x ) { return x; }
};

template<>
class type< _w64_ int >
{
    public :
        enum { isSigned = true };
        enum { isFloat = false };

        static _w64_ int getmax() { return 0x80000000; }
        static _w64_ int getmin() { return 0x7fffffff; }

        static _w64_ int abs( _w64_ int x ) { return ::abs((int) x); }  // cast to int to supress warning from __w64 option
        static _w64_ int round( _w64_ int x ) { return x; }
};

template<>
class type< short >
{
    public :
        enum { isSigned = true };
        enum { isFloat = false };

        static short getmax() { return (short) -0x8000; }
        static short getmin() { return (short)  0x7fff; }

        static short abs( short x ) { return (short) ::abs(x); }
        static short round( short x ) { return x; }
};

template<>
class type< char >
{
    public :
        enum { isSigned = true };
        enum { isFloat = false };

        static char getmax() { return (char) -0x80; }
        static char getmin() { return (char)  0x7f; }

        static char abs( char x ) { return (char) ::abs(x); }
        static char round( char x ) { return x; }
};

template<>
class type< unsigned __int64 >
{
    public :
        enum { isSigned = false };
        enum { isFloat = false };

        static unsigned __int64 getmax() { return 0xffffffffffffffffu; }
        static unsigned __int64 getmin() { return 0x0000000000000000u; }

        static unsigned __int64 abs( unsigned __int64 x ) { return x; }
        static unsigned __int64 round( unsigned __int64 x ) { return x; }
};

template<>
class type< _w64_ unsigned long >
{
    public :
        enum { isSigned = false };
        enum { isFloat = false };

        static unsigned long getmax() { return 0xffffffffu; }
        static unsigned long getmin() { return 0x00000000u; }

        static unsigned long abs( unsigned long x ) { return x; }
        static unsigned long round( unsigned long x ) { return x; }
};

template<>
class type< _w64_ unsigned int >
{
    public :
        enum { isSigned = false };
        enum { isFloat = false };

        static unsigned int getmax() { return 0xffffffffu; }
        static unsigned int getmin() { return 0x00000000u; }

        static unsigned int abs( unsigned int x ) { return x; }
        static unsigned int round( unsigned int x ) { return x; }
};

template<>
class type< unsigned short >
{
    public :
        enum { isSigned = false };
        enum { isFloat = false };

        static unsigned short getmax() { return (unsigned short) 0xffffu; }
        static unsigned short getmin() { return (unsigned short) 0x0000u; }

        static unsigned short abs( unsigned short x ) { return x; }
        static unsigned short round( unsigned short x ) { return x; }
};

template<>
class type< unsigned char >
{
    public :
        enum { isSigned = false };
        enum { isFloat = false };

        static unsigned char getmax() { return (unsigned char) 0xffu; }
        static unsigned char getmin() { return (unsigned char) 0x00u; }

        static unsigned char abs( unsigned char x ) { return x; }
        static unsigned char round( unsigned char x ) { return x; }
};

template<>
class type< float >
{
    public :
        enum { isSigned = true };
        enum { isFloat = true };

        static float getmax() { return  3.4e38f; }
        static float getmin() { return -3.4e38f; }

        static float abs( float x ) { return ::fabsf( x ); }
        static float round( float x ) { return floorf( x + 0.5f ); }
};

template<>
class type< double >
{
    public :
        enum { isSigned = true };
        enum { isFloat = true };

        static double getmax() { return  1.7e308; }
        static double getmin() { return -1.7e308; }

        static double abs( double x ) { return ::fabs( x ); }
        static double round( double x ) { return floor ( x + 0.5 ); }
};




///<summary> This class can be used for iterating something. It
///  allows easy and conform checking if iteration has finished. </summary>
class Iterator
{
    private :
        // abstract iteration position type
        struct __iteration_pos {  };
    public :
        typedef struct __iteration_pos *iteration_pos;

        // the stored position
        iteration_pos m_pos;

    public :
        Iterator() { m_pos = before_iteration_start(); }
        Iterator( const Iterator& iter ) { m_pos = iter.m_pos; }

        Iterator& operator =( const Iterator& iter )
            { m_pos = iter.m_pos; }

        bool Finished() const
            { return m_pos == end_of_iteration(); }

            ///<summary> Returns true if interation has not yet started. </summary>
        bool IsBeforeStart() const
            { return m_pos == before_iteration_start(); }

        void Reset () { m_pos = before_iteration_start(); }
        void Finish() { m_pos = end_of_iteration(); }

        static iteration_pos before_iteration_start() { return (iteration_pos) -1L; }
        static iteration_pos end_of_iteration      () { return (iteration_pos) -2L; }
};




class WaitCursor
{
    public :
        HCURSOR m_hcrsrWaitCursor;
        HCURSOR m_hcrsrPreviousCursor;

    public :
        WaitCursor() { m_hcrsrWaitCursor = ::LoadCursor( NULL, IDC_WAIT ); m_hcrsrPreviousCursor = NULL; Set(); }
        WaitCursor( HCURSOR hcrsrWaitCursor ) { m_hcrsrWaitCursor = hcrsrWaitCursor; m_hcrsrPreviousCursor = NULL; Set(); }
        WaitCursor( const WaitCursor& src ) { m_hcrsrWaitCursor = src.m_hcrsrWaitCursor; m_hcrsrPreviousCursor = NULL; Set(); }
        ~WaitCursor() { Reset(); }

    public :
        void Set()   { if( !m_hcrsrPreviousCursor ) m_hcrsrPreviousCursor = ::SetCursor( m_hcrsrWaitCursor ); }
        void Reset() { if( m_hcrsrPreviousCursor ) { ::SetCursor( m_hcrsrPreviousCursor ); m_hcrsrPreviousCursor = NULL; } }
};




}

#endif
