/*
    File:               argonErrors.cpp
    Description:        Provides easy error handling as well as easy
                        error definition creation, extension and costumization.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         27.05.2006
    last updated:       01.10.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)

    Thanks go to Jovita De Guzman for the italian translations.
*/

#define __argonErrors_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonErrors.h"

namespace ArgonLibrary
{




///<summary>
///  <b>Init()</b> inits this object with the desired parameters.
///</summary>
///<param name="pfnIsSevere"> pointer to IsSevere function. </param>
///<param name="pfnIsPureSuccess"> pointer to IsPureSuccess function. </param>
///<param name="pfnGetDescriptionString"> pointer to GetDescriptionString function. </param>
///<param name="pfnGetDisplayMessage"> pointer to GetDisplayMessage function. </param>
void ErrorGroup_Access::Init( const ErrLangTableStrStruct& ltsClassName,
                              const ErrLangTableStrStruct& ltsGroupName,
                              const GUID&                  groupId,
                              tfnIsSevere                  pfnIsSevere,
                              tfnIsPureSuccess             pfnIsPureSuccess,
                              tfnGetDescriptionString      pfnGetDescriptionString,
                              tfnGetDisplayMessage         pfnGetDisplayMessage )
{
    LockModuleWide lmw;

    if( !m_bInitialized )
    {
        m_ltsClassName = ltsClassName;
        m_ltsGroupName = ltsGroupName;

        m_GroupId = groupId;

        m_pfnIsSevere             = pfnIsSevere;
        m_pfnIsPureSuccess        = pfnIsPureSuccess;
        m_pfnGetDescriptionString = pfnGetDescriptionString;
        m_pfnGetDisplayMessage    = pfnGetDisplayMessage;

        m_bInitialized = true;
    }
}

bool ErrorGroup_Access::GetGroupName( String& str ) const
{
    LanguageTableString lts( m_ltsGroupName.m_pAccess, m_ltsGroupName.m_ID );
    return lts.getStringThreadLocale( str );
}

bool ErrorGroup_Access::GetGroupName( String& str, LCID language ) const
{
    LanguageTableString lts( m_ltsGroupName.m_pAccess, m_ltsGroupName.m_ID );
    return lts.getString( str, language );
}

bool ErrorGroup_Access::GetClassName( String& str ) const
{
    LanguageTableString lts( m_ltsClassName.m_pAccess, m_ltsClassName.m_ID );
    return lts.getStringThreadLocale( str );
}

bool ErrorGroup_Access::GetClassName( String& str, LCID language ) const
{
    LanguageTableString lts( m_ltsClassName.m_pAccess, m_ltsClassName.m_ID );
    return lts.getString( str, language );
}




void ErrorCode::SetSpecializedError( const ErrorCode& ecSpecialized )
{
    if( !m_pSpecialized ) m_pSpecialized = new ErrorSpecializedExtras( ecSpecialized );
    else                  m_pSpecialized->SetSpecializedError( ecSpecialized );
}

void ErrorCode::AppendSpecializedError( const ErrorCode& ecSpecialized )
{
    if( !m_pSpecialized )                             SetSpecializedError( ecSpecialized );
    else if( m_pSpecialized->m_bSpecializedErrorSet ) m_pSpecialized->m_SpecializedError.AppendSpecializedError( ecSpecialized );
    else                                              m_pSpecialized->SetSpecializedError( ecSpecialized );
}

void ErrorCode::SetAdditionalText( PCTSTR pstrText )
{
    if( !m_pSpecialized && !_isstrempty_(pstrText) ) m_pSpecialized = new ErrorSpecializedExtras;
    if( m_pSpecialized )
        { m_pSpecialized->m_strAdditionalText = pstrText; m_pSpecialized->m_strAdditionalText.MakeLF(); }
}

void ErrorCode::AppendAdditionalText( PCTSTR pstrText )
{
    if( !_isstrempty_(pstrText) )
    {
        if( !m_pSpecialized ) m_pSpecialized = new ErrorSpecializedExtras;

        if( m_pSpecialized )
        {
            if( !m_pSpecialized->m_strAdditionalText.IsEmpty() )
                m_pSpecialized->m_strAdditionalText += _T("\n\n");
            m_pSpecialized->m_strAdditionalText += pstrText;
            m_pSpecialized->m_strAdditionalText.MakeLF();
        }
    }
}

bool ErrorCode::GetDescriptionString( String& str, LCID language, bool bInclCascadedErrors ) const
{
    bool ok = GetSafeDescriptionString( str, language );

    if( ok ) str.TrimEnd();

    if( !m_pSpecialized ) return ok;

    if( !m_pSpecialized->m_strAdditionalText.IsEmpty() )
    {
        str += _T("\n\n");
        str += m_pSpecialized->m_strAdditionalText;
    }

    if( !m_pSpecialized->m_bSpecializedErrorSet ) return ok;
    if( !bInclCascadedErrors ) return ok;

    if( !ok )
        return m_pSpecialized->m_SpecializedError.GetDescriptionString( str, language, bInclCascadedErrors );

    str += _T("\n\n");
    str += ErrorsLanguageTable::getString( errltsid_PrecedingErrors, language );
    str += _T(':');

    for( ErrorSpecializedExtras *pSpec = m_pSpecialized ; pSpec ; pSpec = pSpec->m_SpecializedError.m_pSpecialized )
    {
        if( !pSpec->m_bSpecializedErrorSet ) continue;
        ErrorCode *ptr = &pSpec->m_SpecializedError;

        String errorname;
        ptr->GetGroupName( errorname, language );

        if( errorname.IsEmpty() )
            ptr->GetClassName( errorname, language );

        if( !errorname.IsEmpty() )
            errorname += _T(", ");

        String speccode;
        speccode.Sprintf( _T("0x%lx (%ld)"), ptr->m_Code, ptr->m_Code );

        String s;
        if( ptr->GetDescriptionString( s, language, false ) )
        {
            s.Replace( _T('\n'), _T(' ') );

            _size_t_ cnt = str.ReplacestringRet( _T("{#}"), errorname + speccode );
            cnt         += str.ReplacestringRet( _T("{f}"), String(ptr->m_pstrFct) );
            cnt         += str.ReplacestringRet( _T("{m}"), s );
            // Note: We *must* replace {m} at the *end* of all the replacements here,
            // because it can be that the newly inserted string #s# contains some
            // of the {...} itself and they should be overwritten only in the next loop.

            if( cnt == 0 )
            {
                str += _T('\n') + s + _T(" (") + errorname + speccode + _T(')');
                if( ptr->m_pstrFct ) str += _T(' ') + ErrorsLanguageTable::getString( errltsid_inFunction, language ).LowercaseFirstChar() + _T(' ') + String(ptr->m_pstrFct) + _T('.');
            }
        }
        else
            str += _T('\n') + errorname + speccode;
    }

    return ok;
}

void ErrorCode::GetDisplayMessage( String& str, LCID language, bool bInclCascadedErrors ) const
{
    String desc;
    GetDescriptionString( desc, language, bInclCascadedErrors );
    GetSafeDisplayMessage( str, desc, language );
}

bool ErrorCode::GetGroupName( String& str ) const
{
    tryHW { return m_pGroup->GetGroupName( str ); }
    catchHW { return false; }
}

bool ErrorCode::GetGroupName( String& str, LCID language ) const
{
    tryHW { return m_pGroup->GetGroupName( str, language ); }
    catchHW { return false; }
}

bool ErrorCode::GetClassName( String& str ) const
{
    tryHW { return m_pGroup->GetClassName( str ); }
    catchHW { return false; }
}

bool ErrorCode::GetClassName( String& str, LCID language ) const
{
    tryHW { return m_pGroup->GetClassName( str, language ); }
    catchHW { return false; }
}

const ErrorCode *ErrorCode::GetSpecializedError() const
{
    return m_pSpecialized ? ( m_pSpecialized->m_bSpecializedErrorSet ?
        &m_pSpecialized->m_SpecializedError : NULL ) : NULL;
}

void ErrorCode::GetAdditionalText( String& text ) const
{
    if( m_pSpecialized ) text = m_pSpecialized->m_strAdditionalText;
    else text.Empty();
}

bool ErrorCode::GetSafeDescriptionString( String& str, LCID language ) const
{
    tryHW { return m_pGroup->GetDescriptionString( str, m_Code, language ); }
    catchHW { return false; }
}

void ErrorCode::GetSafeDisplayMessage( String& str, const String& desc, LCID language ) const
{
    tryHW { m_pGroup->GetDisplayMessage( str, desc, m_Code, m_pstrFct, language ); }
    catchHW {  }
}




///<summary>
///  Opens a message box and displays the error information. <br/>
///  For most of the parameters, see descriptios to <b>MoreMessageBox::MessageBox()</b>.
///</summary>
///<param name="mbType"> combination of MB_* values. Similar to the system's
///  MessageBox function. If you specify no icon in here, a fitting icon
///  will be used for you (MB_ICONINFORMATION if error code is a success value,
///  MB_ICONERROR if error code is a failure value). </param>
///<param name="dwMmbFlags"> combination of MMB_* flags </param>
///<param name="pMmb"> if not NULL, then this is the MoreMessageBox to use (otherwise
///  an internal own object will be used). This is useful e.g. if you have a checkbox and
///  want to set/get its checked state, etc. </param>
const reterr_t& reterr_t::DisplayMessage( HWND hwndParent, PCTSTR pstrCaption, UINT mbType, DWORD dwMmbFlags, LCID language,
                                          PCTSTR pstrAdditionalText,
                                          PCTSTR pstrButton1, PCTSTR pstrButton2, PCTSTR pstrButton3, PCTSTR pstrMoreButton,
                                          PCTSTR pstrCheck1,
                                          PCTSTR pstrHeader, PCTSTR pstrExtraMsg, PCTSTR pstrExtraLink_Text, PCTSTR pstrExtraLink_Url,
                                          HMODULE hModuleToLoadIconsFrom, UINT_PTR nMsgIcon_IdOrHicon, UINT_PTR nExtraIcon_IdOrHicon,
                                          MoreMessageBox *pMmb ) const
{
    DisplayMessageRet( hwndParent, pstrCaption, mbType, dwMmbFlags, language,
                       pstrAdditionalText,
                       pstrButton1, pstrButton2, pstrButton3, pstrMoreButton,
                       pstrCheck1,
                       pstrHeader, pstrExtraMsg, pstrExtraLink_Text, pstrExtraLink_Url,
                       hModuleToLoadIconsFrom, nMsgIcon_IdOrHicon, nExtraIcon_IdOrHicon,
                       pMmb );
    return *this;
}

///<summary>
///  Opens a message box and displays the error information. <br/>
///  For most of the parameters, see descriptios to <b>MoreMessageBox::MessageBox()</b>.
///</summary>
///<param name="mbType"> combination of MB_* values. Similar to the system's
///  MessageBox function. If you specify no icon in here, a fitting icon
///  will be used for you (MB_ICONINFORMATION if error code is a success value,
///  MB_ICONERROR if error code is a failure value). </param>
///<param name="dwMmbFlags"> combination of MMB_* flags </param>
///<param name="pMmb"> if not NULL, then this is the MoreMessageBox to use (otherwise
///  an internal own object will be used). This is useful e.g. if you have a checkbox and
///  want to set/get its checked state, etc. </param>
INT_PTR reterr_t::DisplayMessageRet( HWND hwndParent, PCTSTR pstrCaption, UINT mbType, DWORD dwMmbFlags, LCID language,
                                     PCTSTR pstrAdditionalText,
                                     PCTSTR pstrButton1, PCTSTR pstrButton2, PCTSTR pstrButton3, PCTSTR pstrMoreButton,
                                     PCTSTR pstrCheck1,
                                     PCTSTR pstrHeader, PCTSTR pstrExtraMsg, PCTSTR pstrExtraLink_Text, PCTSTR pstrExtraLink_Url,
                                     HMODULE hModuleToLoadIconsFrom, UINT_PTR nMsgIcon_IdOrHicon, UINT_PTR nExtraIcon_IdOrHicon,
                                     MoreMessageBox *pMmb ) const
{
    String s;
    GetDisplayMessage( s, language );

    UINT uType = mbType;
    if( 0 == (uType & MB_ICONMASK) )
        uType |= (Succeeded() ? MB_ICONINFORMATION : MB_ICONERROR);

    LanguageTableString langcap( &ErrorsLanguageTable::s_Access, errltsid_ErrorInformationTitle );
    String caption = langcap.getString( language );

    //INT_PTR ret = MessageBox( hwndParent, s, caption,
    //                          uType );

    String msg;
    GetDescriptionString( msg, language, false );

    if( !_isstrempty_(pstrAdditionalText) )
    {
        msg += _T("\n\n");
        msg += pstrAdditionalText;
    }

    MoreMessageBox mmb;
    if( !pMmb ) pMmb = &mmb;

    #ifndef _WIN32_WCE
    if( pMmb->m_strOwnBtn1.IsEmpty()    ) pMmb->m_strOwnBtn1    = pstrButton1;
    if( pMmb->m_strOwnBtn2.IsEmpty()    ) pMmb->m_strOwnBtn2    = pstrButton2;
    if( pMmb->m_strOwnBtn3.IsEmpty()    ) pMmb->m_strOwnBtn3    = pstrButton3;
    if( pMmb->m_strOwnMoreBtn.IsEmpty() ) pMmb->m_strOwnMoreBtn = pstrMoreButton;
    if( pMmb->m_strOwnCheck1.IsEmpty()  ) pMmb->m_strOwnCheck1  = pstrCheck1;
    #endif

    INT_PTR ret = pMmb->MessageBox( hwndParent, msg, s, caption,
                                    uType, dwMmbFlags, language );

    return ret;
}




const LangTableDesc g_ErrorsLangDescs[] = {
    #ifndef _argon_no_LangEnglish
    // English
    { errltsid_ErrorInformationTitle,       LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("Error Information") },
    { errltsid_InternalError,               LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("Internal Error") },
    { errltsid_SystemError,                 LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("System Error") },
    { errltsid_COMError,                    LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("COM Error") },
    { errltsid_UnidentifiedError,           LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("Unidentified Error") },
    { errltsid_ErrorCode,                   LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("Error code") },
    { errltsid_InFunction,                  LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("In function") },
    { errltsid_inFunction,                  LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("in function") },
    { errltsid_PrecedingErrors,             LANG_ENGLISH,       ltmapt_Direct,      (LONG_PTR) _T("Preceding errors") },
    #endif

    #ifndef _argon_no_LangGerman
    // German
    { errltsid_ErrorInformationTitle,       LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("Fehlerbeschreibung") },
    { errltsid_InternalError,               LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("Interner Fehler") },
    { errltsid_SystemError,                 LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("System Fehler") },
    { errltsid_COMError,                    LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("COM Fehler") },
    { errltsid_UnidentifiedError,           LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("Unbekannter Fehler") },
    { errltsid_ErrorCode,                   LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("Fehlercode") },
    { errltsid_InFunction,                  LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("In Funktion") },
    { errltsid_inFunction,                  LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("in Funktion") },
    { errltsid_PrecedingErrors,             LANG_GERMAN,        ltmapt_Direct,      (LONG_PTR) _T("Vorangegangene Fehler") },
    #endif

    #ifndef _argon_no_LangItalian
    // Italian
    { errltsid_ErrorInformationTitle,       LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("Informazione dell'errore") },
    { errltsid_InternalError,               LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("Errore interno") },
    { errltsid_SystemError,                 LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("Errore del sistema") },
    { errltsid_COMError,                    LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("COM Errore") },
    { errltsid_UnidentifiedError,           LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("Errore non identificato") },
    { errltsid_ErrorCode,                   LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("Errore codice") },
    { errltsid_InFunction,                  LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("In funzione") },
    { errltsid_inFunction,                  LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("in funzione") },
    { errltsid_PrecedingErrors,             LANG_ITALIAN,       ltmapt_Direct,      (LONG_PTR) _T("Errori precedenti") },
    #endif

    LANGTABLEDESC_ENDMARK
};

ErrorsLanguageTable g_ErrorsLanguageTable;
LanguageDescriptionTableT<ErrorsLanguageTable> g_ErrorsLanguageDescriptionTable( g_ErrorsLangDescs );




// Note:
//  MAKELCID( MAKELANGID( LANG_ENGLISH, SUBLANG_NEUTRAL ), SORT_DEFAULT )
// is the same like LANG_ENGLISH itself.

const GenErrDesc g_InternalErrorDescs[] = {
    // First the error codes that shall be mapped to other error codes.
    // LANG_NEUTRAL indicates that is should be used regardless of language.
    // If the mapping would match, but its string cannot be retrieved from the
    // mapped error code, the searching will continue in this array so.
    { errOk,                                LANG_NEUTRAL,       gemapt_SysErr,      (LONG_PTR) ERROR_SUCCESS },
    { errFailed,                            LANG_NEUTRAL,       gemapt_HresErr,     (LONG_PTR) E_FAIL },
    { errCritical,                          LANG_NEUTRAL,       gemapt_HresErr,     (LONG_PTR) E_UNEXPECTED },
    { errFileSeekFailed,                    LANG_NEUTRAL,       gemapt_SysErr,      (LONG_PTR) ERROR_SEEK_ON_DEVICE },

    #ifndef _argon_no_LangEnglish
    // English
    { errOk,                                LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Succeeded!") },
    { suxNoOp,                              LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("No operation.") },
    { suxMemoryCopyProduced,                LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("A copy of the memory has been produced.") },
    { errFailed,                            LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Unspecified Failure!") },
    { errCritical,                          LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Critical Failure!") },
    { errAborted,                           LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Operation aborted.") },
    { errUserAborted,                       LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Aborted by user.") },
    { errEndOfData,                         LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Unexpected end of data reached.") },
    { errIllegalMemType,                    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Unsupported memory type.") },
    { errMemoryNotReallocatable,            LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("The memory is not reallocatable.") },
    { errMemoryNotResizeable,               LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("The memory is not resizeable.") },
    { errMemoryDataLocked,                  LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Writing to the memory data is not allowed. It is locked.") },
    { errMemoryReallocLocked,               LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Reallocating or resizing the memory is locked.") },
    { errMemoryMustBeEmpty,                 LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("The memory object must be empty to successfully conclude the operation.") },
    { errMemoryCouldNotAttachPointer,       LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Could not attach the pointer to the shared memory object.") },
    { errMemoryCannotDecrease,              LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Cannot decrease memory size. Only enlarging memory size is allowed.") },
    { errDynMemWrongDimensions,             LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Cannot conclude the operation on the memory object because of incorrect dimension sizes.") },
    { errDynMemNoHeader,                    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("The memory object is missing header data.") },
    { errOutOfRange,                        LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("One or more parameters or indices are out of range.") },
    { errNoDataAvailable,                   LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("No data available.") },
    { errAccessLocked,                      LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Cannot conclude the operation. Access is locked.") },
    { errEOF,                               LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("End of file reached.") },
    { errFileNotOpen,                       LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("No file is open yet.") },
    { errFileAlreadyOpen,                   LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("The file is already open.") },
    { errFileReadFailed,                    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Failed reading from a file.") },
    { errFileWriteFailed,                   LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Failed writing to a file.") },
    { errFileSeekFailed,                    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Failed seeking a file position.") },
    { errFileTellFailed,                    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Failed querying a file position.") },
    { errFileNoStream,                      LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("A file stream is missing.") },
    { errFileStreamError,                   LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("File stream error.") },
    { errFileUnknownType,                   LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Unknown file type.") },
    { errFileReadNotEnoughData,             LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Not enough data to read from a file.") },
    { errMemFileReferenceNotReallocable,    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("The memory cannot be resized since it is referencing external memory.") },
    { errRegNotOpen,                        LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("A registry key is not open yet.") },
    { errRegAlreadyOpen,                    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("A registry key is already open.") },
    { errRegDerivedNotOpen,                 LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("A registry key or private profile is not open yet.") },
    { errRegDerivedAlreadyOpen,             LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("A registry key or private profile is already open.") },
    { errProfileNotOpen,                    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("A private profile is not open yet.") },
    { errProfileAlreadyOpen,                LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("A private profile is already open.") },
    #endif

    #ifndef _argon_no_LangGerman
    // German
    { errOk,                                LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Erfolgreich!") },
    { suxNoOp,                              LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Keine Operation.") },
    { suxMemoryCopyProduced,                LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Eine Kopie des Speichers wurde erzeugt.") },
    { errFailed,                            LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Nicht nher spezifizierter Fehler!") },
    { errCritical,                          LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Kritischer Fehler!") },
    { errAborted,                           LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Operation abgebrochen.") },
    { errUserAborted,                       LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Abbruch durch den Anwender.") },
    { errEndOfData,                         LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Unerwartetes Ende der Daten erreicht.") },
    { errIllegalMemType,                    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Nicht untersttzter Speichertyp.") },
    { errMemoryNotReallocatable,            LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Der Speicher kann nicht reallokiert werden.") },
    { errMemoryNotResizeable,               LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Die Speichergre kann nicht verndert werden.") },
    { errMemoryDataLocked,                  LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Es ist nicht erlaubt, Daten in den Speicher zu schreiben. Er ist gesperrt.") },
    { errMemoryReallocLocked,               LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Reallokieren oder Verndern der Speichergre ist gesperrt.") },
    { errMemoryMustBeEmpty,                 LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Das Speicherobjekt mu leer sein um die Operation erfolgreich zu beenden.") },
    { errMemoryCouldNotAttachPointer,       LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Konnte den Pointer nicht an das Objekt des gemeinsam genutzten Speichers attachen.") },
    { errMemoryCannotDecrease,              LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Die Speichergre kann nicht verkleinert werden. Es ist lediglich mglich, den Speicher zu vergrern.") },
    { errDynMemWrongDimensions,             LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Konnte die Operation an dem Speicher wegen inkorrekter Dimensionsgren nicht durchfhren.") },
    { errDynMemNoHeader,                    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Der Speicher enthlt keine Headerdaten.") },
    { errOutOfRange,                        LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Ein oder mehrere Parameter oder Indices sind auerhalb des erlaubten Bereichs.") },
    { errNoDataAvailable,                   LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Keine Daten verfgbar.") },
    { errAccessLocked,                      LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Die Operation kann nicht durchgefhrt werden. Zugang wird verweigert.") },
    { errEOF,                               LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Dateiende erreicht.") },
    { errFileNotOpen,                       LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Die Datei ist nicht offen.") },
    { errFileAlreadyOpen,                   LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Die Datei ist bereits offen.") },
    { errFileReadFailed,                    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Fehler beim Versuch, von einer Datei zu lesen.") },
    { errFileWriteFailed,                   LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Fehler beim Versuch, in eine Datei zu schreiben.") },
    { errFileSeekFailed,                    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Fehler beim Versuch, den File-Pointer einer Datei auf eine andere Position zu bewegen.") },
    { errFileTellFailed,                    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Fehler beim Versuch, den File-Pointer einer Datei abzufragen.") },
    { errFileNoStream,                      LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Kein Datei-Stream.") },
    { errFileStreamError,                   LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Datei-stream-Fehler.") },
    { errFileUnknownType,                   LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Unbekannter Dateityp.") },
    { errFileReadNotEnoughData,             LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Eine Datei enthlt nicht genug zu lesende Daten.") },
    { errMemFileReferenceNotReallocable,    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Die Speichergre kann nicht gendert werden, da er zur Zeit externen Speicher referenziert.") },
    { errRegNotOpen,                        LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Ein Registry-Key is nicht offen.") },
    { errRegAlreadyOpen,                    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Ein Registry-Key ist bereits offen.") },
    { errRegDerivedNotOpen,                 LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Ein Registry-Key oder Private-Profile ist nicht offen.") },
    { errRegDerivedAlreadyOpen,             LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Ein Registry-Key oder Private-Profile ist bereits offen.") },
    { errProfileNotOpen,                    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Ein Private-Profile ist nicht offen.") },
    { errProfileAlreadyOpen,                LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Ein Private-Profile ist bereits offen.") },
    #endif

    #ifndef _argon_no_LangItalian
    // Italian
    { errOk,                                LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T(" riuscito!") },
    { suxNoOp,                              LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Nessuna operazione.") },
    { suxMemoryCopyProduced,                LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Una copia della memoria  stata prodotta.") },
    { errFailed,                            LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Errore non specificato!") },
    { errCritical,                          LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Errore critico!") },
    { errAborted,                           LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Operazione abortita.") },
    { errUserAborted,                       LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Abortito da utente.") },
    { errEndOfData,                         LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Fine inaspettata dei dati raggiunta.") },
    { errIllegalMemType,                    LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Tipo di memoria non sostenuto.") },
    { errMemoryNotReallocatable,            LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("La memoria non  riallocalizzabile.") },
    { errMemoryNotResizeable,               LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("La memoria non  ridimensionabile.") },
    { errMemoryDataLocked,                  LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Scrivere sulla memoria non  permesso. Modifiche disabilite.") },
    { errMemoryReallocLocked,               LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Riallocando o ridimensionando la memoria disabilitati.") },
    { errMemoryMustBeEmpty,                 LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("L'oggetto della memoria deve essere vuoto per concludere l'operazione con sucesso.") },
    { errMemoryCouldNotAttachPointer,       LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Non si attaccava il cursore alla memoria obiettiva condivisa.") },
    { errMemoryCannotDecrease,              LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Non si pu diminuire la memoria.  permesso solamente ingrandirla.") },
    { errDynMemWrongDimensions,             LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Non si riesce a concludere l'operazione sulla memoria obiettiva a causa delle dimensioni incorrette.") },
    { errDynMemNoHeader,                    LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("La memoria obiettiva manca dati della testata.") },
    { errOutOfRange,                        LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Uno o pi parametri o indice  fuori di serie.") },
    { errNoDataAvailable,                   LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Nessun dato disponibile.") },
    { errAccessLocked,                      LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Non si conclude l'operazione. Accesso disabilitato.") },
    { errEOF,                               LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Fine del file raggiunto.") },
    { errFileNotOpen,                       LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Nessun file  aperto.") },
    { errFileAlreadyOpen,                   LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Il file  gi aperto.") },
    { errFileReadFailed,                    LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Lettura del file non riuscita.") },
    { errFileWriteFailed,                   LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Scrittura sul file non riuscita.") },
    { errFileSeekFailed,                    LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Ricerca di posizione del file non riuscita.") },
    { errFileTellFailed,                    LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Recupero di posizione del file non riuscito.") },
    { errFileNoStream,                      LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Manca un file stream.") },
    { errFileStreamError,                   LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Errore del file stream.") },
    { errFileUnknownType,                   LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Tipo di file sconosciuto.") },
    { errFileReadNotEnoughData,             LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Non abbastanza dati per leggere un file.") },
    { errMemFileReferenceNotReallocable,    LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("La memoria non si pu ridimensionare quando indica memoria esterna.") },
    { errRegNotOpen,                        LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Una chiave della registrazione non  ancora aperta.") },
    { errRegAlreadyOpen,                    LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Una chiave della registrazione  gi aperta.") },
    { errRegDerivedNotOpen,                 LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Una chiave della registrazione o profilo privato non  ancora aperto.") },
    { errRegDerivedAlreadyOpen,             LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Una chiave della registrazione o profilo privato  gi aperto.") },
    { errProfileNotOpen,                    LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Un profilo privato non  ancora aperto.") },
    { errProfileAlreadyOpen,                LANG_ITALIAN,       gemapt_Direct,      (LONG_PTR) _T("Un profilo privato  gi aperto.") },
    #endif

    GENERRDESC_ENDMARK
};

const GUID                                        g_groupidInternalError = GROUPID_INTERNALERROR;
GenericErrorClass                                 g_eclGeneric;                                                                                  // error class for internal errors and user-defined errors
InternalErrorGroup                                g_egrInternal( g_groupidInternalError, ErrLangTableStrStruct().Init(errltsid_InternalError) ); // error group for internal errors defined especially by the Argon Library
GenericErrorDescriptionTableT<InternalErrorGroup> g_edtInternal( g_InternalErrorDescs );                                                         // and its description tables




const GUID                  g_groupidSystemError = GROUPID_SYSTEMERROR;
SystemErrorClass            g_eclSystemError;                           // error class for normal operating system errors
SystemErrorGroup            g_egrSystemError( g_groupidSystemError );   // error group for normal operating system errors
SystemErrorDescriptionTable g_edtSystemError;                           // and its description table

eErrorLangPriority SystemErrorDescriptionTable::GetDescriptionString( String& str, LONG code, LCID language ) const
{
    PTSTR pMsgBuf = NULL;
    DWORD ret;
    eErrorLangPriority pri = errlangprior_none;

    LANGID want_lang_id = LANGIDFROMLCID( language );   // the combined form with PRIMARYLANDID and SUBLANGID

    // First we try to get our desired language.
    ret = ::FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                           FORMAT_MESSAGE_FROM_SYSTEM |
                           FORMAT_MESSAGE_IGNORE_INSERTS,
                           NULL,
                           (DWORD) code,
                           want_lang_id,
                           (PTSTR) &pMsgBuf,
                           0,
                           NULL );

    if( ret > 0 )
        pri = errlangprior_LangId;

    if( ret == 0 )
    {
        // We get the default language if the desired one could not be retrieved.
        ret = ::FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                               FORMAT_MESSAGE_FROM_SYSTEM |
                               FORMAT_MESSAGE_IGNORE_INSERTS,
                               NULL,
                               (DWORD) code,
                               MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                               (PTSTR) &pMsgBuf,
                               0,
                               NULL );

        if( ret > 0 )
            pri = errlangprior_any;
    }

    if( ret > 0 )
        str = pMsgBuf;

    if( pMsgBuf )
        LocalFree( pMsgBuf );

    return pri;
}




const GUID                   g_groupidHresultError = GROUPID_HRESULTERROR;
HresultErrorClass            g_eclHRESULT;                          // error class for HRESULT errors
HresultErrorGroup            g_egrHRESULT( g_groupidHresultError ); // error group for HRESULT errors
HresultErrorDescriptionTable g_edtHRESULT;                          // and its description table

eErrorLangPriority HresultErrorDescriptionTable::GetDescriptionString( String& str, LONG code, LCID language ) const
{
    // Handles generic HRESULTs.
    // If the HRESULT values are DirectX errors or something special,
    // you can add your own error description table object which
    // extracts DirectX error message or so.

    eErrorLangPriority pri = g_edtSystemError.GetDescriptionString( str, code, language );

    // Note: Here, we handle DX9 messages. This will work only if
    // "dxerr9.h" has been included in the "stdafx.h" include file.
    #ifdef _DXERR9_H_   //#ifdef DXGetErrorDescription9
    if( pri < errlangprior_any )
    {
        str = DXGetErrorDescription9( (HRESULT) code );
        pri = errlangprior_any;
    }
    #endif

    // Note: Here, we handle DX8 messages. This will work only if
    // "dxerr8.h" has been included in the "stdafx.h" include file.
    #ifdef _DXERR8_H_   //#ifdef DXGetErrorDescription8
    if( pri < errlangprior_any )
    {
        str = DXGetErrorDescription8( (HRESULT) code );
        pri = errlangprior_any;
    }
    #endif

    return pri;
}




}
