/*
    File:               argonExpressionEvaluator.h
    Description:        Class to evaluate mathematical expressions fast.

    Author:             Roman Komary
    created on:         21.09.2002
    last updated:       11.11.2006 (Roman Komary)

    Copyright (c) 2002-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    The ExpressionEvaluator provides a means to evaluate mathematical
    expressions with the use of variables.
    The variables are stored only in as double values (see type teevar).
    On the left side of the = operator, a destination variable name
    must exist, on the right side, an arithmetic expression can be specified
    with function calls and bracket hierarchies.


    ExpressionEvaluator usage:
    -----------------------------

    Call ExpressionEvaluator's AddExpressions() method to add one or more
    expressions. They will be somewhat compiled but not yet calculated.
    This is done to provide execution speeds as fast as possible
    during evaluation procedure (see below).

    Now if you want, add variable definitions using AddVariables() for variables
    you want to be defined and have values already before starting evaluations.
    Note: As described in the next sentence, SetValue() can also be used to
    create variables automatically if they do not exist already.
    The advantage of AddVariables() is that after every Evaluate() call, you
    do not need to call GetValue() for each variable since GetValue() needs
    to lookup variables by searching for the name which is not fast.
    Using AddVariables() gives the possiblity to provide an array of structs
    which allow easy and fast access of the variables without the need to
    look them up by name before and after each Evaluate() call.

    Before evaluation, you can create and set several variable values using SetValue()
    method. Or you can push values of variables using PushVariables().
    Or access the variables you passed to AddVariables() to set their values.

    Then call Evaluate() to perform the calculations.

    Afterwards you can call GetResult() to get the result value or call
    GetValue() of variables you want to retrieve the calculated values from.
    Or call PopVariables() to restore previously pushed variable values.
    Or access the variables you passed to AddVariables() to get their values.

    You can repeat this procedure as often as you want without the need to
    do the first 2 steps (add expressions and variables) again since
    the ExpressionEvaluator object holds the expressions in compiled
    form for fast evaluation execution speed.

    You can use the variable name mapper (MapVariableName()) to map a new variable
    name to another existing variable name. This way, you can have multiple names
    for the same variable.


    Sample:
    ---------

        zoom = 1.05 + 0.01*( 0.60*sin(0.339*time) + 0.40*sin(0.276*time) );
        rot = 0 + 0.010*( 0.60*sin(0.381*time) + 0.40*sin(0.579*time) );
        decay = decay - 0.01*equal(frame%6,0);
        isnewtime = above( time-lasttime, 0.25 );

    To make this sample work without errors, the following variables must be
    defined first (either using AddVariables() or SetValue()):
        time
        decay
        frame
        lasttime

    The resulting values will be
        zoom
        rot
        decay
        isnewtime

    Note:
    The variables zoom, rot and isnewtime do not need to be added first, because
    they are assigned to and so will be created automatically anyway.
    But if you want to access their values quickly (so not lookup the variables
    always by name using GetValue()), I recommend to add the variables
    also explicitely using AddVariables(). Because then, you can access
    their values easily and fast through pointer access instead of having
    them looked up by their names.


    More complex sample:
    ----------------------

        zoom = 1.05 + 0.01*( 0.60*sin(0.339*time) + 0.40*sin(0.276*time) );
        rot = 0 + 0.010*( 0.60*sin(0.381*time) + 0.40*sin(0.579*time) );
        decay = decay - 0.01*equal(frame%6,0);
        isnewtime = above( time-lasttime, 0.25 );
        vprev = if( isnewtime, vjump, vprev );
        vjump = above(bass_att,1.19);
        dojump = equal(equal(vjump,1)+equal(vprev,0)+isnewtime,3);
        tmp = if( dojump, -1, 1 );
        jump = if( above(jump,-0.1), tmp, -tmp );
        lasttime = if( dojump, time, lasttime );
        att = (bass_att+mid_att+treb_att)/2;
        att = min( att, 1 );
        q1 = 1.5 * att * jump;
        zoom = zoom + 0.3*max(bass_att-1,0);
        num = ( num + dojump )%8;
        bInvert = equal(num,7);
        falloff = if( dojump, 1, exp(falloff*.55)-1 );
        fGammaAdj = 2 + 2*falloff;
        wave_r = wave_r + 0.12*( 0.60*sin(0.733*time) + 0.40*sin(0.345*time) );
        wave_g = wave_g + 0.10*( 0.60*sin(0.600*time) + 0.40*sin(0.456*time) );
        wave_b = wave_b + 0.09*( 0.60*sin(0.510*time) + 0.40*sin(0.550*time) );
        dx=dx+0.004*sin(time*0.974);
        dy=dy+0.004*cos(time*0.705);
        random_bool = if( above(rand(10),5)*dojump, 1-random_bool, random_bool );
        nWaveMode = if( above(rand(10),6)*dojump, 0, 2 );
        wave_g = if( nWaveMode, wave_g, 1 );
        wave_b = if( nWaveMode, wave_b, 1 );

        rot=rot + 0.02*q1/(rad+0.2) + 0.15*sin( rad*11+rand(3*treb*rad) + 3*time )*rad*q1;
        zoom = zoom - random_bool*att*4*min(0,rad-0.2);

    In this sample (similar to the one above), many variables must
    be added first so that the expressions can be evaluated.
    It's too complicated now to list all of them.

    The important thing is to show with this sample how it is possible
    to execute only a specific set of all the added expressions:

    In your C/C++ source code, you could first add the first set of expressions (so
    from line zoom= to line wave_b=) using AddExpressions().
    Then call
        _idx_ end_block_1 = ee.GetMarker();
    to remember the end of the first block.
    (Note: ee is an ExpressionEvaluator object.)

    Then add the second set of expressions (so the last 2 lines rot= and zoom=).
    Again, call
        _idx_ end_block_2 = ee.GetMarker();

    Now when you have added also all necessary variables,
    you can force only a subset of all the expressions to be evaluated.
    E.g. call
        ee.Evaluate( 0, end_block_1 );
    to evaluate the first block.
    Or call
        ee.Evaluate( end_block_1, end_block_2 );
    to evaluate only the second block.
*/

#ifndef __argonExpressionEvaluator_h
#define __argonExpressionEvaluator_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonMath.h"
    #include "argonBase.h"
    #include "argonArray.h"
    #include "argonString.h"
    #include "argonErrors.h"
#endif

namespace ArgonLibrary
{




class ExpressionEvaluator;

typedef double teevar;      // variable type to use for variables in the expression-evaluator
typedef int    teeint;      // and an integer variable type used when calculating binary operations.

#define EETOKENLEN 255      // maximum length (in chars) for a token (variable name etc.)

struct EEConst
{
    PCSTR   m_pszName;
    teevar  m_value;
};

typedef teevar (*TeeFunc)( teevar * );
struct EEFunction
{
    PCSTR   m_pszName;
    int     m_nArgs;
    TeeFunc m_pFunc;
    PCSTR   m_pszDescription;
};

#define EEVARSTORAGESIZE 8
struct EEVar
{
    teevar m_value; // the actual value (must be on first position and must be immediately followed by the 'storage' array)
    teevar m_storage[EEVARSTORAGESIZE]; // additional space for pushing previously calculated values of this variable
};

// Used to add variables to the variable list
struct EEVarItem
{
    PCSTR   m_pszName; // name of the variable to add
    teevar *m_pValue;  // on return, this pointer will point onto the variable value so that it can be set/released easily
};

struct EESavedVar
{
    teevar *m_pVariable;
    teevar  m_savedValue;
};

///<summary>
///  Class <b>VariableSaver</b>. <br/>
///  Important: The variables stored in here are stored
///  as *pointers*. They are valid only as long as the variables
///  exist in the ExpressionEvaluator from which 'this'
///  saver has been filled.
///</summary>
class VariableSaver
{
    friend class VariableSaver;
    friend class ExpressionEvaluator;

    protected :
        Array<EESavedVar> m_aVars;

    public :
        VariableSaver() {  }

        void        Empty() { m_aVars.Empty(); }
        void        Store();
        void        Restore();

        void        AddVariables( const VariableSaver& vs );
        _size_t_    NumVariables() const { return m_aVars.GetSize(); }
        EESavedVar  GetVariableAt( _idx_ ix ) const { return m_aVars[ix]; }
        bool        RemoveVariableAt( _idx_ ix ) { return m_aVars.Remove(ix); }

    protected :
        void AddVar( teevar *pVar );
};

// Following is a list of commands stored in the compiled-expressions-memory
// to be executed.
#define EECMD_PUSHPTR             1 // following must be a pointer to a teevar value
#define EECMD_PUSHVALUE           2 // following must be a value of type teevar
#define EECMD_FUNCCALL            3 // following must be a ptr to the function to be called and a WORD with the number of args (*before* the command, the function's parameters must have been pushed)
#define EECMD_ENDOFEXPR           4 // end of expression. takes the result and stores it and then empties the stack
#define EECMD_ASSIGN            101 // following must be a pointer to a teevar value to be set (and it also will perform a EECMD_PUSHPTR cmd automatically afterwards)
#define EECMD_LE                102
#define EECMD_LT                103
#define EECMD_GE                104
#define EECMD_GT                105
#define EECMD_EQ                106
#define EECMD_NOTEQ             107
#define EECMD_ADD               108
#define EECMD_SUB               109
#define EECMD_LOGICOR           110
#define EECMD_LOGICAND          111
#define EECMD_BINOR             112
#define EECMD_BINAND            113
#define EECMD_MUL               114
#define EECMD_DIV               115
#define EECMD_MOD               117
#define EECMD_POW               118
#define EECMD_NEG               119
#define EECMD_LOGICNOT          120
#define EECMD_BINNOT            121
#define EECMD_ROUND_2           122 // takes 2 args: 1st is ignored, 2nd is rounded (used for Milkdrop mode)

// Modes the Expression Evaluator does support
//  . The normal mode works as you would expect the supported
//  . functions and operations to work.
#define EEMODE_NORMAL             0
#define EEMODE_DEFAULT            EEMODE_NORMAL
//  . Milkdrop (a winamp plugin) has a bit different handling.
//  . The & and | operators just ignore the first value and round
//  . the second one (sometimes jumping to a large value which is
//  . ignored in this implementation). Maybe some other operators
//  . work a bit different, but that is also ignored in here.
#define EEMODE_MILKDROP           1

class ExpressionEvaluator
{
    protected :
        static const EEConst      m_aEEConst[];
        static const EEFunction   m_aEEFunctions[];
        HashTable<StringA, EEVar> m_htVariables;
        // Note: The 'm_htVariables' must be a list rather than an array because
        // when new variables are added, the old ones must stay on their
        // memory location or else the compiled expressions would point
        // to invalid memory addresses.
        int    m_nPushes;   // how many pushes have been done to store variables in the extra storage
        bool   m_bFunctionNamesCaseSensitive;
        bool   m_bVariableNamesCaseSensitive;  // this also includes constant variable names
        DWORD  m_dwMode;    // a EEMODE_* value

        Array<BYTE> m_aCompiledExpressions;

        HashTable<StringA, StringA> m_htVarNameMapper;  // maps variables to other ones.


    public :
        ExpressionEvaluator( bool bFctCaseSens = true, bool bVarCaseSens = true );

    // Options
    public :
        void  SetCaseSensitive( bool bFctCaseSens, bool bVarCaseSens );
            ///<summary> call this *before* you add any expressions to this class. </summary>
        void  SetMode( DWORD mode ) { m_dwMode = mode; }
        DWORD GetMode() const { return m_dwMode; }

    // Variable and Value access
    public :
        bool GetValue( PCSTR pszVarName, teevar& value ) const;
        bool SetValue( PCSTR pszVarName, teevar  value );
        bool GetNextVariable( Iterator& pos, StringA& varname, teevar& value ) const;
        bool AddVariables( EEVarItem *pVarItems ) { return AddVariables( pVarItems, sizeof(EEVarItem) ); }
        bool AddVariables( EEVarItem *pVarItems, _size_t_ cVarItem );
        void PushVariables();
        bool PopVariables();
        bool GrabVariableState ( int state );
        bool StoreVariableState( int state );
        bool GetNextConstant( Iterator& pos, StringA& constname, teevar& value ) const;
    protected :
        const EEConst *FindConst( PCSTR pszName ) const;

    // Variable name mapper
    public :
        bool MapVariableName( PCSTR pszNameToMap, PCSTR pszNameToMapTo );
        void EmptyVarNameMappings() { m_htVarNameMapper.Empty(); }
        StringA GetVariableName( PCSTR pszVarName ) const;

    // Functions
    public :
        bool GetNextFunction( Iterator& pos, StringA& funcname, int& nArgs ) const;
        bool GetNextFunction( Iterator& pos, StringA& funcname, int& nArgs, StringA& funcdescription ) const;
    protected :
        const EEFunction *FindFunc( PCSTR pszName ) const;

    // Expressions
    private :
        PCSTR           m_pszExpression, m_pszE;
        #define         TOKENT_UNDEF      0
        #define         TOKENT_NAME       1
        #define         TOKENT_DELIMITER  2
        #define         TOKENT_NUMBER     3
        #define         TOKENT_HEXNUM     4
        DWORD           m_dwTokenType; // a TOKENT_* value
        char            m_szToken[EETOKENLEN+1];
        #define         EEINITIALSTACKSIZE 256
        Array<teevar,1> m_aStack;  // (dynamic-sized) stack used during evaluation 'Evaluate()'. IMPORTANT: AT THE MOMENT, IT WILL BE SET TO A FIXED HIGH VALUE (EEINITIALSTACKSIZE)
        int             m_cStack;  // and the actual size of the stack (the array itself can be larger to prevent too many memory-reallocations)
        teevar          m_nResult;
        VariableSaver  *m_pVarSvr; // used only temporarily
    public :
        void     Empty( bool bEmptyVars = true );
        reterr_t AddExpressions( PCSTR pszExpr, bool bMore = false, VariableSaver *pVariableSaver = NULL );
            ///<summary> returns string-index where error in <b>AddExpression()</b> did occur (if so) </summary>
        _idx_    ErrorAt() const { return (_idx_) (m_pszE - m_pszExpression); }
        _idx_    GetMarker() const { return (_idx_) m_aCompiledExpressions.GetSize(); }
        bool     Evaluate( _idx_ nStartAtMarker = 0, _idx_ nEndAtMarker = 0 );
        teevar   GetResult() const { return m_nResult; }
        static bool IsOperatorChar( char ch );
    private :
        void     SkipSpaces() { while( isspace(*m_pszE) ) m_pszE++; }
        reterr_t Parse();
        reterr_t Level1();
        reterr_t Level2();
        reterr_t Level3();
        reterr_t Level4();
        reterr_t Level5();
        reterr_t Level6();
        reterr_t Level7();
        reterr_t Level8();
        reterr_t Level9();
        reterr_t Level10();
        reterr_t Level11();
        // And now for the compiled memory:
        void AddWord( WORD w );
        void AddCmd( WORD wCmd ) { AddWord( wCmd ); }
        void AddPtr( void *ptr );
        void AddValue( teevar v );
};




}

#endif
