/*
    File:               argonMemory.h
    Description:        Memory handling classes.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         2.10.2003
    last updated:       14.07.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    Switches:

    _argon_dont_use_SEH     Define _argon_dont_use_SEH if try/catch can also
                            catch asynchronous (hardware) exceptions. This is the
                            case for compiler option /EHac instead of /EHsc.
                            /EHsc usually is the default option.
                            If you don't define _argon_dont_use_SEH, then hardware
                            exceptions will be caught using __try/__except instead
                            of try/catch.
                            Hardware exceptions can be illegal pointer access, etc.
*/

#ifndef _argon_no_includes
    #define _argonErrors_no_Part2
    #include "argonErrors.h"
    // We need to include "argonErrors.h" Part1 *before* defining __argonMemory_h.
#endif

#ifndef __argonMemory_h
#define __argonMemory_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonBase.h"
    #include "argonArray.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4100 )    // unreferenced parameter

    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{

/*
    In here, the following classes are important for you for Memory management:
        MemorySpace
        MemoryShared
        DynMem
    and for the type of memory to use: the enum
        mem_t

    Besides memory management, there are bit-operation-functions:
        _memmovebits_
        _memsetbits_
    and related bit-pointer and bit-length classes:
        BitPtr
        BitCPtr
        BitLen

    For DynMem,
    beside the BitPtr, BitCPtr and BitLen classes, the following
    classes will be important to handle the dimensions of a DynMem object:
        DimPt
        DimSize


    Memory management objects:
    --------------------------------------

    MemorySpace and MemoryShared
    ============================
    can allocate memory dynamically in byte steps.

    You can define blocks (in bytes) as step count for the
    objects resizing memory. The advantage is speed because
    the objects internally will have to reallocate memory
    only when boundaries of the blocks are exceeded when
    resizing memory.

    You can change the type of memory (see enum mem_t) at any
    time as long as the memory types allow reallocation (e.g.
    you cannot change a memory to MemT_External. But you can
    change a MemT_External to any other type).
    For MemT_FileMapping, it is possible to create a named memory,
    but then it cannot be resized without loosing the named
    memory and will gain a non-named, new memory instead.

    The memories can be locked in different aspects.
    Locking data will prevent the memory objects returning write-to-
    pointers (#WritePtr()#) to forbid writing to the memory data.
    Locking reallocation will return an error every time
    the current memory shall be reallocated or resized.
    Btw: Locking data will also disallow memory to be made
    smaller because this would cause loss of data.

    Difference between MemorySpace and MemoryShared:
    The difference is: A MemorySpace object has the memory for its
    own and for its own alone.
    A MemoryShared object can share the same memory with other
    MemoryShared objects within the same process.
    This is achieved internally by reserving some header information before
    the actual memory data start.
    When querying the memory start pointer (#ConstPtr()# or #WritePtr()#),
    you can get the pointer to the really allocated memory start (so the header start)
    by subtracting #GetInternalContentOffset()#. This can be important
    e.g. with file mapping objects if you have the handle to the
    file mapping object and need to get the offset to know the
    start of the memory data after the header info.
    MemorySpace objects do not store any header informations.

    Another advantage of MemoryShared objects is size of the objects
    themselves. MemoryShared are tried to be kept as small as possible.
    On 32bit systems, they are 12 bytes in size.
    A MemorySpace object's size is larger.


    DynMem
    ======
    has all the advantages of MemoryShared objects,
    plus it allows allocating memory also in other units than only bytes.
    You can even allocate in steps of bits.
    In addition, it is capable to handle multiple dimensions of memory.
    There is no limitation for the dimension count and size.

    E.g. DynMem is very useful for handling pictures (especially
    pictures or bitmaps with color depth less than 8 bits), or
    n-dimensional matrices.
    DynMem also supports row-padding which is necessary for bitmaps
    so that a new pixel row starts at a 32 bit boundary.
*/

// Now we define the tryHW and catchHW exception handler macros.
// We use them to catch asynchronous (that is hardware) exceptions
// like illegal pointer access.
// Define _argon_dont_use_SEH if try/catch can also catch asynchronous
// (hardware) exceptions. This is the case for compiler option /EHac instead
// of /EHsc.
// If you don't define _argon_dont_use_SEH, then hardware
// exceptions will be caught using __try/__except instead
// of try/catch.
// Usage:
//      tryHW { .... }
//      catchHW { .... }
#ifndef tryHW
#ifdef _argon_dont_use_SEH
    #define tryHW       try
    #define catchHW     catch(...)
#else
    #define tryHW       __try
    #define catchHW     __except( EXCEPTION_EXECUTE_HANDLER )
    // The disadvantages of using SEH (structured exception handling):
    // - It is not fully designed for C++.
    // - It is not in the C++ standard (so maybe not all compilers support it).
    // - In a function with __try/__except, no local object destructors
    //   can be executed.
    // The advantage:
    // - It can catch hardware exceptions (like illegal pointer access)
    //   even if the compiler option /EHsc is specified (this is the default,
    //   it means that try/catch can catch only synchronous exceptions).
#endif
#endif
// Note: By default (so by not defining _argon_dont_use_SEH) tryHW/catchHW
// map to __try/__except because the compiler does not use SEH for
// try/catch by default (/EHsc).
// IMPORTANT: Since tryHW/catchHW can map to __try/__except, beware of the
// restrictions of __try/__except usage: No local C++ object unwinding (dtor)
// is allowed in the function where the __try/__except statements are.
// In such a case, make sub-functions where in there, the object unwindings
// are done.

// #BitLen#, #BitPtr# and #BitCPtr# need the following declarations.
// Note: See below for the BitPtr versions of these bit-memory functions.
void *_memmovebits_( void *pdest, int destbit, const void *psrce, int srcebit, INT64 bitcount );
void *_memsetbits_ ( void *pdest, int destbit, char bit, INT64 bitcount );

///<summary> The <b>BitLen</b> class handles a size value which is described
///  in byte- and/or bit-sizes. The size in bits is stored internally as
///  unsigned 64-bit value. </summary>
class BitLen
{
    friend class BitLen;

    protected :
        UINT64 m_cbits;  // size in bits.
        // This value is expected to be an unsigned value. So do not
        // change it to signed ever. (E.g. #DynMem# does expect it that way).

    public :
        BitLen() { m_cbits = 0; }
        BitLen( INT64 bits ) { bits >= 0 ? SetCBits( (UINT64) bits ) : SetError(); } // passing -1 will set an error
        //BitLen( const BitLen& src ) { m_cbits = src.m_cbits; }
        //  Note: We let the compiler generate copy ctor and operator automatically.
        UINT64 cbytes() const { return IsByteable() ? m_cbits >> 3 : 0; }
        UINT64 cbits () const { return m_cbits; }
        void Empty() { m_cbits = 0; }
        void SetError() { m_cbits = (UINT64) 0xffffffffffffffff; }
        bool IsError() const { return m_cbits == (UINT64) 0xffffffffffffffff; }
        void SetCBytes( INT64 bytes ) { bytes > 0 ? m_cbits = (((UINT64) bytes) << 3) : m_cbits = 0; }
        void SetCBits ( INT64 bits  ) { bits  > 0 ? m_cbits = (UINT64) bits : m_cbits = 0; }
        bool IsByteable() const { return ((int) m_cbits & 7) == 0; }
        bool IsEmpty() const { return IsError() || m_cbits == 0; }
        // #EvalByteSize()# will evaluate the minimum size in bytes.
        UINT64 EvalByteSize() const { return IsError() ? 0 : ((m_cbits + 8 - 1) >> 3); }
        //BitLen& operator = ( const BitLen& src ) { m_cbits = src.m_cbits; return *this; }
        //  Note: We let the compiler generate copy ctor and operator automatically.
        BitLen& operator +=( const BitLen& bl ) { ( IsError() || bl.IsError() )? SetError() : SetCBits( (INT64) m_cbits + (INT64) bl.m_cbits ); return *this; }
        BitLen& operator -=( const BitLen& bl ) { ( IsError() || bl.IsError() )? SetError() : SetCBits( (INT64) m_cbits - (INT64) bl.m_cbits ); return *this; }
        BitLen  operator + ( const BitLen& bl ) { return ( IsError() || bl.IsError() )? (INT64) -1 : (INT64) m_cbits + (INT64) bl.m_cbits; }
        BitLen  operator - ( const BitLen& bl ) { return ( IsError() || bl.IsError() )? (INT64) -1 : (INT64) m_cbits - (INT64) bl.m_cbits; }
};

class BitCPtr;

///<summary> The <b>BitPtr</b> class is something like a pointer but with bit-addressing.
///  See <b>BitLen</b> class for the corresponding size-in-bits class. </summary>
class BitPtr
{
    friend class BitPtr;
    friend class BitCPtr;

    protected :
        void  *m_ptr;   // (byte)address of the byte being addressed
        LONG32 m_bitnr; // bit-nr inside the byte to address ( 0 = bit0, 7 = bit7 )

    public :
        BitPtr() { m_ptr = NULL; m_bitnr = 7; }
        BitPtr( void *pointer ) { Set( pointer, 0 ); }
        BitPtr( void *pointer, INT64 bit_offset ) { Set( pointer, bit_offset ); }
        //BitPtr( const BitPtr& src ) { m_ptr = src.m_ptr; m_bitnr = src.m_bitnr; }
        //  Note: We let the compiler generate copy ctor and operator automatically.

        void *ptr  () const { return m_ptr; }
        int   bitnr() const { return (int) m_bitnr; }

        // Methods to set/get a value at the bit-position that #this# #BitPtr# points to.
        void SetBit     ( char     v ) { if( !IsNull() ) _memmovebits_( m_ptr, m_bitnr, &v, 0, 1             ); } // only bit 0 of #v# is used
        void SetChar    ( char     v ) { if( !IsNull() ) _memmovebits_( m_ptr, m_bitnr, &v, 7, 8 * sizeof(v) ); }
        void SetShort   ( short    v ) { if( !IsNull() ) _memmovebits_( m_ptr, m_bitnr, &v, 7, 8 * sizeof(v) ); }
        void SetInt     ( int      v ) { if( !IsNull() ) _memmovebits_( m_ptr, m_bitnr, &v, 7, 8 * sizeof(v) ); }
        void SetLong    ( long     v ) { if( !IsNull() ) _memmovebits_( m_ptr, m_bitnr, &v, 7, 8 * sizeof(v) ); }
        void SetLongLong( LONGLONG v ) { if( !IsNull() ) _memmovebits_( m_ptr, m_bitnr, &v, 7, 8 * sizeof(v) ); }
        void SetInt64   ( INT64    v ) { if( !IsNull() ) _memmovebits_( m_ptr, m_bitnr, &v, 7, 8 * sizeof(v) ); }
        void SetIntPtr  ( INT_PTR  v ) { if( !IsNull() ) _memmovebits_( m_ptr, m_bitnr, &v, 7, 8 * sizeof(v) ); }
        char     GetBit     () const { CHAR     v = 0; if( !IsNull() ) _memmovebits_( &v, 0, m_ptr, m_bitnr, 1             ); return v; }
        char     GetChar    () const { CHAR     v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        short    GetShort   () const { SHORT    v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        int      GetInt     () const { INT      v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        long     GetLong    () const { LONG     v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        LONGLONG GetLongLong() const { LONGLONG v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        INT64    GetInt64   () const { INT64    v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        INT_PTR  GetIntPtr  () const { INT_PTR  v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }

        // Note: The #.m_ptr# and #.m_bitnr# items are intended to be passed directly
        // to the parameters of #_memmovebits_()# and #_memsetbits_()#.
        // Note: In the following methods, bit-addresses will consist of a address to
        // a byte (void-pointer) and a bit-offset.
        // IMPORTANT: A bit-offset is different to a bit-nr. While a bit-nr can only
        // be a value from 0 to 7 and indicates directly the bit number, a bit-offset
        // can exceed this limits (also be negative) and the order of bits in a byte
        // is numbered in the other direction. Examples for a bit-offset and bit-nr:
        //      (   pointer, bit_offset )   |  (       ptr, bitnr )
        //       0x00010000           0     |   0x00010000      7
        //       0x00010000           7     |   0x00010000      0
        //       0x00010000           8     |   0x00010001      7
        //       0x00010001           0     |   0x00010001      7
        //       0x00010000           9     |   0x00010001      6
        //       0x00010000          10     |   0x00010001      5
        //       0x00010000          -1     |   0x0000ffff      0
        //       0x00010000          -8     |   0x0000ffff      7
        //       0x00010000          -9     |   0x0000fffe      0
        BitPtr& Set( void *pointer, INT64 bit_offset );
        BitPtr& SetNull() { m_ptr = NULL; m_bitnr = 7; return *this; }
        //BitPtr& operator = ( const BitPtr& src ) { m_ptr = src.m_ptr; m_bitnr = src.m_bitnr; return *this; }
        //  Note: We let the compiler generate copy ctor and operator automatically.
        BitPtr& operator +=( INT64 bit_offset );
        BitPtr& operator -=( INT64 bit_offset ) { return operator +=( -bit_offset ); }
        BitPtr& operator +=( const BitLen& len );
        BitPtr& operator -=( const BitLen& len );
        BitPtr  operator + ( INT64 bit_offset );
        BitPtr  operator - ( INT64 bit_offset ) { return operator +( -bit_offset ); }
        BitPtr  operator + ( const BitLen& len );
        BitPtr  operator - ( const BitLen& len );
        INT64   operator - ( const BitCPtr& bitptr2 );
        INT64   operator - ( const BitPtr&  bitptr2 );
        bool IsNull() const { return( !m_ptr && m_bitnr == 7 ); } // if this bitpointer is a NULL-pointer
        bool IsEqual   ( const BitCPtr& bitptr2 ) const;
        bool IsEqual   ( const BitPtr&  bitptr2 ) const { return m_ptr == bitptr2.m_ptr && m_bitnr == bitptr2.m_bitnr; }
        bool IsLessThan( const BitCPtr& bitptr2 ) const;
        bool IsLessThan( const BitPtr&  bitptr2 ) const { return m_ptr < bitptr2.m_ptr || ( m_ptr == bitptr2.m_ptr && m_bitnr > bitptr2.m_bitnr ); }
        bool operator ==( const BitCPtr& bp ) const;
        bool operator ==( const BitPtr&  bp ) const { return  IsEqual(bp);                    }
        bool operator !=( const BitCPtr& bp ) const;
        bool operator !=( const BitPtr&  bp ) const { return !IsEqual(bp);                    }
        bool operator < ( const BitCPtr& bp ) const;
        bool operator < ( const BitPtr&  bp ) const { return  IsLessThan(bp);                 }
        bool operator > ( const BitCPtr& bp ) const;
        bool operator > ( const BitPtr&  bp ) const { return !IsLessThan(bp) && !IsEqual(bp); }
        bool operator <=( const BitCPtr& bp ) const;
        bool operator <=( const BitPtr&  bp ) const { return  IsLessThan(bp) ||  IsEqual(bp); }
        bool operator >=( const BitCPtr& bp ) const;
        bool operator >=( const BitPtr&  bp ) const { return !IsLessThan(bp);                 }
};

///<summary>
///  The <b>BitCPtr</b> is the same as a <b>BitPtr</b> but it has the
///  restrictions of a const pointer. That is, the content the pointer points
///  to shall not be allowed to be edited.
///  But the pointer itself can be changed to point to another location.
///</summary>
class BitCPtr
{
    friend class BitPtr;
    friend class BitCPtr;

    protected :
        void  *m_ptr;   // (byte)address of the byte being addressed
        LONG32 m_bitnr; // bit-nr inside the byte to address ( 0 = bit0, 7 = bit7 )

    public :
        BitCPtr() { m_ptr = NULL; m_bitnr = 7; }
        BitCPtr( const void *pointer ) { Set( pointer, 0 ); }
        BitCPtr( const void *pointer, INT64 bit_offset ) { Set( pointer, bit_offset ); }
        BitCPtr( const BitPtr& src ) { m_ptr = src.m_ptr; m_bitnr = src.m_bitnr; }
        //BitCPtr( const BitCPtr& src ) { m_ptr = src.m_ptr; m_bitnr = src.m_bitnr; }
        //  Note: We let the compiler generate copy ctor and operator automatically.

        const void *ptr  () const { return m_ptr; }
        int         bitnr() const { return (int) m_bitnr; }

        // Methods to get a value at the bit-position that #this# #BitCPtr# points to.
        // Setting is not allowed since this is a to be intended as const pointer.
        char     GetBit     () const { char     v = 0; if( !IsNull() ) _memmovebits_( &v, 0, m_ptr, m_bitnr, 1             ); return v; }
        char     GetChar    () const { char     v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        short    GetShort   () const { short    v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        int      GetInt     () const { int      v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        long     GetLong    () const { long     v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        LONGLONG GetLongLong() const { LONGLONG v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        INT64    GetInt64   () const { INT64    v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }
        INT_PTR  GetIntPtr  () const { INT_PTR  v = 0; if( !IsNull() ) _memmovebits_( &v, 7, m_ptr, m_bitnr, 8 * sizeof(v) ); return v; }

        // Note: The #.m_ptr# and #.m_bitnr# items are intended to be passed directly
        // to the parameters of #_memmovebits_()# and #_memsetbits_()#.
        // Note: In the following methods, bit-addresses will consist of a address to
        // a byte (void-pointer) and a bit-offset.
        // IMPORTANT: A bit-offset is different to a bit-nr. While a bit-nr can only
        // be a value from 0 to 7 and indicates directly the bit number, a bit-offset
        // can exceed this limits (also be negative) and the order of bits in a byte
        // is numbered in the other direction. Examples for a bit-offset and bit-nr:
        //      (   pointer, bit_offset )   |  (       ptr, bitnr )
        //       0x00010000           0     |   0x00010000      7
        //       0x00010000           7     |   0x00010000      0
        //       0x00010000           8     |   0x00010001      7
        //       0x00010001           0     |   0x00010001      7
        //       0x00010000           9     |   0x00010001      6
        //       0x00010000          10     |   0x00010001      5
        //       0x00010000          -1     |   0x0000ffff      0
        //       0x00010000          -8     |   0x0000ffff      7
        //       0x00010000          -9     |   0x0000fffe      0
        BitCPtr& Set( const void *pointer, INT64 bit_offset );
        BitCPtr& SetNull() { m_ptr = NULL; m_bitnr = 7; return *this; }
        //BitCPtr& operator = ( const BitCPtr& src ) { m_ptr = src.m_ptr; m_bitnr = src.m_bitnr; return *this; }
        //  Note: We let the compiler generate copy ctor and operator automatically.
        BitCPtr& operator = ( const BitPtr& src ) { m_ptr = src.m_ptr; m_bitnr = src.m_bitnr; return *this; }
        BitCPtr& operator +=( INT64 bit_offset );
        BitCPtr& operator -=( INT64 bit_offset ) { return operator +=( -bit_offset ); }
        BitCPtr& operator +=( const BitLen& len );
        BitCPtr& operator -=( const BitLen& len );
        BitCPtr  operator + ( INT64 bit_offset );
        BitCPtr  operator - ( INT64 bit_offset ) { return operator +( -bit_offset ); }
        BitCPtr  operator + ( const BitLen& len );
        BitCPtr  operator - ( const BitLen& len );
        INT64        operator - ( const BitCPtr& bitptr2 );
        INT64        operator - ( const BitPtr&  bitptr2 );
        bool IsNull() const { return( !m_ptr && m_bitnr == 7 ); } // if this bitpointer is a NULL-pointer
        bool IsEqual   ( const BitCPtr& bitptr2 ) const { return m_ptr == bitptr2.m_ptr && m_bitnr == bitptr2.m_bitnr; }
        bool IsEqual   ( const BitPtr&  bitptr2 ) const { return m_ptr == bitptr2.m_ptr && m_bitnr == bitptr2.m_bitnr; }
        bool IsLessThan( const BitCPtr& bitptr2 ) const { return m_ptr < bitptr2.m_ptr || ( m_ptr == bitptr2.m_ptr && m_bitnr > bitptr2.m_bitnr ); }
        bool IsLessThan( const BitPtr&  bitptr2 ) const { return m_ptr < bitptr2.m_ptr || ( m_ptr == bitptr2.m_ptr && m_bitnr > bitptr2.m_bitnr ); }
        bool operator ==( const BitCPtr& bp ) const { return  IsEqual(bp);                    }
        bool operator ==( const BitPtr&  bp ) const { return  IsEqual(bp);                    }
        bool operator !=( const BitCPtr& bp ) const { return !IsEqual(bp);                    }
        bool operator !=( const BitPtr&  bp ) const { return !IsEqual(bp);                    }
        bool operator < ( const BitCPtr& bp ) const { return  IsLessThan(bp);                 }
        bool operator < ( const BitPtr&  bp ) const { return  IsLessThan(bp);                 }
        bool operator > ( const BitCPtr& bp ) const { return !IsLessThan(bp) && !IsEqual(bp); }
        bool operator > ( const BitPtr&  bp ) const { return !IsLessThan(bp) && !IsEqual(bp); }
        bool operator <=( const BitCPtr& bp ) const { return  IsLessThan(bp) ||  IsEqual(bp); }
        bool operator <=( const BitPtr&  bp ) const { return  IsLessThan(bp) ||  IsEqual(bp); }
        bool operator >=( const BitCPtr& bp ) const { return !IsLessThan(bp);                 }
        bool operator >=( const BitPtr&  bp ) const { return !IsLessThan(bp);                 }
};

inline bool BitPtr::IsEqual    ( const BitCPtr& bitptr2 ) const { return m_ptr == bitptr2.m_ptr && m_bitnr == bitptr2.m_bitnr; }
inline bool BitPtr::IsLessThan ( const BitCPtr& bitptr2 ) const { return m_ptr < bitptr2.m_ptr || ( m_ptr == bitptr2.m_ptr && m_bitnr > bitptr2.m_bitnr ); }
inline bool BitPtr::operator ==( const BitCPtr& bp ) const { return  IsEqual(bp);                    }
inline bool BitPtr::operator !=( const BitCPtr& bp ) const { return !IsEqual(bp);                    }
inline bool BitPtr::operator < ( const BitCPtr& bp ) const { return  IsLessThan(bp);                 }
inline bool BitPtr::operator > ( const BitCPtr& bp ) const { return !IsLessThan(bp) && !IsEqual(bp); }
inline bool BitPtr::operator <=( const BitCPtr& bp ) const { return  IsLessThan(bp) ||  IsEqual(bp); }
inline bool BitPtr::operator >=( const BitCPtr& bp ) const { return !IsLessThan(bp);                 }

// These are the BitPtr versions of the bit-memory functions declared above.
BitPtr _memmovebits_( BitPtr dest, BitCPtr srce, BitLen size );
BitPtr _memsetbits_ ( BitPtr dest, char bit, BitLen size );

typedef LONG32 dim_t;  // this must be a value of 32 bit at minimum

///<summary>
///  The classes <b>DimSize</b> and <b>DimPt</b> are used similar
///  to the SIZE and POINT structures but with the difference to support
///  any desired dimension (while SIZE and POINT are 2-dimensional).
///  The <b>Size()</b> method will return the dimension.
///  The main difference between <b>DimSize</b> and <b>DimPt</b> is, that
///  if there are added new elements to a <b>DimSize</b> (or also just when
///  they are cleared using <b>Clear()</b>), they will be intialized with
///  the value 1, while new elements in a <b>DimPt</b> will be initialized
///  with zeros. <br/>
///  Note: When I am speaking of a dimension-extent (or short: dim-extent),
///  then I mean one element of the array of a <b>DimSize</b> or
///  <b>DimPt</b> object (because they are derived from <b>Array</b>).
///</summary>
class DimSize : public Array< dim_t >
{
    public :
        DimSize() : Array<dim_t>() {  }
        DimSize( const DimSize& dimsize ) : Array<dim_t>( dimsize ) {  } // copy constructor
        DimSize( _size_t_         cSize ) : Array<dim_t>( cSize   ) { Clear(); } // since in constructors the vtable does exist completely, force a call to <b>Clear()</b> in here.
        DimSize( dim_t *pElements, _size_t_ cSize ) : Array<dim_t>( pElements, cSize ) {  }

        void SetSize0( _size_t_ cSize ) { Array<dim_t>::SetSize( cSize ); } // new elements will be filled with zeros.
        virtual void SetSize( _size_t_ cSize ); // new elements will be filled with ones instead of zeros.

        virtual void Clear();   // all elements will be filled with ones instead of zeros.
        bool IsEmpty() const;   // returns TRUE if even one dimension-extent in the array is zero

        bool IsLessThan     ( const DimSize& dimsize2 ) const;
        bool IsLessEqualThan( const DimSize& dimsize2 ) const;

        static DimSize EvalMinimum( const DimSize& dimsize1, const DimSize& dimsize2 );
        static DimSize EvalMaximum( const DimSize& dimsize1, const DimSize& dimsize2 );
};

///<summary>
///  The classes <b>DimSize</b> and <b>DimPt</b> are used similar
///  to the SIZE and POINT structures but with the difference to support
///  any desired dimension (while SIZE and POINT are 2-dimensional).
///  The <b>Size()</b> method will return the dimension.
///  The main difference between <b>DimSize</b> and <b>DimPt</b> is, that
///  if there are added new elements to a <b>DimSize</b> (or also just when
///  they are cleared using <b>Clear()</b>), they will be intialized with
///  the value 1, while new elements in a <b>DimPt</b> will be initialized
///  with zeros. <br/>
///  Note: When I am speaking of a dimension-extent (or short: dim-extent),
///  then I mean one element of the array of a <b>DimSize</b> or
///  <b>DimPt</b> object (because they are derived from <b>Array</b>).
///</summary>
class DimPt : public Array< dim_t >
{
    public :
        DimPt() : Array<dim_t>() {  }
        DimPt( const DimPt& dimpt ) : Array<dim_t>( dimpt ) {  } // copy constructor
        DimPt( _size_t_     cSize ) : Array<dim_t>( cSize ) {  }
        DimPt( dim_t *pElements, _size_t_ cSize ) : Array<dim_t>( pElements, cSize ) {  }

        void SetToEndOfDim( const DimSize& dimsize );

        bool Increment( const DimPt& step, const DimSize& dimsize, bool *pbDimOverflow = NULL ) { return Increment( step, DimPt(dimsize.GetSize()), dimsize, pbDimOverflow ); }
        bool Increment( const DimPt& step, const DimPt&   dimpos,  const DimSize& dimsize, bool *pbDimOverflow = NULL );
        bool Decrement( const DimPt& step, const DimSize& dimsize, bool *pbDimUnderflow = NULL ) { return Decrement( step, DimPt(dimsize.GetSize()), dimsize, pbDimUnderflow ); }
        bool Decrement( const DimPt& step, const DimPt&   dimpos,  const DimSize& dimsize, bool *pbDimUnderflow = NULL );

        static DimPt EvalMinimum( const DimPt& dimpt1, const DimPt& dimpt2 );
        static DimPt EvalMaximum( const DimPt& dimpt1, const DimPt& dimpt2 );
};

typedef DWORD bitcounter_t;
#define bitcounter_t_bits       ( 8 * sizeof(bitcounter_t) )

///<summary>
///  This is a counter with a desired number of bits. You can increment and
///  decrement it and check the state of one or of all bits. <br/>
///  Note: The desired number of bits can be any value wanted ( >= 1 ).
///</summary>
class BitCounter : public Array< bitcounter_t >
{
    friend class BitCounter;

    protected :
        _size_t_ m_cBits;
        bool     m_bOverflowed;
    private :
        static DWORD s_aUnusedMask[];

    public :
        BitCounter( _size_t_ cbits )
            : Array<bitcounter_t>( (tmax(cbits, (_size_t_) 1) + bitcounter_t_bits - 1) / bitcounter_t_bits )
        { m_cBits = tmax(cbits, (_size_t_) 1); m_bOverflowed = false; }
        BitCounter( const BitCounter& bc ) : Array<bitcounter_t>( bc )
        { m_cBits = bc.m_cBits; m_bOverflowed = bc.m_bOverflowed; }

        // prefix inc/dec
        const BitCounter& operator++()      { Inc(); return *this; }
        const BitCounter& operator--()      { Dec(); return *this; }
        // postfix inc/dec
        const BitCounter  operator++(int n) { BitCounter result(*this); Inc(); return result; }
        const BitCounter  operator--(int n) { BitCounter result(*this); Dec(); return result; }

        void Inc();
        void Dec();

        int operator[]( int ix ) const { return GetBit( ix ); }
        int GetBit( _idx_ ix ) const;

        bool IsOverflowed() const { return m_bOverflowed; }

        _size_t_ CBits() const { return m_cBits; }
};




enum mem_t {
    MemT_Malloc         = 0,    // Allocate on the heap using #malloc()# or #realloc()#.
    MemT_New,                   // Allocate on the heap using the #new# operator.
    MemT_FileMapping,           // Allocate as file mapping object.
    MemT_External,              // Indicates foreign memory being pointed to.
    _memt_end,
    MemT_Default        = MemT_Malloc,
    // Note: Because of #MemDat_Settings::m_memtCreate# below, there are
    // at maximum 16 different memory types supported.

    // Remember: These type values above will be used also as indices
    // into #MemReallocator::s_aReallocFcts[]#. You will have to
    // update this array if you change the #mem_t# enum.

    MemT_Dont_Change    = 0x100,    // This will preserve the old memory type.
                                    //  If there is no old one, the memory type will be set to #MemT_Default#.
    MemT_Dont_Change_Dynamic,       // Like #MemT_Dont_Change#, but #MemT_External# type will be automatically converted to the creation type stored in #MemDat_Settings::m_memtCreate#

    MemT_None           = 0xffff,

    _pad_mem_t          = 0x7fff    // nevertheless, the compiler makes the enum 32 bits large. seems I cannot force it to 16 bit size
};

///<summary> To determine if the passed memory type indicates a real memory type or if it is a flag value. </summary>
inline bool IsRealMemType( mem_t memtype ) { return memtype < _memt_end; }
///<summary> To determine if a memory type is capable of very large allocation sizes. </summary>
inline bool IsVeryLargeMemType( mem_t memtype ) { return memtype == MemT_FileMapping; }
///<summary> To determine if for a specific memory type, header usage is supported (useful only for shareable memory). </summary>
inline bool IsMemTypeHeaderSupported( mem_t memtype ) { return memtype != MemT_External && memtype < _memt_end; }
///<summary> To determine if for a specific memory type, memory can be allocated (and reallocated). </summary>
inline bool IsMemTypeAllocateable( mem_t memtype ) { return memtype != MemT_External && memtype < _memt_end; }

#pragma pack (4)

struct MemDat_Settings;

// the absolute memory basics info
struct MemDat_Basic
{
    unsigned int m_MemType      : 16;   // type of the current memory (if currently no mem is allocated -> it is the type for memory to be allocated later)
    unsigned int m_bNamed       : 1;    // if the memory is named. Currently only MemT_FileMapping memories can have a name.
    PBYTE    m_pMemory;                 // points to current memory location
    _size_t_ m_cMemory;                 // size of current memory in bytes available for the user (this can be smaller than #m_cAllocated#) (does not include any header info size)
    _size_t_ m_cAllocated;              // real size of the memory allocated in bytes (including padding to reach multiples of #m_cBlock#) (does not include any header info size)
    WORD     m_wDataLockCount;          // if changing the content data of the memory is locked
    WORD     m_wReallocLockCount;       // if reallocating (and resizing) the memory is locked

    void Init( mem_t memtype );
    void Init( mem_t memtype, WORD wDataLockCount, WORD wReallocLockCount );
    void Init( const MemDat_Basic& other_basic );

    bool IsEmptyMem() const { return m_pMemory == NULL; }
    bool IsAnythingLocked() const { return m_wDataLockCount > 0 || m_wReallocLockCount > 0; }
};

// This is the (base class) struct with more memory
// infos additional to the #MemDat_Basic#.
struct MemDat_Extended {  };

// extended infos for filemapping memory
struct MemDat_FileMapping : public MemDat_Extended
{
    HANDLE  m_hFileMapping;
    // note: If the memory is named, you can check that by using #MemDat_Basic::m_bNamed#

    void Init() { m_hFileMapping = NULL; }
};

// And the union of all the available #MemDat_Extended# derivatives
union MemDat_UnionExtended
{
    MemDat_FileMapping  m_filemapping;

    bool Init( mem_t memtype );
    bool CopyFrom( mem_t memtype, const MemDat_Extended *ptr );
    static _size_t_ StructSize( mem_t memtype );
};

struct MemDat_Settings
{
    unsigned int m_cBlock           : 27;   // size of allocation block in bytes. Memory will be allocated in #m_cBlock# steps.
    unsigned int m_bClearNewMem     : 1;    // if to clear newly allocated (or appended) memory. (default = off)
    unsigned int m_memtCreate       : 4;    // memtype to use when to allocate new memory (if there is no memory allocated yet)

    void Init() { m_cBlock = 8; m_bClearNewMem = 0; m_memtCreate = MemT_Default; }
    _size_t_ GetBlockSize() const { return tmax( (_size_t_) (UINT) m_cBlock, (_size_t_) 1 ); }

    void CopyPropertiesFrom( const MemDat_Settings& settings )
    { m_cBlock = settings.m_cBlock; m_bClearNewMem = settings.m_bClearNewMem; m_memtCreate = settings.m_memtCreate; }
};

struct MemDat_ForeignMem
{
    unsigned int m_cForeignSize : 31;   // size of the foreign memory in bytes (if #m_bForeignMem# is 1)
    unsigned int m_bForeignMem  : 1;    // 1 if we are working with external memory.

    void Init() { Empty(); }
    void Empty() { m_bForeignMem = 0; m_cForeignSize = 0; }
};

struct MemDat_RefCnt
{
    typedef LONG refcnt_t;  // must be a LONG for the Interlocked functions (must be aligned on 32-bit boundary)
    enum refcnt_v { illegal = -1 };     // the reserved value for illegal/undefined value

    refcnt_t m_Counter;
    bool     m_bEnabled;

    void InitEnabled () { m_Counter = 1; m_bEnabled = true;  }
    void InitDisabled() { m_Counter = 1; m_bEnabled = false; }
    refcnt_t Inc() { return m_bEnabled ? (refcnt_t) InterlockedIncrement( &m_Counter ) : 1; }
    refcnt_t Dec() { return m_bEnabled ? (refcnt_t) InterlockedDecrement( &m_Counter ) : 1; }
    refcnt_t Get() { return m_bEnabled ? (refcnt_t) InterlockedCompareExchange( &m_Counter, 0, 0 ) : 1; }
    bool IsLastRef() { return m_bEnabled ? Get() <= 1 : true; }
    bool IsLastRef_Unsafe() const { return m_bEnabled ? m_Counter <= 1 : true; }
    bool IsShared() { return m_bEnabled ? Get() > 1 : false; }
    bool IsShared_Unsafe() const { return m_bEnabled ? m_Counter > 1 : false; }
};

// The header with id and basics info used for "shareable" memory.
// This header will be stored directly before the actual requested
// memory space when sharing memory.
struct MemDat_ShareHdr
{
    #define MEMDAT_SHHDRID      makeFourcc( 'S', 'M', 'E', 'M' )
    DWORD32         m_4cc;      // this must be MEMDAT_SHHDRID to indicate a valid header info
    DWORD           m_pid;      // the process-id of the process that created the memory. Other processes are not allowed to share that memory. It may be shared only within that process-id.
    MemDat_Basic    m_Basic;    // its #.m_pMemory# must point to the address directly following this header info (but *only* if reference counting is enabled)
    MemDat_RefCnt   m_RefCnt;
};
// Note: If the memory needs extended info (derived from #MemDat_Extended#)
// (which depends on memory type), then this info will be placed directly
// before the header info struct.

#define MEMDAT_TOTAL_HEADER_BLOCKING    8   // number of bytes the total header before the actual memory must be multiples of in size.

#pragma pack ()

///<summary>
///  Class <b>MemReallocator</b> is an internally only used class
///  to allocate memory and retrieve informations on it.
///</summary>
class MemReallocator
{
    public :
        MemReallocator() {  }

    // Realloc functions without changing MemType.
    // Passing zero as size will cause to free the memory.
    private :
        typedef reterr_t (*Realloc_T)( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc );
        static Realloc_T s_aReallocFcts[];
    protected :
        static UINT GetNumReallocFcts();
        static reterr_t ReallocSameMemType( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc );
    private :
        static reterr_t Realloc_Malloc     ( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc );
        static reterr_t Realloc_New        ( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc );
        static reterr_t Realloc_FileMapping( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc );
        static reterr_t Realloc_External   ( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc );

    // Reallocating memory that cannot be shared.
    public :
        static reterr_t ReallocNonShareableMemory( MemDat_Basic& basic, MemDat_Settings& settings, MemDat_UnionExtended& ext, _size_t_ newsize, mem_t memtype, PCWSTR pstrMemName = NULL, _size_t_ maxsize = (_size_t_) -1L, bool bOverrideLocks = false );
        static reterr_t ReallocShareableMemory( PBYTE& ptrMem, MemDat_Settings& settings, MemDat_ForeignMem& foreign, _size_t_ newsize, mem_t memtype, _size_t_ maxsize = (_size_t_) -1L, bool bOverrideLocks = false, bool bForceRealloc = false );
        static MemDat_ShareHdr *GetHeaderPtr_Safe( const BYTE *ptrMem, bool bIncRefCnt, bool bForbidLocked );
        static MemDat_ShareHdr *GetHeaderPtr     ( const BYTE *ptrMem ) { return ptrMem ? ((MemDat_ShareHdr*) ptrMem) - 1 : NULL; }
        static MemDat_ShareHdr *GetHeaderPtr_Safe( MemDat_Extended *& pExt, const BYTE *ptrMem, bool bIncRefCnt, bool bForbidLocked );
        static MemDat_ShareHdr *GetHeaderPtr     ( MemDat_Extended *& pExt, const BYTE *ptrMem );
        static MemDat_Extended *GetExtendedPtr   ( const MemDat_ShareHdr *pHdr )
        { return pHdr ? (MemDat_Extended*) (((PBYTE)pHdr) - MemDat_UnionExtended::StructSize( (mem_t) pHdr->m_Basic.m_MemType )) : NULL; }
        static _size_t_ GetTotalHeaderSize( const MemDat_ShareHdr *pHdr );
        static _size_t_ GetTotalHeaderSize( mem_t memtype );
        static bool GetBasic_Safe( MemDat_Basic& basic, MemDat_ShareHdr *pHdr );
        static bool DecRefCnt_Safe( MemDat_ShareHdr *pHdr );
};



class DynMem;
typedef retval_t<PBYTE>  _retPByte_t_;
typedef retval_t<BitPtr> _retBitPtr_t_;

#pragma pack (4)

// Note: The following memory classes #MemorySpace# and #MemoryShared#
// do not have any virtual methods to hold the size of the class
// as small as possible (especially for #MemoryShared#).

class MemorySpace
{
    private :
        MemDat_Basic         m_Basic;
        MemDat_UnionExtended m_Extended;
        MemDat_Settings      m_Settings;

    public :
        MemorySpace() { Init(); }
        MemorySpace( _size_t_ count ) { Init(); SetSize( count ); }
        MemorySpace( _size_t_ count, mem_t memtype ) { Init(); SetSizeAndType( count, memtype ); }
        MemorySpace( _size_t_ count, mem_t memtype, PCSTR  pstrMemName )
        { Init(); SetSizeAndType( count, memtype, pstrMemName ); }
        MemorySpace( _size_t_ count, mem_t memtype, PCWSTR pstrMemName )
        { Init(); SetSizeAndType( count, memtype, pstrMemName ); }
        MemorySpace( LPCVOID ptr, _size_t_ count, mem_t memtype = MemT_Default )
        { Init(); if( memtype == MemT_External ) SetExternalMem(ptr, count); else { SetMemType(memtype); Add(ptr, count); } }
        MemorySpace( const MemorySpace& mem ) { Init(); CopyFrom( mem ); } // copy-ctor
        MemorySpace& operator =( const MemorySpace& mem ) { CopyFrom( mem ); return *this; }
        ~MemorySpace() { ForceFree(); }

        const BYTE  *ConstPtr() const { return m_Basic.m_pMemory; }
        _retPByte_t_ WritePtr();

        bool IsDataLocked   () const { return m_Basic.m_wDataLockCount    > 0; } // if the memory content data is locked (may not be changed) (plus: memory size decreasing is forbidden or else we would loose data bytes)
        bool IsReallocLocked() const { return m_Basic.m_wReallocLockCount > 0; } // if reallocating the memory of any kind is locked

        _size_t_ GetSize() const          { return m_Basic.m_cMemory;    } // Returns the size *not* including any padding bytes (caused by block size).
        _size_t_ GetAllocatedSize() const { return m_Basic.m_cAllocated; } // Returns the size *including* the padding bytes (caused by block size or maxsize).
        reterr_t SetSize       ( _size_t_ count ) { return MemReallocator::ReallocNonShareableMemory( m_Basic, m_Settings, m_Extended, count, MemT_Dont_Change_Dynamic ); }
        reterr_t SetSizeAndType( _size_t_ count, mem_t memtype, PCSTR  pstrMemName );
        reterr_t SetSizeAndType( _size_t_ count, mem_t memtype, PCWSTR pstrMemName = NULL )
        { return MemReallocator::ReallocNonShareableMemory( m_Basic, m_Settings, m_Extended, count, memtype, pstrMemName ); }
        reterr_t ForceEnlargeToSize( _size_t_ count );
        reterr_t SetExternalMem( LPCVOID ptr, _size_t_ count );
        reterr_t Free() { return MemReallocator::ReallocNonShareableMemory( m_Basic, m_Settings, m_Extended, 0, MemT_Dont_Change_Dynamic ); }
        reterr_t ForceFree()    // treat this with caution. Use it only if absolutely necessary. It will ignore any locks.
        { return MemReallocator::ReallocNonShareableMemory( m_Basic, m_Settings, m_Extended, 0, MemT_Dont_Change_Dynamic, NULL, (_size_t_) -1L, true ); }
        bool     IsEmpty() const { return m_Basic.IsEmptyMem(); } // if some memory is allocated or if it is empty

        _size_t_ GetBlockSize() const { return m_Settings.GetBlockSize(); }
        void     SetBlockSize( _size_t_ blocksize = 1 ) { m_Settings.m_cBlock = (UINT) blocksize; }

        reterr_t     SetData( LPCVOID ptr, _size_t_ count );
        reterr_t     Add( _size_t_ count );
        reterr_t     Add( LPCVOID ptr, _size_t_ count );
        _retPByte_t_ Add_Ptr( _size_t_ count = 1 );
        reterr_t     Append( const MemorySpace& mem );
        reterr_t     Insert( _idx_ idx, _size_t_ count, LPCVOID ptr );
        _retPByte_t_ Insert_Ptr( _idx_ idx, _size_t_ count = 1 );
        reterr_t     Remove( _idx_ idx, _size_t_ count = 1 );
        reterr_t     Clear();
        reterr_t     Clear( _idx_ idx, _size_t_ count );
        reterr_t     ClearPadding();
        reterr_t     ClearAllocatedBuffer();

        mem_t    GetMemType() const          { return (mem_t) m_Basic.m_MemType; }
        mem_t    GetMemTypeCreate() const    { return (mem_t) m_Settings.m_memtCreate; } // the memory type this object will use when it has to create completely new memory
        reterr_t SetMemType( mem_t memtype )
        { return MemReallocator::ReallocNonShareableMemory( m_Basic, m_Settings, m_Extended, m_Basic.m_cMemory, memtype ); }

        HANDLE   GetFileMappingHandle() const
        { return m_Basic.m_MemType == MemT_FileMapping ? m_Extended.m_filemapping.m_hFileMapping : NULL; }

        reterr_t CopyPropertiesFrom( const MemorySpace& mem );
        reterr_t CopyFrom( const MemorySpace& mem );
        reterr_t Exchange( MemorySpace& mem );

        bool IsClearNewMemEnabled() const     { return m_Settings.m_bClearNewMem != 0; }
        void EnableClearNewMem   ( bool bEn ) { m_Settings.m_bClearNewMem = bEn; }

    // misc
    protected :
        bool CheckRangeLimits( _idx_& idx, _size_t_& count ) { return CheckRangeLimits( idx, count, m_Basic.m_cMemory, 0 ); }
    public :
        static bool CheckRangeLimits( _idx_& idx, _size_t_& count, _size_t_ memsize, _idx_ left_limit = 0 );

    protected :
        void Init()
        { m_Settings.Init(); m_Basic.Init( (mem_t) m_Settings.m_memtCreate ); m_Extended.Init( (mem_t) m_Basic.m_MemType ); }

    // Locking.
    // You can use #MemLockData#, #MemLockRealloc#
    // or you can use the methods below to achieve locks.
    public :
        reterr_t LockData     () { m_Basic.m_wDataLockCount++;    return true; }
        reterr_t UnlockData   () { m_Basic.m_wDataLockCount--;    return true; }
        reterr_t LockRealloc  () { m_Basic.m_wReallocLockCount++; return true; }
        reterr_t UnlockRealloc() { m_Basic.m_wReallocLockCount--; return true; }
};

class MemoryShared
{
    friend class DynMem;

    private :
        PBYTE             m_pMemory;
        MemDat_ForeignMem m_ForeignMem;
        MemDat_Settings   m_Settings;

    public :
        MemoryShared() { Init();  }
        MemoryShared( _size_t_ count ) { Init(); SetSize( count ); }
        MemoryShared( _size_t_ count, mem_t memtype ) { Init(); SetSizeAndType( count, memtype ); }
        MemoryShared( LPCVOID ptr, _size_t_ count, mem_t memtype = MemT_Default )
        { Init(); if( memtype == MemT_External ) SetExternalMem(ptr, count); else { SetMemType(memtype); Add(ptr, count); } }
        MemoryShared( const MemoryShared& mem ) { Init(); ShareFrom( mem ); } // copy-ctor
        MemoryShared& operator =( const MemoryShared& mem ) { ShareFrom( mem ); return *this; }
        ~MemoryShared() { ForceFree(); }

        bool     IsShared() const;
        reterr_t Isolate             () { return IsolateToSizeAndType( GetSize(), MemT_Dont_Change_Dynamic ); } // forces the object to have its own non-shared memory.
        reterr_t IsolateToSize       ( _size_t_ count ) { return IsolateToSizeAndType( count, MemT_Dont_Change_Dynamic ); }
        reterr_t IsolateToSizeAndType( _size_t_ count, mem_t memtype );

        const BYTE  *ConstPtr() const { return m_pMemory; }
        _retPByte_t_ WritePtr();
        _retPByte_t_ WritePtr_Retain();
        _retPByte_t_ WritePtr_Retain( _size_t_ count );
        _retPByte_t_ WritePtr_RetainMinimum( _size_t_ minsize );
        _retPByte_t_ WritePtr_Discard( _size_t_ count );
        _retPByte_t_ WritePtr_Discard( _size_t_ count, mem_t memtype );

        bool IsDataLocked   () const; // if the memory content data is locked (may not be changed) (plus: memory size decreasing is forbidden or else we would loose data bytes)
        bool IsReallocLocked() const; // if reallocating the memory of any kind is locked

        _size_t_ GetSize() const; // Returns the size *not* including any padding bytes (caused by block size).
        _size_t_ GetAllocatedSize() const; // Returns the size *including* the padding bytes (caused by block size or maxsize).
        reterr_t SetSize       ( _size_t_ count ) { return MemReallocator::ReallocShareableMemory( m_pMemory, m_Settings, m_ForeignMem, count, MemT_Dont_Change_Dynamic ); }
        reterr_t SetSizeAndType( _size_t_ count, mem_t memtype )
        { return MemReallocator::ReallocShareableMemory( m_pMemory, m_Settings, m_ForeignMem, count, memtype ); }
        reterr_t ForceEnlargeToSize( _size_t_ count );
        reterr_t SetExternalMem( LPCVOID ptr, _size_t_ count );
        reterr_t AttachPointer( LPCVOID ptr );
        static bool IsAttachablePointer( LPCVOID ptr, MemDat_Basic *pBasic = NULL );
        reterr_t Free() { return MemReallocator::ReallocShareableMemory( m_pMemory, m_Settings, m_ForeignMem, 0, MemT_Dont_Change_Dynamic ); }
        reterr_t ForceFree()    // treat this with caution. Use it only if absolutely necessary. It will ignore any locks.
        { return MemReallocator::ReallocShareableMemory( m_pMemory, m_Settings, m_ForeignMem, 0, MemT_Dont_Change_Dynamic, (_size_t_) -1L, true ); }
        bool     IsEmpty() const { return m_pMemory == NULL; } // if some memory is allocated or if it is empty

        _size_t_ GetBlockSize() const { return m_Settings.GetBlockSize(); }
        void     SetBlockSize( _size_t_ blocksize = 1 ) { m_Settings.m_cBlock = (UINT) blocksize; }

        reterr_t     SetData( LPCVOID ptr, _size_t_ count );
        reterr_t     Add( _size_t_ count );
        reterr_t     Add( LPCVOID ptr, _size_t_ count );
        _retPByte_t_ Add_Ptr( _size_t_ count = 1 );
        reterr_t     Append( const MemoryShared& mem );
        reterr_t     Insert( _idx_ idx, _size_t_ count, LPCVOID ptr );
        _retPByte_t_ Insert_Ptr( _idx_ idx, _size_t_ count = 1 );
        reterr_t     Remove( _idx_ idx, _size_t_ count = 1 );
        reterr_t     Clear();
        reterr_t     Clear( _idx_ idx, _size_t_ count );
        reterr_t     ClearPadding();
        reterr_t     ClearAllocatedBuffer();

        mem_t    GetMemType() const;
        mem_t    GetMemTypeCreate() const { return (mem_t) m_Settings.m_memtCreate; } // the memory type this object will use when it has to create completely new memory
        reterr_t SetMemType( mem_t memtype )
        { return MemReallocator::ReallocShareableMemory( m_pMemory, m_Settings, m_ForeignMem, GetSize(), memtype ); }

        HANDLE   GetFileMappingHandle() const;
        _size_t_ GetInternalContentOffset() const;

        reterr_t CopyPropertiesFrom( const MemoryShared& mem );
        reterr_t ShareFrom( const MemoryShared& mem );
        reterr_t Exchange( MemoryShared& mem );

        bool IsClearNewMemEnabled() const     { return m_Settings.m_bClearNewMem != 0; }
        void EnableClearNewMem   ( bool bEn ) { m_Settings.m_bClearNewMem = bEn; }

    // misc
    public :
        static bool CheckRangeLimits( _idx_& idx, _size_t_& count, _size_t_ memsize, _idx_ left_limit = 0 )
        { return MemorySpace::CheckRangeLimits( idx, count, memsize, left_limit ); }

    protected :
        void Init()
        { m_Settings.Init(); m_pMemory = NULL; m_ForeignMem.Init(); }

    // Locking.
    // You can use #MemLockData#, #MemLockRealloc#
    // or you can use the methods below to achieve locks.
    public :
        reterr_t LockData     ();
        reterr_t UnlockData   ();
        reterr_t LockRealloc  ();
        reterr_t UnlockRealloc();
};

///<summary>
///  <b>DynMem</b> is an N-dimensional, bit-oriented, resizeable and
///  shareable memory class with selectable memory type.
///  <br/><br/>
///
///  Handles a resizeable memory array.
///  The array can have any desired dimension.
///  The data memory can be allocated as different memory types.
///  <br/>
///  Since the memory can be of any desired dimension, the dimension
///  extents themselves must be stored in memory. This will be done
///  on the heap and you cannot select any different memory type
///  for the dimension extent memories.
///  But if the dimension is 3 or less, then the extents will be
///  stored in the object itself not needing any addition heap memory
///  for the extents. (This is done to reduce heap fragmentation).
///</summary>

class DynMem
{
    protected :
    // main data-array memory settings
        BitLen  m_cMemUnit;             // Size of a memory-unit in bit/bytes
        BitLen  m_cAlign;               // Alignment for all dimension-extents.
                                         // (Note: In reality, this is only for the first dim-extent, because all following dim-extents
                                         // will be automatically aligned to this value because they are multiples of the first dim-extent then.)
                                         // Can be in bits even if #m_cMemUnits# is in bytes (in that case, the data memory will be non-byteable)
                                         // If empty -> no alignment.

    // additional memory-size-affecting memory settings
        BitLen  m_cHeaderArea;          // If a memory hole should be held before the real memory content for additional use

        // Note: Following values of the variables above may *not* be changed when
        // currently a memory array is allocated (dimension other than zero and
        // no header):
        //  - m_cMemUnit
        //  - m_cAlign


    public :
        DynMem() { Init(); }
        DynMem( mem_t memtype ) : m_Mem(0, memtype) { Init(); }
        DynMem( const DynMem& dm ) { Init(); ShareFrom( dm ); }
        DynMem& operator =( const DynMem& dm ) { ShareFrom( dm ); return *this; }
        ~DynMem() { ForceFree(); }

        bool IsByteable() const; // if handling with byte-addresses (char *) is possible or if bit-addresses (BitPtr) must be used
        // Note: If a #DynMem# is byteable: -> for the #BitPtr# and #BitLen# values you
        // can work with the byte values and ignore the bit values.

        bool     IsShared() const { return m_Mem.IsShared(); }
        reterr_t Isolate ()       { return m_Mem.Isolate (); } // forces the object to have its own non-shared memory.

        BitCPtr       ConstHeaderPtr() const;  // pointer to the header (or NULL if no header available)
        BitCPtr       ConstPtr() const;        // pointer of the start of the array (*after* the header)
        BitCPtr       ConstPtrAt( _idx_ ix ) const;
        BitCPtr       ConstPtrAt( _idx_ ix1, _idx_ ix2 ) const;
        BitCPtr       ConstPtrAt( const DimPt& pos ) const;
        BitCPtr       ConstPtrOfArea( _idx_ ix, _size_t_ ci ) const;
        BitCPtr       ConstPtrOfArea( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2 ) const;
        BitCPtr       ConstPtrOfArea( const DimPt& pos, const DimSize& size ) const;
        _retBitPtr_t_ WriteHeaderPtr();        // pointer to the header (or NULL if no header available)
        _retBitPtr_t_ WritePtr();              // pointer of the start of the array (*after* the header)
        _retBitPtr_t_ WritePtrAt( _idx_ ix );
        _retBitPtr_t_ WritePtrAt( _idx_ ix1, _idx_ ix2 );
        _retBitPtr_t_ WritePtrAt( const DimPt& pos );
        _retBitPtr_t_ WritePtrOfArea( _idx_ ix, _size_t_ ci );
        _retBitPtr_t_ WritePtrOfArea( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2 );
        _retBitPtr_t_ WritePtrOfArea( const DimPt& pos, const DimSize& size );

        //operator BitCPtr() const { return ConstPtr(); } // pointer of the start of the array (*after* the header)
        //operator BitPtr ()       { return WritePtr(); } // pointer of the start of the array (*after* the header)

        bool IsDataLocked   () const { return m_Mem.IsDataLocked   (); } // if the memory content data is locked (may not be changed) (plus: memory size decreasing is forbidden or else we would loose data bytes)
        bool IsReallocLocked() const { return m_Mem.IsReallocLocked(); } // if reallocating the memory of any kind is locked

        mem_t    GetMemType() const          { return m_Mem.GetMemType(); }
        mem_t    GetMemTypeCreate() const    { return m_Mem.GetMemTypeCreate(); } // the memory type this object will use when it has to create completely new memory
        reterr_t SetMemType( mem_t memtype ) { return m_Mem.SetMemType( memtype ); } // Note: This method works also when memory is allocated already.

        HANDLE   GetFileMappingHandle() const { return m_Mem.GetFileMappingHandle(); }
        _size_t_ GetInternalHeaderOffsetInBytes() const;
        BitLen   GetInternalContentDataOffset() const;

        BitLen   GetMemUnitSize() const { return m_cMemUnit; }
        reterr_t SetMemUnitSizeInBytes( _size_t_ bytememunitsize ) { return SetMemUnitSize( 8 * bytememunitsize ); }
        reterr_t SetMemUnitSize( BitLen memunitsize );

        _size_t_ GetBlockSizeInBytes() const { return m_Mem.GetBlockSize(); }
        void SetBlockSizeInBytes( _size_t_ blocksize = 1 ) // This will work at any time, but it will not cause any memory to be reallocated.
        { m_Mem.SetBlockSize( blocksize ); }

        BitLen   GetAlignment() const { return m_cAlign; } // returns the alignment value in bit/bytes for each row (= just for the first dim-extent, for following dim-extents you cannot specify additional alignments)
        reterr_t SetAlignment( BitLen alignment );        // sets    the alignment value in bit/bytes for each row (= just for the first dim-extent, for following dim-extents you cannot specify additional alignments)

        reterr_t SetSize( _size_t_ size ); // set array size with dimension 1; header size and mem type remain as before
        reterr_t SetSize( _size_t_ size1, _size_t_ size2 ); // set array size with dimension 2; header size and mem type remain as before
        reterr_t SetSize( const DimSize& size ) { return SetSize( size, m_cHeaderArea, MemT_Dont_Change_Dynamic ); } // set array size; header size and mem type remain as before
        reterr_t SetSize( const DimSize& size, BitLen headersize, mem_t memtype ); // set array size, header size and mem type all at once
        const DimSize& GetSize() const { return m_oDim; } // get size (and dimension) of the whole array (without header and block-sizing)
        _size_t_ GetExtentSize( _idx_ nExtent ) const { return (_size_t_) m_oDim[nExtent]; } // get the count of items of *one* dimension-extent

        BitLen   GetHeaderSize() const { return m_cHeader; } // size of only the header in bit/bytes
        reterr_t SetHeaderSizeInBytes( _size_t_ byteheadersize ) { return SetHeaderSize( 8 * byteheadersize ); }
        reterr_t SetHeaderSize( BitLen headersize ); // set size (in bit/bytes) of the header not affecting the array (except moving the array to make place for the header)

        BitLen   GetArraySize    () const { return m_cArraySize; } // size of whole array in bit/bytes (without header and block-sizing)
        BitLen   GetMemSize      () const { return EvalNettoMemSize( m_cArraySize, m_cHeader ); } // size (in bit/bytes) of the whole memory (header + array) but without block-sizing
        _size_t_ GetAllocatedSize() const { return m_Mem.GetAllocatedSize(); } // size (in bytes) of the whole memory (header + array) *including* the block-sizing, so the completely allocated memory
        _size_t_ GetDimension    () const { return m_oDim.GetSize(); } // gets the dimension of the whole array

        reterr_t Free     () { return      Free_Array( MemT_Dont_Change_Dynamic ); } // free the complete memory (array and header)
        reterr_t ForceFree() { return ForceFree_Array( MemT_Dont_Change_Dynamic ); } // free the complete memory (array and header)

        reterr_t      Add( _size_t_ add ); // adds new memory-units to the first dim-extent.
        reterr_t      Add( _size_t_ add1, _size_t_ add2 ); // adds new memory-units to the 1st + 2nd dim-extent
        reterr_t      Add( const DimSize& add ); // adds new memory-units to the array
        _retBitPtr_t_ Add_Ptr( _size_t_ add = 1 ); // adds new memory-units to the first dim-extent. Returns the pointer to the first new unit.
        reterr_t      Insert( _idx_ ix, _size_t_ ci ); // insert new memory-units into the current memory
        reterr_t      Insert( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2 ); // insert new memory-units into the current memory
        reterr_t      Insert( const DimPt& pos, const DimSize& size ); // insert new memory-units into the current memory
        _retBitPtr_t_ Insert_Ptr( _idx_ ix, _size_t_ ci = 1 ); // insert new memory-units into the current memory
        reterr_t      Remove( _idx_ ix, _size_t_ ci = 1 ); // remove memory-units from the memory array
        reterr_t      Remove( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2 ); // remove memory-units from the memory array
        reterr_t      Remove( const DimPt& pos, const DimSize& size ); // remove memory-units from the memory array

        reterr_t Move( _idx_ idest, _idx_ isrce, _size_t_ size, bool bClip = false ); // move a part of the 1-dimensional memory within the 1-dimensional memory array
        reterr_t Move( _idx_ idest1, _idx_ idest2, _idx_ isrce1, _idx_ isrce2, _size_t_ size1, _size_t_ size2, bool bClip = false ); // move a part of the 2-dimensional memory within the 2-dimensional memory array
        reterr_t Move( DimPt dest, DimPt srce, DimSize size, bool bClip = false ); // move a part of the memory within the memory array

        reterr_t ClearHeader(); // clears only the header
        reterr_t Clear(); // clears the whole memory array, but *not* the header
        reterr_t Clear( _idx_ ix, _size_t_ ci, bool bClip = false ); // clears the specified area of the memory array
        reterr_t Clear( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2, bool bClip = false ); // clears the specified area of the memory array
        reterr_t Clear( DimPt pos, DimSize size, bool bClip = false ); // clears the specified area of the memory array
        reterr_t ClearPadding() { return m_Mem.ClearPadding(); } // see remarks to #MemoryShared::ClearPadding()#

        reterr_t CopyPropertiesFrom( const DynMem& dm ); // copies only the properties (memunit-size, mem-type, blocksize etc.) from another 'DynMem' object
        reterr_t ShareFrom( const DynMem& dm ); // deletes all previous memory and shares another #DynMem#'s memory.
        reterr_t Exchange( DynMem& dm ); // exchanges all contents (including the memory) of 'this' and a second 'DynMem' object

        // Note: #IsClearNewMemEnabled()# and #EnableClearNewMem()# methods
        // do not exist, because #DynMem# objects always clear the new memory
        // (but not padding bytes).


    protected :
        void   Init();
        BitLen EvalArraySize() const { return EvalArraySize( m_oDim ); }
        BitLen EvalArraySize( const DimSize& oDim ) const;
        BitLen EvalNettoMemSize( const DimSize& oDim, BitLen cHeader ) const;
        BitLen EvalNettoMemSize( BitLen cArraySize, BitLen cHeader ) const;
        BitLen EvalArrayOffset( const DimPt& oPos, const DimSize& oDim, bool bAllowOutsideArray = false ) const;
        BitLen EvalOffset( const DimPt& oPos, const DimSize& oDim, BitLen cHeader, bool bAllowOutsideArray = false ) const;
        BitLen EvalOffset( const DimPt& oPos, bool bAllowOutsideArray = false ) const { return EvalOffset( oPos, m_oDim, m_cHeader, bAllowOutsideArray ); }
        BitLen EvalStartOfArray() const { return m_cHeader; }


    // Managing n-dimensional array memory
    private :
        BitLen  m_cHeader;              // the current size of the header
        DimSize m_oDim;                 // the dimension of the dynamic memory array
        BitLen  m_cArraySize;           // size of the array in bit/bytes (without header)
    protected :
        reterr_t ReAlloc_Array  ( const DimSize& oDim_Desired,
                                  BitLen cHeader_Desired,
                                  mem_t memtype_desired );
        reterr_t Free_Array     ( mem_t memtype ); // after freeing any memory, the memtype will be set to the desired type
        reterr_t ForceFree_Array( mem_t memtype ); // after freeing any memory, the memtype will be set to the desired type


    // Linear Memory allocation and management
    // The following objects and methods should be used
    // only by the #*_Array()# methods above
    // (except for querying or updating states like locking).
    private :
        MemoryShared m_Mem;


    // Locking.
    // You can use #MemLockData#, #MemLockRealloc#
    // or you can use the methods below to achieve locks.
    public :
        reterr_t LockData     () { return m_Mem.LockData     (); }
        reterr_t UnlockData   () { return m_Mem.UnlockData   (); }
        reterr_t LockRealloc  () { return m_Mem.LockRealloc  (); }
        reterr_t UnlockRealloc() { return m_Mem.UnlockRealloc(); }
};

#pragma pack ()



// The following MemLock* classes can be used to lock
//  the memory objects above.
// You can create multiple MemLock objects of the same type
//  (e.g. two MemLockData objects) for the same memory object,
// CAUTION: Locking will cause the memory to isolated (for
//  MemoryShared and DynMem objects) and therefore could cause the
//  memory to be reallocated.
//  So previous pointers to the memory content could be invalid.

enum memlock_obj_t {
    MemLockObjMemorySpace       = 1,
    MemLockObjMemoryShared,
    MemLockObjDynMem
};

class MemLockData
{
    private :
        PVOID         m_pMemory;
        memlock_obj_t m_objectT;
        bool          m_bLocked;

    public :
        MemLockData( MemorySpace  *pMem, bool bInitialLock = true );
        MemLockData( MemoryShared *pMem, bool bInitialLock = true );
        MemLockData( DynMem       *pMem, bool bInitialLock = true );
        ~MemLockData() { Unlock(); }

        bool Lock();
        bool Unlock();

    private :
        // Since we use SEH (__try/__except) in Lock() and Unlock(), local object unwinding
        // is not allowed in these functions. But reterr_t (cast to bool) will need
        // unwinding (its dtor will be called), so we have to include that
        // in separate functions _Lock() and _Unlock().
        bool _Lock();
        bool _Unlock();
};

class MemLockRealloc
{
    private :
        PVOID         m_pMemory;
        memlock_obj_t m_objectT;
        bool          m_bLocked;

    public :
        MemLockRealloc( MemorySpace  *pMem, bool bInitialLock = true );
        MemLockRealloc( MemoryShared *pMem, bool bInitialLock = true );
        MemLockRealloc( DynMem       *pMem, bool bInitialLock = true );
        ~MemLockRealloc() { Unlock(); }

        bool Lock();
        bool Unlock();

    private :
        // Since we use SEH (__try/__except) in Lock() and Unlock(), local object unwinding
        // is not allowed in these functions. But reterr_t (cast to bool) will need
        // unwinding (its dtor will be called), so we have to include that
        // in separate functions _Lock() and _Unlock().
        bool _Lock();
        bool _Unlock();
};




}

#endif

#ifndef _argon_no_includes
    #undef _argonErrors_no_Part2
    #include "argonErrors.h"
    // We need to include "argonErrors.h" Part2 *after* the memory objects.
#endif
