/*
    File:               argonRectF.h
    Description:        Point, Size and Rectangle classes for floating point values.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         03.11.2006
    last updated:       03.11.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonRectF_h
#define __argonRectF_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonRect.h"
#endif

namespace ArgonLibrary
{




typedef struct tagPOINTF
{
    float x;
    float y;
} POINTF, *PPOINTF;

typedef struct tagSIZEF
{
    float cx;
    float cy;
} SIZEF, *PSIZEF;

typedef struct tagRECTF
{
    float left;
    float top;
    float right;
    float bottom;
} RECTF, *PRECTF;

typedef const POINTF *PCPOINTF;
typedef const SIZEF  *PCSIZEF;
typedef const RECTF  *PCRECTF;

namespace privateRectF
{
    template< class T > void tswap( T& a, T& b ) { T h = a; a = b; b = h; }
    template< class T > T tmin( T a, T b ) { return ( a <= b )? a : b; }
    template< class T > T tmax( T a, T b ) { return ( a >= b )? a : b; }

    inline float roundf( float x ) { return floorf(x + 0.5f); }
}

class _PtF_;
class _SizeF_;
class _RectF_;

class _PtF_ : public tagPOINTF
{
    public :
        _PtF_() {  } // randomly filled
        _PtF_( float _x, float _y ) { x = _x; y = _y; }
        _PtF_( POINTF         _pt ) { *(POINTF*)this = _pt; }
        _PtF_( POINT          _pt ) { x = (float) _pt.x; y = (float) _pt.y; }
        _PtF_( SIZEF        _size ) { *(POINTF*)this = *(POINTF*) &_size; }
        _PtF_( SIZE         _size ) { x = (float) _size.cx; y = (float) _size.cy; }
        _PtF_( DWORD           dw ) { x = (float) (SHORT) LOWORD(dw); y = (float) (SHORT) HIWORD(dw); }
        _PtF_( const WINDOWPOS *pwp ) { x = pwp ? (float) pwp->x : 0.0f; y = pwp ? (float) pwp->y : 0.0f; }

        _PtF_& Set     ( float _x, float _y ) { x = _x; y = _y; return *this; }
        _PtF_& SetX    ( float _x ) { x = _x; return *this; }
        _PtF_& SetY    ( float _y ) { y = _y; return *this; }
        _PtF_& SetEmpty()           { x = y = 0.0f; return *this; }
        _PtF_& Offset( float xofs, float yofs ) { x += xofs;    y += yofs;    return *this; }
        _PtF_& Offset( POINTF  pt )             { x += pt.x;    y += pt.y;    return *this; }
        _PtF_& Offset( SIZEF size )             { x += size.cx; y += size.cy; return *this; }
        _PtF_& Scale( float xscale, float yscale ) { x *= xscale;   y *= yscale;   return *this; }
        _PtF_& Scale( SIZEF  scale )               { x *= scale.cx; y *= scale.cy; return *this; }
        _PtF_& Resize( float size ) { x *= size; y *= size; return *this; }
        _PtF_& Divide( float invsize ) { if( invsize != 0.0f ) { x /= invsize; y /= invsize; } return *this; }
        _PtF_& StayInPseudoRect( RECTF rc );

        operator PPOINTF() { return this; }

        bool IsZero() const { return( x == 0.0f && y == 0.0f ); }

        _SizeF_&       GetAsSize();
        const _SizeF_& GetAsSize() const;
        _RectF_ GetEmptyRect () const; // top-left and bottom-right point will be set to 'this' position
        _RectF_ GetOneDotRect() const; // top-left will be set to 'this' position and the rect's size will be 1 pixel in each direction

        _PtF_& GetTL( RECTF rc ) { x = rc.left;  y = rc.top;    return *this; }
        _PtF_& GetTR( RECTF rc ) { x = rc.right; y = rc.top;    return *this; }
        _PtF_& GetBL( RECTF rc ) { x = rc.left;  y = rc.bottom; return *this; }
        _PtF_& GetBR( RECTF rc ) { x = rc.right; y = rc.bottom; return *this; }

        _PtF_& Absolute() { x = fabsf(x); y = fabsf(y); return *this; }
        _PtF_& Negate  () { x = -x; y = -y; return *this; }
        _PtF_& Flip    () { privateRectF::tswap( x, y ); return *this; }

        bool operator ==( POINTF pt ) const { return( x == pt.x && y == pt.y ); }
        bool operator !=( POINTF pt ) const { return( x != pt.x || y != pt.y ); }
        _PtF_& operator +=( SIZEF      size ) { x += size.cx; y += size.cy; return *this; }
        _PtF_& operator -=( SIZEF      size ) { x -= size.cx; y -= size.cy; return *this; }
        _PtF_& operator +=( POINTF       pt ) { x += pt.x;    y += pt.y;    return *this; }
        _PtF_& operator -=( POINTF       pt ) { x -= pt.x;    y -= pt.y;    return *this; }
        _PtF_& operator *=( float     scale ) { x *= scale;   y *= scale;   return *this; }
        _PtF_& operator *=( double    scale ) { x = (float) (x * scale); y = (float) (y * scale); return *this; }
        _PtF_& operator /=( float  invscale ) { if( invscale != 0.0f ) { x /= invscale; y /= invscale; } return *this; }
        _PtF_& operator /=( double invscale ) { if( invscale != 0.0  ) { x = (float) (x / invscale); y = (float) (y / invscale); } return *this; }

        _PtF_ operator +( SIZEF      size ) const { return _PtF_( x + size.cx, y + size.cy ); }
        _PtF_ operator -( SIZEF      size ) const { return _PtF_( x - size.cx, y - size.cy ); }
        _PtF_ operator -()                  const { return _PtF_( -x, -y ); }
        _PtF_ operator +( POINTF       pt ) const { return _PtF_( x + pt.x,  y + pt.y  ); }
        _PtF_ operator *( float     scale ) const { return _PtF_( x * scale, y * scale ); }
        _PtF_ operator *( double    scale ) const { return _PtF_( (float) (x * scale), (float) (y * scale) ); }
        _PtF_ operator /( float  invscale ) const { if( invscale != 0.0f ) return _PtF_( x / invscale, y / invscale ); else return *this; }
        _PtF_ operator /( double invscale ) const { if( invscale != 0.0  ) return _PtF_( (float) (x / invscale), (float) (y / invscale) ); else return *this; }

        _SizeF_ operator -( POINTF pt ) const;

        _RectF_ operator +( PCRECTF prc ) const;
        _RectF_ operator -( PCRECTF prc ) const;

        _Pt_ toPt      () const { return _Pt_( (int) x, (int) y ); }
        _Pt_ toPt_floor() const { return _Pt_( (int) floorf(x), (int) floorf(y) ); }
        _Pt_ toPt_ceil () const { return _Pt_( (int) ceilf (x), (int) ceilf (y) ); }
        _Pt_ toPt_round() const { return _Pt_( (int) privateRectF::roundf(x), (int) privateRectF::roundf(y) ); }
};

class _SizeF_ : public tagSIZEF
{
    public :
        _SizeF_() {  } // random filled
        _SizeF_( float _cx, float _cy ) { cx = _cx; cy = _cy; }
        _SizeF_( SIZEF          _size ) { * (SIZEF*) this = _size; }
        _SizeF_( SIZE           _size ) { cx = (float) _size.cx; cy = (float) _size.cy; }
        _SizeF_( POINTF           _pt ) { * (SIZEF*) this = * (SIZEF*) &_pt; }
        _SizeF_( POINT            _pt ) { cx = (float) _pt.x; cy = (float) _pt.y; }
        _SizeF_( DWORD             dw ) { cx = (float) (SHORT) LOWORD(dw); cy = (float) (SHORT) HIWORD(dw); }
        _SizeF_( RECTF            _rc ) { cx = _rc.right - _rc.left; cy = _rc.bottom - _rc.left; }
        _SizeF_( const WINDOWPOS *pwp ) { cx = pwp ? (float) pwp->cx : 0.0f; cy = pwp ? (float) pwp->cy : 0.0f; }

        _SizeF_& Set     ( float _cx, float _cy ) { cx = _cx; cy = _cy; return *this; }
        _SizeF_& SetCx   ( float _cx ) { cx = _cx; return *this; }
        _SizeF_& SetCy   ( float _cy ) { cy = _cy; return *this; }
        _SizeF_& SetEmpty()          { cx = cy = 0; return *this; }
        _SizeF_& Scale( float xscale, float yscale ) { cx *= xscale;   cy *= yscale;   return *this; }
        _SizeF_& Scale( SIZEF  scale )               { cx *= scale.cx; cy *= scale.cy; return *this; }
        _SizeF_& Resize( float size ) { cx *= size; cy *= size; return *this; }
        _SizeF_& Divide( float invsize ) { if( invsize != 0.0f ) { cx /= invsize; cy /= invsize; } return *this; }

        double GetAspectRatio() const { return cy != 0.0f ? (double) cx / (double) cy : 0.0; }
        _SizeF_& SetAspectRatio( SIZEF sizOfAspectRatio, DWORD option );
        _SizeF_& SetAspectRatio( double aspect_ratio, DWORD option );

        operator PSIZEF() { return this; }

        bool IsEmpty() const { return( cx <= 0.0f || cy <= 0.0f ); }
        bool IsNull () const { return( cx == 0.0f && cy == 0.0f ); }

        _PtF_&       GetAsPt()       { return * (_PtF_*) this; }
        const _PtF_& GetAsPt() const { return * (_PtF_*) this; }
        _RectF_ GetOriginRect() const; // top-left will be (0,0) and rect's size will be set to 'this' size
        _PtF_   GetCenter() const { return _PtF_( cx * 0.5f, cy * 0.5f ); }

        _SizeF_& MaxSize( PSIZEF psize2 ) { cx = privateRectF::tmax( cx, psize2->cx ); cy = privateRectF::tmax( cy, psize2->cy ); return *this; } // get the maximum of the sizes; only allow to make size greater. cx and cy may not be smaller than wanted
        _SizeF_& MinSize( PSIZEF psize2 ) { cx = privateRectF::tmin( cx, psize2->cx ); cy = privateRectF::tmin( cy, psize2->cy ); return *this; } // get the minimum of the sizes; only allow to make size smaller. cx and cy may not be greater than wanted

        _SizeF_& Absolute() { cx = fabsf(cx); cy = fabsf(cy); return *this; }
        _SizeF_& Negate  () { cx = -cx; cy = -cy; return *this; }
        _SizeF_& Flip    () { privateRectF::tswap( cx, cy ); return *this; }

        bool operator ==( SIZEF size ) const { return( cx == size.cx && cy == size.cy ); }
        bool operator !=( SIZEF size ) const { return( cx != size.cx || cy != size.cy ); }
        bool operator < ( SIZEF size ) const { return( cx <  size.cx && cy <  size.cy ); } // only if both extents are smaller -> then the complete size is smaller
        bool operator > ( SIZEF size ) const { return( cx >  size.cx && cy >  size.cy ); } // only if both extents are greater -> then the complete size is greater
        bool operator <=( SIZEF size ) const { return( cx <= size.cx || cy <= size.cy ); } // if one extent is smaller or equal -> the complete size is smaller or equal
        bool operator >=( SIZEF size ) const { return( cx >= size.cx || cy >= size.cy ); } // if one extent is greater or equal -> the complete size is greater or equal
        _SizeF_& operator +=( SIZEF      size ) { cx += size.cx; cy += size.cy; return *this; }
        _SizeF_& operator -=( SIZEF      size ) { cx -= size.cx; cy -= size.cy; return *this; }
        _SizeF_& operator *=( float     scale ) { cx *= scale;   cy *= scale;   return *this; }
        _SizeF_& operator *=( double    scale ) { cx = (float) (cx * scale); cy = (float) (cy * scale); return *this; }
        _SizeF_& operator /=( float  invscale ) { if( invscale != 0.0f ) { cx /= invscale; cy /= invscale; } return *this; }
        _SizeF_& operator /=( double invscale ) { if( invscale != 0.0  ) { cx = (float) (cx / invscale); cy = (float) (cy / invscale); } return *this; }

        _SizeF_ operator +( SIZEF      size ) const { return _SizeF_( cx + size.cx, cy + size.cy ); }
        _SizeF_ operator -( SIZEF      size ) const { return _SizeF_( cx - size.cx, cy - size.cy ); }
        _SizeF_ operator -()                  const { return _SizeF_( -cx, -cy ); }
        _SizeF_ operator *( float     scale ) const { return _SizeF_( cx * scale, cy * scale ); }
        _SizeF_ operator *( double    scale ) const { return _SizeF_( (float) (cx * scale), (float) (cy * scale) ); }
        _SizeF_ operator /( float  invscale ) const { if( invscale != 0.0f ) return _SizeF_( cx / invscale, cy / invscale ); else return *this; }
        _SizeF_ operator /( double invscale ) const { if( invscale != 0.0  ) return _SizeF_( (float) (cx / invscale), (float) (cy / invscale) ); else return *this; }

        _PtF_ operator +( POINTF pt ) const { return _PtF_( cx + pt.x, cy + pt.y ); }
        _PtF_ operator -( POINTF pt ) const { return _PtF_( cx - pt.x, cy - pt.y ); }

        _RectF_ operator +( PCRECTF prc ) const;

        _Size_ toSize      () const { return _Size_( (int) cx, (int) cy ); }
        _Size_ toSize_floor() const { return _Size_( (int) floorf(cx), (int) floorf(cy) ); }
        _Size_ toSize_ceil () const { return _Size_( (int) ceilf (cx), (int) ceilf (cy) ); }
        _Size_ toSize_round() const { return _Size_( (int) privateRectF::roundf(cx), (int) privateRectF::roundf(cy) ); }
};

class _RangeF_
{
    public :
        float m_start;
        float m_length;

    public :
        _RangeF_() {  } // random filled
        _RangeF_( float _start, float _length ) { m_start = _start; m_length = _length; }
        _RangeF_( const _Range_& _range ) { m_start = (float) _range.m_start; m_length = (float) _range.m_length; }
        _RangeF_( _Range_        _range ) { m_start = (float) _range.m_start; m_length = (float) _range.m_length; }

        _RangeF_& SetStartLength( float _start, float _length ) { m_start = _start; m_length = _length; return *this; }
        _RangeF_& SetStartEnd   ( float _start, float _end )    { m_start = _start; m_length = _end - _start; return *this; }
        _RangeF_& SetStart      ( float _start )                { m_length -= _start - m_start; m_start = _start; return *this; }
        _RangeF_& SetEnd        ( float _end )                  { m_length = _end - m_start; return *this; }
        _RangeF_& SetEmpty      ()                              { m_start = m_length = 0.0f; return *this; }

        bool IsEmpty() const { return m_length <= 0.0f; }
        bool IsNull () const { return m_start == 0.0f && m_length == 0.0f; }
        bool PosInRange( float pos ) const { return pos >= m_start && pos < m_start + m_length; }

        _RangeF_& Offset ( float d )          { m_start += d; return *this; }
        _RangeF_& Inflate( float d )          { m_start -= d; m_length += d + d; return *this; }
        _RangeF_& Inflate( float s, float e ) { m_start -= s; m_length += s + e; return *this; }
        _RangeF_& Deflate( float d )          { m_start += d; m_length -= d + d; return *this; }
        _RangeF_& Deflate( float s, float e ) { m_start += s; m_length -= s + e; return *this; }

        _RangeF_& MoveStartTo ( float pos ) { m_start = pos; return *this; }
        _RangeF_& MoveCenterTo( float pos ) { m_start = pos - (m_length * 0.5f); return *this; }
        _RangeF_& MoveEndTo   ( float pos ) { m_start = pos - m_length; return *this; }

        bool      IntersectRange       ( const _RangeF_& range2 );
        _RangeF_& IntersectRangeRet    ( const _RangeF_& range2 ) { IntersectRange( range2 ); return *this; }
        bool      IntersectRangeEmul   ( const _RangeF_& range2 ) { return IntersectRange( range2 ); }
        _RangeF_& IntersectRangeEmulRet( const _RangeF_& range2 ) { IntersectRangeEmul( range2 ); return *this; }
        bool      UnionRange           ( const _RangeF_& range2 );
        _RangeF_& UnionRangeRet        ( const _RangeF_& range2 ) { UnionRange( range2 ); return *this; }
        bool      UnionRangeEmul       ( const _RangeF_& range2 );
        _RangeF_& UnionRangeEmulRet    ( const _RangeF_& range2 ) { UnionRangeEmul( range2 ); return *this; }

        _RangeF_& operator = ( const _RangeF_& _range ) { m_start = _range.m_start; m_length = _range.m_length; return *this; }
        bool operator ==( const _RangeF_& _range ) const { return m_start == _range.m_start && m_length == _range.m_length; }
        bool operator !=( const _RangeF_& _range ) const { return m_start != _range.m_start || m_length != _range.m_length; }
        _RangeF_& operator +=( float ofs ) { m_start += ofs; return *this; }
        _RangeF_& operator -=( float ofs ) { m_start -= ofs; return *this; }
        _RangeF_& operator &=( const _RangeF_& _range ) { return IntersectRangeEmulRet( _range ); }
        _RangeF_& operator |=( const _RangeF_& _range ) { return UnionRangeEmulRet    ( _range ); }

        float GetStart () const { return m_start; }
        float GetEnd   () const { return m_start + m_length; }
        float GetLength() const { return m_length; }
        float GetCenter() const { return m_start + (m_length * 0.5f); }

        _Range_ toRange      () const { return _Range_().SetStartEnd( (int) GetStart(), (int) GetEnd() ); }
        _Range_ toRange_floor() const { return _Range_().SetStartEnd( (int) floorf(GetStart()), (int) floorf(GetEnd()) ); }
        _Range_ toRange_ceil () const { return _Range_().SetStartEnd( (int) ceilf (GetStart()), (int) ceilf (GetEnd()) ); }
        _Range_ toRange_round() const { return _Range_().SetStartEnd( (int) privateRectF::roundf(GetStart()), (int) privateRectF::roundf(GetEnd()) ); }
};

class _RectF_ : public tagRECTF
{
    public :
        _RectF_() {  } // random filled
        _RectF_( float l, float t, float r, float b ) { left = l; top = t; right = r; bottom = b; }
        _RectF_( const RECTF&  rc ) { * (RECTF*) this = rc; }
        _RectF_( const RECT&   rc ) { left = (float) rc.left; top = (float) rc.top; right = (float) rc.right; bottom = (float) rc.bottom; }
        _RectF_( PCRECTF      prc ) { if( prc ) * (RECTF*) this = *prc; }
        _RectF_( PCRECT       prc ) { if( prc ) { left = (float) prc->left; top = (float) prc->top; right = (float) prc->right; bottom = (float) prc->bottom; } }
        _RectF_( SIZEF       size ) { left = top = 0.0f; right = size.cx; bottom = size.cy; }
        _RectF_( POINTF topleft, SIZEF         size ) { right = ( left = topleft.x ) + size.cx; bottom = ( top = topleft.y ) + size.cy; }
        _RectF_( POINTF topleft, POINTF bottomright ) { left = topleft.x; top = topleft.y; right = bottomright.x; bottom = bottomright.y; }
        _RectF_( const WINDOWPOS *pwp ) { left = pwp ? (float) pwp->x : 0.0f; top = pwp ? (float) pwp->cy : 0.0f; right = left + (pwp ? (float) pwp->cx : 0.0f); bottom = top + (pwp ? (float) pwp->cy : 0.0f); }

        float Cx() const { return( right - left ); }
        float Cy() const { return( bottom - top ); }
        float Width () const { return( right - left ); }
        float Height() const { return( bottom - top ); }
        _RectF_& SetCx( float cx ) { right = left + cx; return *this; }
        _RectF_& SetCy( float cy ) { bottom = top + cy; return *this; }
        _RectF_& SetWidth ( float cx ) { right = left + cx; return *this; }
        _RectF_& SetHeight( float cy ) { bottom = top + cy; return *this; }
        _RectF_& SetSize( float cx, float cy ) { right = left + cx; bottom = top + cy; return *this; } // same meaning as 'SetSizeTL()'
        _RectF_& SetSize( SIZEF         size ) { return SetSizeTL( size ); }                           // same meaning as 'SetSizeTL()'
        _RectF_& SetSizeTL( float cx, float cy ) { right = left   + cx; bottom = top + cy; return *this; } // the Top-Left corner stays in place
        _RectF_& SetSizeTR( float cx, float cy ) { left  = right  - cx; bottom = top + cy; return *this; } // the Top-Right corner stays in place
        _RectF_& SetSizeBL( float cx, float cy ) { right = left   + cx; top = bottom - cy; return *this; } // the Bottom-Left corner stays in place
        _RectF_& SetSizeBR( float cx, float cy ) { left  = right  - cx; top = bottom - cy; return *this; } // the Bottom-Right corner stays in place
        _RectF_& SetSizeTC( float cx, float cy ) { right = ( left = Xm() - (cx * 0.5f) ) + cx; bottom = top + cy; return *this; } // the Top-horzCenter corner stays in place
        _RectF_& SetSizeBC( float cx, float cy ) { right = ( left = Xm() - (cx * 0.5f) ) + cx; top = bottom - cy; return *this; } // the Bottom-horzCenter corner stays in place
        _RectF_& SetSizeCL( float cx, float cy ) { right = left   + cx; bottom = ( top = Ym() - (cy * 0.5f) ) + cy; return *this; } // the vertCenter-Left corner stays in place
        _RectF_& SetSizeCR( float cx, float cy ) { left  = right  - cx; bottom = ( top = Ym() - (cy * 0.5f) ) + cy; return *this; } // the vertCenter-Right corner stays in place
        _RectF_& SetSizeCC( float cx, float cy ) { right = ( left = Xm() - (cx * 0.5f) ) + cx; bottom = ( top = Ym() - (cy * 0.5f) ) + cy; return *this; } // the Center of the rect (vertCenter-horzCenter) stays in place
        _RectF_& SetSizeTL( SIZEF size )         { return SetSizeTL( size.cx, size.cy ); } // the Top-Left corner stays in place
        _RectF_& SetSizeTR( SIZEF size )         { return SetSizeTR( size.cx, size.cy ); } // the Top-Right corner stays in place
        _RectF_& SetSizeBL( SIZEF size )         { return SetSizeBL( size.cx, size.cy ); } // the Bottom-Left corner stays in place
        _RectF_& SetSizeBR( SIZEF size )         { return SetSizeBR( size.cx, size.cy ); } // the Bottom-Right corner stays in place
        _RectF_& SetSizeTC( SIZEF size )         { return SetSizeTC( size.cx, size.cy ); } // the Top-horzCenter corner stays in place
        _RectF_& SetSizeBC( SIZEF size )         { return SetSizeBC( size.cx, size.cy ); } // the Bottom-horzCenter corner stays in place
        _RectF_& SetSizeCL( SIZEF size )         { return SetSizeCL( size.cx, size.cy ); } // the vertCenter-Left corner stays in place
        _RectF_& SetSizeCR( SIZEF size )         { return SetSizeCR( size.cx, size.cy ); } // the vertCenter-Right corner stays in place
        _RectF_& SetSizeCC( SIZEF size )         { return SetSizeCC( size.cx, size.cy ); } // the Center of the rect (vertCenter-horzCenter) stays in place
        _RectF_& LimitCx     ( float cxmin, float cxmax ) { LimitCxMax( cxmax ); LimitCxMin( cxmin ); return *this; }
        _RectF_& LimitCxMin  ( float cxmin )              { if( right - left < cxmin ) right = left + cxmin; return *this; } // width may not be smaller than 'cxmin'
        _RectF_& LimitCxMax  ( float cxmax )              { if( right - left > cxmax ) right = left + cxmax; return *this; } // width may not be greater than 'cxmax'
        _RectF_& LimitCy     ( float cymin, float cymax ) { LimitCyMax( cymax ); LimitCyMin( cymin ); return *this; }
        _RectF_& LimitCyMin  ( float cymin )              { if( bottom - top < cymin ) bottom = top + cymin; return *this; } // height may not be smaller than 'cymin'
        _RectF_& LimitCyMax  ( float cymax )              { if( bottom - top > cymax ) bottom = top + cymax; return *this; } // height may not be greater than 'cymax'
        _RectF_& LimitSizeMin( float cxmin, float cymin ) { LimitCxMin( cxmin      ); LimitCyMin( cymin      ); return *this; } // width and height may not be smaller than wanted
        _RectF_& LimitSizeMin( SIZEF sizemin )            { LimitCxMin( sizemin.cx ); LimitCyMin( sizemin.cy ); return *this; } // width and height may not be smaller than wanted
        _RectF_& LimitSizeMax( float cxmax, float cymax ) { LimitCxMax( cxmax      ); LimitCyMax( cymax      ); return *this; } // width and height may not be greater than wanted
        _RectF_& LimitSizeMax( SIZEF sizemax )            { LimitCxMax( sizemax.cx ); LimitCyMax( sizemax.cy ); return *this; } // width and height may not be greater than wanted
        _RectF_& MaxSize     ( SIZEF size2 )              { return LimitSizeMin( size2 ); } // get the maximum of the sizes; only allow to make size greater. width and height may not be smaller than wanted
        _RectF_& MinSize     ( SIZEF size2 )              { return LimitSizeMax( size2 ); } // get the minimum of the sizes; only allow to make size smaller. width and height may not be greater than wanted
        _SizeF_ Size() const { return _SizeF_( right - left, bottom - top ); }
        _PtF_& TopLeft          ()       { return *(_PtF_*) ((POINTF*)this);    }
        _PtF_& BottomRight      ()       { return *(_PtF_*)(((POINTF*)this)+1); }
        const _PtF_& TopLeft    () const { return *(_PtF_*) ((POINTF*)this);    }
        const _PtF_& BottomRight() const { return *(_PtF_*)(((POINTF*)this)+1); }
        _PtF_ GetTopLeft     () const { return _PtF_( left,                   top                  ); }
        _PtF_ GetTopCenter   () const { return _PtF_( (left+right) * 0.5f,    top                  ); }
        _PtF_ GetTopRight    () const { return _PtF_( right,                  top                  ); }
        _PtF_ GetVCenterRight() const { return _PtF_( right,                 (top + bottom) * 0.5f ); }
        _PtF_ GetBottomRight () const { return _PtF_( right,                  bottom               ); }
        _PtF_ GetBottomCenter() const { return _PtF_( (left + right) * 0.5f,  bottom               ); }
        _PtF_ GetBottomLeft  () const { return _PtF_( left,                   bottom               ); }
        _PtF_ GetVCenterLeft () const { return _PtF_( left,                  (top + bottom) * 0.5f ); }
        _PtF_ GetCenter      () const { return _PtF_( (left + right) * 0.5f, (top + bottom) * 0.5f ); }
        float Xm() const { return( (left + right) * 0.5f ); }
        float Ym() const { return( (top + bottom) * 0.5f ); }
        _RectF_&      SwapLeftRight()             { SwapLeftRight( (PRECTF)this ); return *this; }
        _RectF_&      SwapTopBottom()             { SwapTopBottom( (PRECTF)this ); return *this; }
        static PRECTF SwapLeftRight( PRECTF prc ) { privateRectF::tswap( prc->left, prc->right ); return prc; }
        static PRECTF SwapTopBottom( PRECTF prc ) { privateRectF::tswap( prc->top, prc->bottom ); return prc; }
        _RangeF_ GetHorzRange() const { return _RangeF_().SetStartEnd( left, right ); }
        _RangeF_ GetVertRange() const { return _RangeF_().SetStartEnd( top, bottom ); }
        void SetHorzRange( const _RangeF_& range ) { right  = ( left = range.m_start ) + range.m_length; }
        void SetVertRange( const _RangeF_& range ) { bottom = ( top  = range.m_start ) + range.m_length; }
        double GetAspectRatio() const { float _cy = Cy(); return _cy != 0.0f ? (double) Cx() / (double) _cy : 0.0; }

        operator PRECTF ()       { return this; }
        operator PCRECTF() const { return this; }

        bool IsRectEmpty() const { return( right <= left || bottom <= top ); }
        bool IsRectNull () const { return( left == 0.0f && right == 0.0f && top == 0.0f && bottom == 0.0f ); }
        bool PtInRect( POINTF pt ) const { return( pt.x >= left && pt.x < right && pt.y >= top && pt.y < bottom ); }

        _RectF_& SetRect     ( float x1, float y1, float x2, float y2 ) { left = x1; top = y1; right = x2; bottom = y2; return *this; }
        _RectF_& SetRect     ( POINTF topleft,     SIZEF         size ) { right = ( left = topleft.x ) + size.cx; bottom = ( top = topleft.y ) + size.cy; return *this; }
        _RectF_& SetRect     ( POINTF topleft,     POINTF bottomright ) { left = topleft.x; top = topleft.y; right = bottomright.x; bottom = bottomright.y; return *this; }
        _RectF_& SetRectEmpty()                                   { memset( this, 0, sizeof(RECTF) ); return *this; }
        _RectF_& CopyRect ( PCRECTF prc )       { if( prc ) *(RECTF*)this = *prc; return *this; }
        bool     EqualRect( PCRECTF prc ) const { return( memcmp( this, prc, sizeof(RECTF) ) == 0 ); }
        _RectF_& CopyHorizontal( PCRECTF prc ) { if( prc ) { left = prc->left; right  = prc->right;  } return *this; } // copies the left and right values
        _RectF_& CopyVertical  ( PCRECTF prc ) { if( prc ) { top  = prc->top;  bottom = prc->bottom; } return *this; } // copies the top and bottom values

        _RectF_& InflateRect( float x, float y ) { left -= x; top -= y; right += x; bottom += y; return *this; }
        _RectF_& InflateRect( SIZEF       size ) { left -= size.cx; top -= size.cy; right += size.cx; bottom += size.cy; return *this; }
        _RectF_& InflateRect( PCRECTF      prc ) { left -= prc->left; top -= prc->top; right += prc->right; bottom += prc->bottom; return *this; }
        _RectF_& InflateRect( float l, float t, float r, float b ) { left -= l; top -= t; right += r; bottom += b; return *this; }
        _RectF_& DeflateRect( float x, float y ) { left += x; top += y; right -= x; bottom -= y; return *this; }
        _RectF_& DeflateRect( SIZEF       size ) { left += size.cx; top += size.cy; right -= size.cx; bottom -= size.cy; return *this; }
        _RectF_& DeflateRect( PCRECTF      prc ) { left += prc->left; top += prc->top; right -= prc->right; bottom -= prc->bottom; return *this; }
        _RectF_& DeflateRect( float l, float t, float r, float b ) { left += l; top += t; right -= r; bottom -= b; return *this; }

        _RectF_& NormalizeRect();

        _RectF_& OffsetRect( float x, float y ) { left += x; top += y; right += x; bottom += y; return *this; }
        _RectF_& OffsetRect( POINTF        pt ) { left += pt.x; top += pt.y; right += pt.x; bottom += pt.y; return *this; }
        _RectF_& OffsetRect( SIZEF       size ) { left += size.cx; top += size.cy; right += size.cx; bottom += size.cy; return *this; }
        _RectF_& MoveRectTo( float x, float y ) { float cx = Cx(); float cy = Cy(); right = ( left = x ) + cx; bottom = ( top = y ) + cy; return *this; } // Moves the rect to the position without resizing (same as 'MoveRectTLTo()')
        _RectF_& MoveRectTo( POINTF        pt ) { return MoveRectTo( pt.x, pt.y ); } // Moves the rect to the position without resizing (same as 'MoveRectTLTo()')
        _RectF_& MoveRectToOrigin() { return MoveRectTo( 0.0f, 0.0f ); }
        _RectF_& MoveRectToFrame( _RectF_ rcFrame, DWORD flags );
        _RectF_& RectStayInFrame( RECTF   rcFrame, DWORD flags = RSIFF_DEFAULT );
        _RectF_& MoveRectTLTo( float x, float y ) { float cx = Cx(); float cy = Cy(); right = ( left  = x ) + cx; bottom = ( top    = y ) + cy; return *this; } // Moves the rect's Top-Left corner to the position without resizing (same as 'MoveRectTo()')
        _RectF_& MoveRectTRTo( float x, float y ) { float cx = Cx(); float cy = Cy(); left  = ( right = x ) - cx; bottom = ( top    = y ) + cy; return *this; } // Moves the rect's Top-Right corner to the position without resizing
        _RectF_& MoveRectBLTo( float x, float y ) { float cx = Cx(); float cy = Cy(); right = ( left  = x ) + cx; top    = ( bottom = y ) - cy; return *this; } // Moves the rect's Bottom-Left corner to the position without resizing
        _RectF_& MoveRectBRTo( float x, float y ) { float cx = Cx(); float cy = Cy(); left  = ( right = x ) - cx; top    = ( bottom = y ) - cy; return *this; } // Moves the rect's Bottom-Right corner to the position without resizing
        _RectF_& MoveRectTCTo( float x, float y ) { float cx = Cx(); float cy = Cy(); right = ( left  = x - (cx * 0.5f) ) + cx; bottom = ( top    = y ) + cy; return *this; } // Moves the rect's Top-horzCenter corner to the position without resizing
        _RectF_& MoveRectBCTo( float x, float y ) { float cx = Cx(); float cy = Cy(); right = ( left  = x - (cx * 0.5f) ) + cx; top    = ( bottom = y ) - cy; return *this; } // Moves the rect's Bottom-horzCenter corner to the position without resizing
        _RectF_& MoveRectCLTo( float x, float y ) { float cx = Cx(); float cy = Cy(); right = ( left  = x ) + cx; bottom = ( top = y - (cy * 0.5f) ) + cy; return *this; } // Moves the rect's vertCenter-Left corner to the position without resizing
        _RectF_& MoveRectCRTo( float x, float y ) { float cx = Cx(); float cy = Cy(); left  = ( right = x ) - cx; bottom = ( top = y - (cy * 0.5f) ) + cy; return *this; } // Moves the rect's vertCenter-Right corner to the position without resizing
        _RectF_& MoveRectCCTo( float x, float y ) { float cx = Cx(); float cy = Cy(); right = ( left  = x - (cx * 0.5f) ) + cx; bottom = ( top = y - (cy * 0.5f) ) + cy; return *this; } // Moves the rect's Center (vertCenter-horzCenter) to the position without resizing
        _RectF_& MoveRectTLTo( POINTF pt )        { return MoveRectTLTo( pt.x, pt.y ); } // Moves the rect's Top-Left corner to the position without resizing (same as 'MoveRectTo()')
        _RectF_& MoveRectTRTo( POINTF pt )        { return MoveRectTRTo( pt.x, pt.y ); } // Moves the rect's Top-Right corner to the position without resizing
        _RectF_& MoveRectBLTo( POINTF pt )        { return MoveRectBLTo( pt.x, pt.y ); } // Moves the rect's Bottom-Left corner to the position without resizing
        _RectF_& MoveRectBRTo( POINTF pt )        { return MoveRectBRTo( pt.x, pt.y ); } // Moves the rect's Bottom-Right corner to the position without resizing
        _RectF_& MoveRectTCTo( POINTF pt )        { return MoveRectTCTo( pt.x, pt.y ); } // Moves the rect's Top-horzCenter corner to the position without resizing
        _RectF_& MoveRectBCTo( POINTF pt )        { return MoveRectBCTo( pt.x, pt.y ); } // Moves the rect's Bottom-horzCenter corner to the position without resizing
        _RectF_& MoveRectCLTo( POINTF pt )        { return MoveRectCLTo( pt.x, pt.y ); } // Moves the rect's vertCenter-Left corner to the position without resizing
        _RectF_& MoveRectCRTo( POINTF pt )        { return MoveRectCRTo( pt.x, pt.y ); } // Moves the rect's vertCenter-Right corner to the position without resizing
        _RectF_& MoveRectCCTo( POINTF pt )        { return MoveRectCCTo( pt.x, pt.y ); } // Moves the rect's Center (vertCenter-horzCenter) to the position without resizing
        _RectF_& ZoomRect( float xscale, float yscale ) { left *= xscale; top *= yscale; right *= xscale; bottom *= yscale; return *this; } // Multiplies each coord of the rect
        _RectF_& ZoomRect( SIZEF  scale )               { return ZoomRect( scale.cx, scale.cy ); } // Multiplies each coord of the rect
        _RectF_& DivRect( float xinvscale, float yinvscale ) { if( xinvscale != 0.0f && yinvscale != 0.0f ) { left /= xinvscale; top /= yinvscale; right /= xinvscale; bottom /= yinvscale; } return *this; } // Divides each coord of the rect
        _RectF_& DivRect( float  invscale )                  { return DivRect( invscale, invscale ); } // Divides each coord of the rect
        _RectF_& ZoomSize( float xscale, float yscale ) { right = left + (right - left) * xscale; bottom = top + (bottom - top) * yscale; return *this; } // Multiplies only the size (width and height) of the rect
        _RectF_& ZoomSize( SIZEF  scale )               { return ZoomSize( scale.cx, scale.cy ); } // Multiplies only the size (width and height) of the rect
        _RectF_& DivSize( float xinvscale, float yinvscale ) { if( xinvscale != 0.0f && yinvscale != 0.0f ) { right = left + (right - left) / xinvscale; bottom = top + (bottom - top) / yinvscale; } return *this; } // Divides only the size (width and height) of the rect
        _RectF_& DivSize( SIZEF  divsize )                   { return DivSize( divsize.cx, divsize.cy ); } // Divides only the size (width and height) of the rect
        _RectF_& SetAspectRatio( SIZEF sizOfAspectRatio, DWORD option );
        _RectF_& SetAspectRatio( double aspect_ratio, DWORD option );
        _RectF_ MirrorRect            ( POINTF pt ) const;
        _RectF_ MirrorOnHorizontalAxis( float   y ) const;
        _RectF_ MirrorOnVerticalAxis  ( float   x ) const;

        _RectF_& QuarterizeRect( PRECTF prcTL, PRECTF prcTR, PRECTF prcBL, PRECTF prcBR );

        static bool AreAttachedRects( PCRECTF prc1, PCRECTF prc2 );
        bool        AreAttachedRects( PCRECTF prc2 ) const { return AreAttachedRects( this, prc2 ); }
        static bool AreHorizontallyAttachedRects( PCRECTF prc1, PCRECTF prc2 );
        bool        AreHorizontallyAttachedRects( PCRECTF prc2 ) const { return AreHorizontallyAttachedRects( this, prc2 ); }
        static bool AreVerticallyAttachedRects( PCRECTF prc1, PCRECTF prc2 );
        bool        AreVerticallyAttachedRects( PCRECTF prc2 ) const { return AreVerticallyAttachedRects( this, prc2 ); }

        bool     IntersectRectEmul   ( PCRECTF prc )                { IntersectRectEmulRet( this, prc );  return !IsRectEmpty(); }
        bool     IntersectRectEmul   ( PCRECTF prc1, PCRECTF prc2 ) { IntersectRectEmulRet( prc1, prc2 ); return !IsRectEmpty(); }
        _RectF_& IntersectRectEmulRet( PCRECTF prc )                { IntersectRectEmulRet( this, prc );  return *this; }
        _RectF_& IntersectRectEmulRet( PCRECTF prc1, PCRECTF prc2 );
        bool     UnionRectEmul       ( PCRECTF prc )                { UnionRectEmulRet( this, prc );  return !IsRectEmpty(); }
        bool     UnionRectEmul       ( PCRECTF prc1, PCRECTF prc2 ) { UnionRectEmulRet( prc1, prc2 ); return !IsRectEmpty(); }
        _RectF_& UnionRectEmulRet    ( PCRECTF prc )                { UnionRectEmulRet( this, prc );  return *this; }
        _RectF_& UnionRectEmulRet    ( PCRECTF prc1, PCRECTF prc2 );

        _RectF_& operator = ( const RECTF& rc ) { *(RECTF*)this = rc; return *this; }
        bool     operator ==( const RECTF& rc ) const { return( memcmp( this, &rc, sizeof(RECTF) ) == 0 ); }
        bool     operator !=( const RECTF& rc ) const { return( memcmp( this, &rc, sizeof(RECTF) ) != 0 ); }
        _RectF_& operator +=( POINTF       pt ) { left  += pt.x;    top    += pt.y;    right += pt.x; bottom += pt.y; return *this; }
        _RectF_& operator +=( SIZEF      size ) { right += size.cx; bottom += size.cy; return *this; }
        _RectF_& operator +=( PCRECTF     prc ) { InflateRect( prc ); return *this; }
        _RectF_& operator -=( POINTF       pt ) { left  -= pt.x;    top    -= pt.y;    right -= pt.x; bottom -= pt.y; return *this; }
        _RectF_& operator -=( SIZEF      size ) { right -= size.cx; bottom -= size.cy; return *this; }
        _RectF_& operator -=( PCRECTF     prc ) { DeflateRect( prc ); return *this; }
        _RectF_& operator &=( const RECTF& rc ) { return IntersectRectEmulRet( &rc ); }
        _RectF_& operator |=( const RECTF& rc ) { return UnionRectEmulRet    ( &rc ); }
        _RectF_& operator *=( float     scale ) { return ZoomRect( scale, scale ); }
        _RectF_& operator *=( double    scale ) { return ZoomRect( (float) scale, (float) scale ); }
        _RectF_& operator /=( float  invscale ) { return DivRect( invscale, invscale ); }
        _RectF_& operator /=( double invscale ) { return DivRect( (float) invscale, (float) invscale ); }

        _RectF_ operator +( POINTF  pt ) const { return _RectF_( left + pt.x, top + pt.y, right + pt.x,    bottom + pt.y    ); }
        _RectF_ operator -( POINTF  pt ) const { return _RectF_( left - pt.x, top - pt.y, right - pt.x,    bottom - pt.y    ); }
        _RectF_ operator +( SIZEF size ) const { return _RectF_( left,        top,        right + size.cx, bottom + size.cy ); }
        _RectF_ operator -( SIZEF size ) const { return _RectF_( left,        top,        right - size.cx, bottom - size.cy ); }
        // IMPORTANT: The 'operator+(SIZEF)' and 'operator-(SIZEF)' operators
        // in here have a different implmenetation than 'CRect'.
        // 'CRect' will offset the rect itself while '_RectF_' offsets only
        // the rect's bottom right corner.
        _RectF_ operator +( PCRECTF      prc ) const { _RectF_ rc( this ); rc.InflateRect( prc ); return rc; }
        _RectF_ operator -( PCRECTF      prc ) const { _RectF_ rc( this ); rc.DeflateRect( prc ); return rc; }
        _RectF_ operator &( const RECTF& rc2 ) const { _RectF_ rc; rc.IntersectRectEmul( this, &rc2 ); return rc; }
        _RectF_ operator |( const RECTF& rc2 ) const { _RectF_ rc; rc.UnionRectEmul    ( this, &rc2 ); return rc; }
        _RectF_ operator *( float     scale ) const { return _RectF_( this ).ZoomRect( scale, scale ); }
        _RectF_ operator *( double    scale ) const { return _RectF_( this ).ZoomRect( (float) scale, (float) scale ); }
        _RectF_ operator /( float  invscale ) const { return _RectF_( this ).DivRect ( invscale, invscale ); }
        _RectF_ operator /( double invscale ) const { return _RectF_( this ).DivRect ( (float) invscale, (float) invscale ); }
        _RectF_ MulDiv( double dMultiplier, double dDivisor ) const;

        _Rect_ toRect      () const { return _Rect_( (int) left, (int) top, (int) right, (int) bottom ); }
        _Rect_ toRect_floor() const { return _Rect_( (int) floorf(left), (int) floorf(top), (int) floorf(right), (int) floorf(bottom) ); }
        _Rect_ toRect_ceil () const { return _Rect_( (int) ceilf (left), (int) ceilf (top), (int) ceilf (right), (int) ceilf (bottom) ); }
        _Rect_ toRect_round() const { return _Rect_( (int) privateRectF::roundf(left), (int) privateRectF::roundf(top), (int) privateRectF::roundf(right), (int) privateRectF::roundf(bottom) ); }
};




}

#endif
