/*
    File:               argonRegistry.cpp
    Description:        Registry, PrivateProfiles.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         13.11.2003
    last updated:       26.10.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonRegistry_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonRegistry.h"

#ifndef _argon_no_warning_disables
    #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)
    #pragma warning( disable: 4996 )    // warning C4996: 'sscanf' was declared deprecated
    // Note: Usage of #sscanf()# and #swscanf()# are no security problem in here because
    // we do not use any "%s" sequences in the format strings.
    #endif
#endif

namespace ArgonLibrary
{




// Use this macro for #RegKey# methods and derived classes to
// automatically enable and restore the correct locale.
#ifndef _WIN32_WCE
#define REGLOC      StoreLocale stloc; if( m_bUse_C_Locale ) _set_C_locale_();
#else
#define REGLOC
#endif

reterr_t RegKey::Close()
{
    if( !m_hKey ) return true;  // return interr (errRegNotOpen);

    LONG err = ::RegCloseKey( m_hKey );

    m_hKey            = NULL;
    m_bCreated        = FALSE;
    m_dwLastValueSize = 0;

    return syserr (err);
}

reterr_t RegKey::_Create( HKEY hkeyParent, PCTSTR pstrSubkey, bool bCreate, REGSAM samDesired )
{
    if( m_hKey ) return interr (errRegAlreadyOpen);
    if( isOpen() ) return interr (errRegDerivedAlreadyOpen);

    String subkey = pstrSubkey;
    if( !hkeyParent || subkey.IsEmpty() ) return syserr (ERROR_INVALID_PARAMETER);

    subkey.OsPath();
    DWORD dwDisposition;
    LONG err;

    if( bCreate ) {
        err = RegCreateKeyEx( hkeyParent,
                              subkey,
                              0,
                              NULL,
                              0,
                              samDesired,
                              NULL,
                              &m_hKey,
                              &dwDisposition );
    } else {
        err = RegOpenKeyEx( hkeyParent,
                            subkey,
                            0,
                            samDesired,
                            &m_hKey );
        dwDisposition = REG_OPENED_EXISTING_KEY;
    }

    m_bCreated = false;

    if( err != ERROR_SUCCESS ) {
        m_hKey   = NULL;
        m_samKey = 0;
    } else {
        m_samKey = samDesired;
        switch( dwDisposition ) {
            case REG_CREATED_NEW_KEY     : m_bCreated = true;  break;
            case REG_OPENED_EXISTING_KEY : m_bCreated = false; break;
        }
    }

    m_dwLastValueSize = 0;

    return syserr (err);
}



// Gets next SubKey
//  : nSubKey = index
//  : pszSubKey = will contain the result
//  : zlenSubKey = size of 'pszSubKey' buffer in characters
//  > false = Error or no more SubKeys found
bool RegKey::EnumKey( _idx_ nSubKey, PTSTR pstrSubKey, _size_t_ zlenSubKey )
{
    if( !m_hKey ) return false;
    if( nSubKey < 0 || !pstrSubKey || zlenSubKey <= 0 ) return false;

    DWORD dw = (DWORD) zlenSubKey;
    FILETIME ft;

    LONG err = RegEnumKeyEx( m_hKey,
                             (DWORD) nSubKey,
                             pstrSubKey,
                             &dw,
                             NULL,
                             NULL,
                             NULL,
                             &ft );

    if( err == ERROR_NO_MORE_ITEMS )
        return false;

    return err == ERROR_SUCCESS;
}

bool RegKey::EnumKey( _idx_ nSubKey, String& strSubKey )
{
    TCHAR subkey[REGKEY_MAXNAMELEN];

    bool ret = EnumKey( nSubKey, subkey, _countof(subkey) );

    if( ret ) strSubKey = subkey;
    else      strSubKey.Empty();

    return ret;
}

// Gets next Value
//  : nValue = index
//  : pszValue = will contain the result
//  : zlenValue = size of 'pszValue' buffer in characters
//  : pdwType = pointer to DWORD which will contain the type of data
//            NULL (default) = no type will be retrieved.
//  > false = Error or no more Values found
bool RegKey::EnumValue( _idx_ nValue, PTSTR pstrValue, _size_t_ zlenValue, PDWORD pdwType )
{
    if( !m_hKey ) return false;
    if( !pstrValue || zlenValue <= 0 ) return false;

    DWORD dw = (DWORD) zlenValue;

    LONG err = RegEnumValue( m_hKey,
                             (DWORD) nValue,
                             pstrValue,
                             &dw,
                             NULL,
                             pdwType,
                             NULL,
                             NULL );

    if( err == ERROR_NO_MORE_ITEMS )
        return false;

    return err == ERROR_SUCCESS;
}

bool RegKey::EnumValue( _idx_ nValue, String& strValue, PDWORD pdwType )
{
    TCHAR value[REGKEY_MAXNAMELEN];

    bool ret = EnumValue( nValue, value, _countof(value), pdwType );

    if( ret ) strValue = value;
    else      strValue.Empty();

    return ret;
}



#ifndef _WIN32_WCE

// Gets the Type and Size of a Value
//  : cSizeBytes = initial size value +
//                 on return, the real size for a buffer in bytes will be returned.
// Note: When the value is of type string (REG_SZ or REG_EXPAND_SZ) -> then
// the size returned will also include the terminating 0-byte of the string.
reterr_t RegKey::GetValueTypeSize( PCSTR pstrValue, DWORD& dwType, _size_t_& cSizeBytes )
{
    if( !m_hKey ) return interr (errRegNotOpen);

    m_dwLastValueSize = (DWORD) cSizeBytes;

    LONG err = RegQueryValueExA( m_hKey,
                                 pstrValue,
                                 NULL,
                                 &dwType,
                                 NULL,
                                 &m_dwLastValueSize );

    if( err == ERROR_SUCCESS )
        cSizeBytes = (_size_t_) m_dwLastValueSize;

    return syserr (err);
}

// Gets the Type and Size of a Value
//  : cSizeBytes = initial size value +
//                 on return, the real size for a buffer in bytes will be returned.
// Note: When the value is of type string (REG_SZ or REG_EXPAND_SZ) -> then
// the size returned will also include the terminating 0-byte of the string.
reterr_t RegKey::GetValueTypeSize( PCWSTR pstrValue, DWORD& dwType, _size_t_& cSizeBytes )
{
    if( !m_hKey ) return interr (errRegNotOpen);

    m_dwLastValueSize = (DWORD) cSizeBytes;

    LONG err = RegQueryValueExW( m_hKey,
                                 pstrValue,
                                 NULL,
                                 &dwType,
                                 NULL,
                                 &m_dwLastValueSize );

    if( err == ERROR_SUCCESS )
        cSizeBytes = (_size_t_) m_dwLastValueSize;

    return syserr (err);
}

#else

// Gets the Type and Size of a Value
//  : cSizeBytes = initial size value +
//                 on return, the real size for a buffer in bytes will be returned.
// Note: When the value is of type string (REG_SZ or REG_EXPAND_SZ) -> then
// the size returned will also include the terminating 0-byte of the string.
reterr_t RegKey::GetValueTypeSize( PCTSTR pstrValue, DWORD& dwType, _size_t_& cSizeBytes )
{
    if( !m_hKey ) return interr (errRegNotOpen);

    m_dwLastValueSize = (DWORD) cSizeBytes;

    LONG err = RegQueryValueEx( m_hKey,
                                pstrValue,
                                NULL,
                                &dwType,
                                NULL,
                                &m_dwLastValueSize );

    if( err == ERROR_SUCCESS )
        cSizeBytes = (_size_t_) m_dwLastValueSize;

    return syserr (err);
}

#endif

// returns true, if the value exists
bool RegKey::isValue( PCTSTR pstrValue )
{
    DWORD dw = m_dwLastValueSize;
    DWORD dummy1;
    _size_t_ dummy2;

    bool ok = GetValueTypeSize( pstrValue, dummy1, dummy2 );

    m_dwLastValueSize = dw;

    return ok;
}

// returns true, if the value is of type REG_SZ or REG_EXPAND_SZ
bool RegKey::isString( PCTSTR pstrValue )
{
    DWORD dw = m_dwLastValueSize;
    DWORD type;
    _size_t_ dummy;

    bool ok = GetValueTypeSize( pstrValue, type, dummy );

    if( ok && (type != REG_SZ && type != REG_EXPAND_SZ) )
        ok = false;

    m_dwLastValueSize = dw;

    return ok;
}

// returns true, if the value is of type REG_DWORD
bool RegKey::isDWord( PCTSTR pstrValue )
{
    DWORD dw = m_dwLastValueSize;
    DWORD type;
    _size_t_ dummy;

    bool ok = GetValueTypeSize( pstrValue, type, dummy );

    if( ok && type != REG_DWORD )
        ok = false;

    m_dwLastValueSize = dw;

    return ok;
}

// returns true, if the value is of type REG_BINARY
bool RegKey::isBinary( PCTSTR pstrValue )
{
    DWORD dw = m_dwLastValueSize;
    DWORD type;
    _size_t_ dummy;

    bool ok = GetValueTypeSize( pstrValue, type, dummy );

    if( ok && type != REG_BINARY )
        ok = false;

    m_dwLastValueSize = dw;

    return ok;
}

bool RegKey::isKey( PCTSTR pstrSubkey )
{
    if( !this ) return false;
    return isKey( m_hKey, pstrSubkey );
}

bool RegKey::isKey( HKEY hkey, PCTSTR pstrKey )
{
    if( !hkey ) return false;
    if( _isstrempty_(pstrKey) ) return false;

    RegKey key;
    return key.Open( hkey, pstrKey, KEY_READ );
}



reterr_t RegKey::DeleteSubKey( PCTSTR pstrSubKey )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( _isstrempty_(pstrSubKey) ) return syserr (ERROR_INVALID_PARAMETER);

    // Before deleting the key, delete all subkeys the key has (this is necessary
    // because in WinNT and higher, a key must not contain any subkeys to delete it).
    RegKey key;
    reterr_t err = key.Open( m_hKey, pstrSubKey, KEY_ALL_ACCESS );
    if( !err ) return err;

    TCHAR subsubkey[REGKEY_MAXNAMELEN];
    while( key.EnumKey( 0, subsubkey, _countof(subsubkey) ) )
        if( !( err = key.DeleteSubKey(subsubkey) ) )
            return err;

    key.Close();

    return syserr (RegDeleteKey( m_hKey, pstrSubKey ));
}

reterr_t RegKey::DeleteRootKey( HKEY hRootKey, PCTSTR pstrSubKey )
{
    if( _isstrempty_(pstrSubKey) ) return syserr (ERROR_INVALID_PARAMETER);

    // Before deleting the key, delete all subkeys the key has (this is necessary
    // because in WinNT and higher, a key must not contain any subkeys to delete it).
    RegKey key;
    reterr_t err = key.Open( hRootKey, pstrSubKey, KEY_ALL_ACCESS );
    if( !err ) return err;

    TCHAR subsubkey[REGKEY_MAXNAMELEN];
    while( key.EnumKey( 0, subsubkey, _countof(subsubkey) ) )
        if( !( err = key.DeleteSubKey(subsubkey) ) )
            return err;

    key.Close();

    return syserr (RegDeleteKey( hRootKey, pstrSubKey ));
}

reterr_t RegKey::DeleteValue( PCTSTR pstrValue )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( _isstrempty_(pstrValue) ) return syserr (ERROR_INVALID_PARAMETER);

    return syserr (RegDeleteValue( m_hKey, pstrValue ));
}

//  > false = Error (not all values could be deleted)
reterr_t RegKey::DeleteAllValues()
{
    // Note: Do not check 'm_hKey' in here, because this method
    // will also be used by derived classes which do not use
    // the 'm_hKey' member.

    TCHAR value[REGKEY_MAXNAMELEN];
    reterr_t err;
    while( EnumValue( 0, value, _countof(value) ) )
        if( !( err = DeleteValue(value) ) )
            return err;

    return true;
}

// Deletes all these values with names matching one of the masks in 'pszMasks'.
//  : pszMasks = a list of masks (separated by ';' characters).
//               The syntax of these masks is identical to the syntax of '_matchpathmasks_()'.
//             Note: Because ( as described at '_matchimasks_()' in "StdExt-cpp.h" ) the
//              '.' and ';' chars have special meanings and if you want a mask to explicitely
//              contain one of these chars to be searched for, do specify "[.]" instead of
//              '.', and "[;]" instead of ';'.
//              Similar you can treat the special chars '*' and '?'.
//              Also instead of '[' use "[[]".
//  > false = Error (not all desired values could be deleted)
reterr_t RegKey::DeleteSomeValues( PCTSTR pstrMasks )
{
    // Note: Do not check 'm_hKey' in here, because this method
    // will also be used by derived classes which do not use
    // the 'm_hKey' member.

    if( !pstrMasks ) return syserr (ERROR_INVALID_PARAMETER);

    TCHAR value[REGKEY_MAXNAMELEN];
    reterr_t err;
    _idx_ i = 0;
    while( EnumValue( i, value, _countof(value) ) )
    {
        if( _matchpathmasks_( value, pstrMasks ) )
        {
            if( !( err = DeleteValue(value) ) )
                return err;
            // Now start right away from the beginning just
            // for savety if the order of values in a key
            // is shuffled by the system. (I don't know if
            // this can happen, so we do this to be sure).
            i = 0;
        } else
            ++i;
    }

    return true;
}



///<summary> Sets a value of type REG_NONE. </summary>
reterr_t RegKey::SetRegNone( PCTSTR pstrValue )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueEx( m_hKey,
                              pstrValue,
                              0,
                              REG_NONE,
                              NULL,
                              0 );

    return syserr (err);
}

// Sets a DWORD value
reterr_t RegKey::SetDWord( PCTSTR pstrValue, DWORD val )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueEx( m_hKey,
                              pstrValue,
                              0,
                              REG_DWORD,
                              (const BYTE*) &val,
                              sizeof(val) );

    return syserr (err);
}

// Gets a DWORD value
// If the method fails, the value will *not* be changed
reterr_t RegKey::GetDWord( PCTSTR pstrValue, DWORD& val )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue ) return syserr (ERROR_INVALID_PARAMETER);

    DWORD type; _size_t_ size = sizeof(val);
    reterr_t ret = GetValueTypeSize( pstrValue, type, size );
    if( !ret ) return ret;

    if( type != REG_DWORD )
        return syserr (ERROR_INVALID_DATA);

    m_dwLastValueSize = (DWORD) sizeof(val);

    DWORD dw;
    LONG err = RegQueryValueEx( m_hKey,
                                pstrValue,
                                NULL,
                                NULL,
                                (PBYTE) &dw,
                                &m_dwLastValueSize );

    if( err == ERROR_SUCCESS )
        val = dw;

    return syserr (err);
}

reterr_t RegKey::GetBoolFromYesNo( PCTSTR pstrValue, bool& val )
{
    String str;
    reterr_t err = GetString( pstrValue, str );
    if( !err ) return err;

    if(      str.Striis( _T("yes") ) ) val = true;
    else if( str.Striis( _T("no" ) ) ) val = false;
    else return syserr (ERROR_INVALID_PARAMETER);

    return true;
}

// Sets Binary Data
reterr_t RegKey::SetBinary( PCTSTR pstrValue, const void *pData, _size_t_ cData )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue || !pData || cData <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueEx( m_hKey,
                              pstrValue,
                              0,
                              REG_BINARY,
                              (const BYTE*) pData,
                              cData );

    return syserr (err);
}

// Gets Binary Data
// If the method fails, the value will *hopefully* not be changed ( depends on 'RegQueryValueEx()' )
// ( Tested: It seems so that 'RegQueryValueEx()' does not change the value if it cannot be found in the Registry. )
reterr_t RegKey::GetBinary( PCTSTR pstrValue, void *pData, _size_t_ cData )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue || !pData || cData <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    DWORD type; _size_t_ size = cData;
    reterr_t ret = GetValueTypeSize( pstrValue, type, size );
    if( !ret ) return ret;

    if( type != REG_BINARY )
        return syserr (ERROR_INVALID_DATA);

    m_dwLastValueSize = (DWORD) cData;

    LONG err = RegQueryValueEx( m_hKey,
                                pstrValue,
                                NULL,
                                NULL,
                                (PBYTE) pData,
                                &m_dwLastValueSize );

    return syserr (err);
}

///<summary> Similar to <b>GetBinary()</b>, but if the provided
///  buffer is too small (<b>cData</b> is too small), the <b>GetBinary()</b>
///  will fail while <b>GetBinaryPart()</b> will return only the
///  requested leading part of the binary data. </summary>
///<remarks> The emulation this way is done that the whole data
///  is read and only the desired part is returned. </remarks>
reterr_t RegKey::GetBinaryPart( PCTSTR pstrValue, void *pData, _size_t_ cData )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue || !pData || cData <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    DWORD type; _size_t_ size = cData;
    reterr_t ret = GetValueTypeSize( pstrValue, type, size );
    if( !ret ) return ret;

    if( type != REG_BINARY )
        return syserr (ERROR_INVALID_DATA);

    if( size > cData )
    {
        // Special handling.

        Array<BYTE> mem;
        PBYTE ptr = mem.AddNew( size );
        if( !ptr )
            return syserr (ERROR_OUTOFMEMORY);

        m_dwLastValueSize = (DWORD) size;

        LONG err = RegQueryValueEx( m_hKey,
                                    pstrValue,
                                    NULL,
                                    NULL,
                                    (PBYTE) ptr,
                                    &m_dwLastValueSize );

        if( err == ERROR_SUCCESS )
            memcpy( pData, ptr, cData );

        return syserr (err);
    }

    // Same handling as in #GetBinary()#.

    m_dwLastValueSize = (DWORD) cData;

    LONG err = RegQueryValueEx( m_hKey,
                                pstrValue,
                                NULL,
                                NULL,
                                (PBYTE) pData,
                                &m_dwLastValueSize );

    return syserr (err);
}

#ifndef _WIN32_WCE

// Sets a String
reterr_t RegKey::SetString( PCSTR pstrValue, PCSTR pStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueExA( m_hKey,
                               pstrValue,
                               0,
                               REG_SZ,
                               (const BYTE*) pStr,
                               (_strlen_(pStr) + 1) * sizeof(*pStr) );

    return syserr (err);
}

// Sets a String
reterr_t RegKey::SetString( PCWSTR pstrValue, PCWSTR pStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueExW( m_hKey,
                               pstrValue,
                               0,
                               REG_SZ,
                               (const BYTE*) pStr,
                               (_strlen_(pStr) + 1) * sizeof(*pStr) );

    return syserr (err);
}

// Sets an Expand-String
reterr_t RegKey::SetExpandString( PCSTR pstrValue, PCSTR pStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueExA( m_hKey,
                               pstrValue,
                               0,
                               REG_EXPAND_SZ,
                               (const BYTE*) pStr,
                               (_strlen_(pStr) + 1) * sizeof(*pStr) );

    return syserr (err);
}

// Sets an Expand-String
reterr_t RegKey::SetExpandString( PCWSTR pstrValue, PCWSTR pStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueExW( m_hKey,
                               pstrValue,
                               0,
                               REG_EXPAND_SZ,
                               (const BYTE*) pStr,
                               (_strlen_(pStr) + 1) * sizeof(*pStr) );

    return syserr (err);
}

// Gets a String
// If the method fails, the value will *hopefully* not be changed ( depends on 'RegQueryValueEx()' )
// ( Tested: It seems so that 'RegQueryValueEx()' does not change the value if it cannot be found in the Registry. )
reterr_t RegKey::GetString( PCSTR pstrValue, PSTR pStr, _size_t_ zlenStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr || zlenStr <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    DWORD type; _size_t_ size = zlenStr * sizeof(*pStr);
    reterr_t ret = GetValueTypeSize( pstrValue, type, size );
    if( !ret ) return ret;

    if( type != REG_SZ && type != REG_EXPAND_SZ )
        return syserr (ERROR_INVALID_DATA);

    m_dwLastValueSize = zlenStr * sizeof(*pStr);

    LONG err = RegQueryValueExA( m_hKey,
                                 pstrValue,
                                 NULL,
                                 NULL,
                                 (PBYTE) pStr,
                                 &m_dwLastValueSize );

    if( err == ERROR_SUCCESS )
        if( pStr && zlenStr > 0 ) pStr[zlenStr - 1] = 0;

    return syserr (err);
}

// Gets a String
// If the method fails, the value will *hopefully* not be changed ( depends on 'RegQueryValueEx()' )
// ( Tested: It seems so that 'RegQueryValueEx()' does not change the value if it cannot be found in the Registry. )
reterr_t RegKey::GetString( PCWSTR pstrValue, PWSTR pStr, _size_t_ zlenStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr || zlenStr <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    DWORD type; _size_t_ size = zlenStr * sizeof(*pStr);
    reterr_t ret = GetValueTypeSize( pstrValue, type, size );
    if( !ret ) return ret;

    if( type != REG_SZ && type != REG_EXPAND_SZ )
        return syserr (ERROR_INVALID_DATA);

    m_dwLastValueSize = zlenStr * sizeof(*pStr);

    LONG err = RegQueryValueExW( m_hKey,
                                 pstrValue,
                                 NULL,
                                 NULL,
                                 (PBYTE) pStr,
                                 &m_dwLastValueSize );

    if( err == ERROR_SUCCESS )
        if( pStr && zlenStr > 0 ) pStr[zlenStr - 1] = 0;

    return syserr (err);
}

#else

// Sets a String
reterr_t RegKey::SetString( PCTSTR pstrValue, PCTSTR pStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueEx( m_hKey,
                              pstrValue,
                              0,
                              REG_SZ,
                              (const BYTE*) pStr,
                              (_strlen_(pStr) + 1) * sizeof(*pStr) );

    return syserr (err);
}

// Sets an Expand-String
reterr_t RegKey::SetExpandString( PCTSTR pstrValue, PCTSTR pStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr ) return syserr (ERROR_INVALID_PARAMETER);

    LONG err = RegSetValueEx( m_hKey,
                              pstrValue,
                              0,
                              REG_EXPAND_SZ,
                              (const BYTE*) pStr,
                              (_strlen_(pStr) + 1) * sizeof(*pStr) );

    return syserr (err);
}

// Gets a String
// If the method fails, the value will *hopefully* not be changed ( depends on 'RegQueryValueEx()' )
// ( Tested: It seems so that 'RegQueryValueEx()' does not change the value if it cannot be found in the Registry. )
reterr_t RegKey::GetString( PCTSTR pstrValue, PTSTR pStr, _size_t_ zlenStr )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pStr || zlenStr <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    DWORD type; _size_t_ size = zlenStr * sizeof(*pStr);
    reterr_t ret = GetValueTypeSize( pstrValue, type, size );
    if( !ret ) return ret;

    if( type != REG_SZ && type != REG_EXPAND_SZ )
        return syserr (ERROR_INVALID_DATA);

    m_dwLastValueSize = zlenStr * sizeof(*pStr);

    LONG err = RegQueryValueEx( m_hKey,
                                pstrValue,
                                NULL,
                                NULL,
                                (PBYTE) pStr,
                                &m_dwLastValueSize );

    if( err == ERROR_SUCCESS )
        if( pStr && zlenStr > 0 ) pStr[zlenStr - 1] = 0;

    return syserr (err);
}

#endif

// Gets a variable length String
// If the method fails, the string will not be changed.
reterr_t RegKey::GetString( PCTSTR pstrValue, String& str )
{
    if( !m_hKey ) return interr (errRegNotOpen);

    DWORD    val_type;
    _size_t_ val_size;
    reterr_t ret = GetValueTypeSize( pstrValue, val_type, val_size );
    if( !ret ) return ret;

    if( val_type != REG_SZ && val_type != REG_EXPAND_SZ ) return syserr (ERROR_INVALID_DATA);
    if( val_size <= 0 ) return syserr (ERROR_INVALID_DATA);

    String tmp;
    PTSTR ptmp = tmp.GetBuffer( val_size / sizeof(TCHAR) - 1 );
    if( !ptmp )
        return syserr (ERROR_OUTOFMEMORY);

    ret = GetString( pstrValue, ptmp, val_size / sizeof(TCHAR) );

    tmp.ReleaseBuffer();

    if( ret )
        str = tmp;

    return ret;
}

reterr_t RegKey::SetStringList( PCTSTR pstrValue, PCTSTR pStrLst )
{
    if( !pStrLst ) return syserr (ERROR_INVALID_PARAMETER);
    return SetBinary( pstrValue, (const void*) pStrLst, (_strlstlen_(pStrLst) + 2) * sizeof(*pStrLst) );
}

reterr_t RegKey::GetStringList( PCTSTR pstrValue, PTSTR pStrLst, _size_t_ zzlenStrLst )
{
    if( !pStrLst || zzlenStrLst < 2 ) return syserr (ERROR_INVALID_PARAMETER);
    return GetBinary( pstrValue, (void*) pStrLst, zzlenStrLst * sizeof(*pStrLst) );
}

reterr_t RegKey::SetFloatAsString( PCTSTR pstrValue, float val )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    REGLOC;

    TCHAR s[128];
    _sprintf_( s, _T("%g"), val );

    return SetString( pstrValue, s );
}

reterr_t RegKey::GetFloatFromString( PCTSTR pstrValue, float& val )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    TCHAR s[128];
    reterr_t err = GetString( pstrValue, s, _countof(s) );
    if( !err ) return err;

    REGLOC;

    val = (float) _stof_( s );

    return true;
}

reterr_t RegKey::SetDoubleAsString( PCTSTR pstrValue, double val )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    REGLOC;

    TCHAR s[128];
    _sprintf_( s, _T("%g"), val );

    return SetString( pstrValue, s );
}

reterr_t RegKey::GetDoubleFromString( PCTSTR pstrValue, double& val )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    TCHAR s[128];
    reterr_t err = GetString( pstrValue, s, _countof(s) );
    if( !err ) return err;

    REGLOC;

    val = (double) _stof_( s );

    return true;
}

// Sets the string of the string-list at the speicified index.
// The string-list must be consecitive strings separated by '\0' chars. The end of
// the string-list is indicated by *two* '\0' chars.
reterr_t RegKey::SetIndexIntoStringList( PCTSTR pstrValue, _idx_ nIndex, PCTSTR pstrStringList )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);
    if( nIndex < 0 || !pstrStringList ) return syserr (ERROR_INVALID_PARAMETER);

    PCTSTR pstr = pstrStringList;

    while( nIndex > 0 && *pstr ) {
        pstr += _strlen_( pstr ) + 1;
        --nIndex;
    }

    return SetString( pstrValue, pstr );
}

// Gets the string value and returns the index of that string found in the string-list.
// If value-string is not found in the string-list, -1 will be returned as index (and
// errEndOfData will be returned then as error code).
reterr_t RegKey::GetIndexIntoStringList( PCTSTR pstrValue, _idx_& nIndex, PCTSTR pstrStringList )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    nIndex = -1;
    _size_t_ cnt = 0;
    _size_t_ maxlen = 0;
    PCTSTR pstr = pstrStringList;

    while( *pstr ) {
        ++cnt;
        _size_t_ l = _strlen_( pstr );
        if( l > maxlen ) maxlen = l;
        pstr += l + 1;
    }
    if( cnt == 0 || maxlen == 0 )
        return interr (errEndOfData);
    ++maxlen;

    String str;
    reterr_t err = GetString( pstrValue, str );
    if( !err ) return err;

    PCTSTR ptr = str;
    pstr = pstrStringList;

    for( nIndex = 0 ; nIndex < (_idx_) cnt ; ++nIndex, pstr += _strlen_( pstr ) + 1 )
        if( _stris_( pstr, ptr ) )
            return true;

    nIndex = -1;

    return interr (errEndOfData);
}

// Sets the bits of the dword but where each bit has a name and store the names.
//  : dwBits = the bits which shall be stored
//  : pstrStringList_Names = a string-list which represent the names.
//                           Caution: the names themselves *must not* contain any spaces,
//                            because spaces will be used to separate the names in the string
//                            finally stored in the registry-string-value.
//  : pstrStringList_BitNrs = a string-list of bit-numbers where each substring is the number
//                             of the bit with the name taken from the according substring
//                             in 'pstrStringList_Names'.
// The string stored in the registry-string-value will contain several
//  names separated by #chSep# characters.
// Note:
//  A string-list must be consecitive strings separated by '\0' chars. The end of
//  the string-list is indicated by *two* '\0' chars.
// NOTE:
//  If there were bits set in 'dwBits' which could not be stored because they were not
//  defined in 'pszStringList_BitNrs' and 'pszStringList_Names', then *no* error will
//  be returned. The bits simply will be ignored without causing any error.
reterr_t RegKey::SetDWordBitsAsStringList( PCTSTR pstrValue, DWORD dwBits,
                                           PCTSTR pstrStringList_Names, PCTSTR pstrStringList_BitNrs,
                                           TCHAR chSep )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);
    if( !pstrStringList_Names || !pstrStringList_BitNrs )
        return syserr (ERROR_INVALID_PARAMETER);

    REGLOC;

    String strResult;
    for( int bit = 0 ; bit < (int) sizeof(DWORD) * 8 ; ++bit )
        if( dwBits & (1 << bit) )
        {
            // first find the bit-number in #pstrStringList_BitNrs#
            _idx_ idx_found = -1;

            _idx_ idx = 0;
            PCTSTR curr_bitnr_str = pstrStringList_BitNrs;
            while( curr_bitnr_str )
            {
                while( _isspace_( *curr_bitnr_str ) )
                    ++curr_bitnr_str;

                if( _stoi_( curr_bitnr_str ) == bit ) {
                    idx_found = idx;
                    break;
                }

                ++idx;
                curr_bitnr_str = _fwd_to_next_strlst_str_( curr_bitnr_str );
            }

            // now add the name as word to the string to be stored
            if( idx_found >= 0 )
            {
                PCTSTR name_str = _get_strlst_str_ptr_( pstrStringList_Names, idx_found );
                if( !_isstrempty_( name_str ) ) {
                    if( !strResult.IsEmpty() ) strResult += chSep;
                    strResult += name_str;
                }
            }
        }

    return SetString( pstrValue, strResult );
}

// Gets the bits of a dword stored as a string with names where each name represents a bit.
// Params like for 'SetDWordBitsAsStringList()', except that 'dwBits' will be
// the result rather than the input.
// IMPORTANT:
//  The names in 'pstrStringList_Names' are case-sensitive !
// NOTE:
//  If in the string stored in the registry-string-value does contain some words which do
//  not match to any name of 'pstrStringList_Names', then *no* error will be returned.
//  These words simply will be ignored without causing any error.
reterr_t RegKey::GetDWordBitsAsStringList( PCTSTR pstrValue, DWORD& dwBits,
                                           PCTSTR pstrStringList_Names, PCTSTR pstrStringList_BitNrs,
                                           TCHAR chSep )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);
    if( !pstrStringList_Names || !pstrStringList_BitNrs )
        return syserr (ERROR_INVALID_PARAMETER);

    REGLOC;

    String strStored;
    reterr_t err = GetString( pstrValue, strStored );
    if( !err ) return err;

    TCHAR sep[2] = { chSep, (TCHAR) 0 };

    DWORD dw = 0;
    _idx_ iparse = 0;
    for( _idx_ iword = 0 ;; ++iword )
    {
        String word;
        iparse = strStored.Findnextword( word, iparse, 1, findword_SeparatorChars, sep );
        if( iparse < 0 ) break;

        // first find the name in #pstrStringList_Names# which is identical to #word#
        _idx_ idx_found = _findstrlststr_( word, pstrStringList_Names );

        // now find the bit-number to the found name and store it in 'dw'
        if( idx_found >= 0 )
        {
            PCTSTR bitnr_str = _get_strlst_str_ptr_( pstrStringList_BitNrs, idx_found );
            if( !_isstrempty_( bitnr_str ) ) {
                int bit = _stoi_( bitnr_str );
                if( bit >= 0 && bit < (int) sizeof(DWORD) * 8 )
                    dw |= (1 << bit);
            }
        }
    }

    dwBits = dw;

    return true;
}

reterr_t RegKey::SetColorrefAsString( PCTSTR pstrValue, COLORREF crColor )
{
    return SetColorRGBAsString( pstrValue, GetRValue(crColor), GetGValue(crColor), GetBValue(crColor) );
}

reterr_t RegKey::GetColorrefFromString( PCTSTR pstrValue, COLORREF& crColor )
{
    BYTE a, r, g, b;
    reterr_t err = GetColorARGBFromString( pstrValue, a, r, g, b );
    if( !err ) return err;

    crColor = RGB(r, g, b);

    return true;
}

reterr_t RegKey::SetColorRGBAsString( PCTSTR pstrValue, BYTE r, BYTE g, BYTE b )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    REGLOC;

    TCHAR s[80];
    _sprintf_( s, _T("(r%03u,g%03u,b%03u)"),
               (UINT) r, (UINT) g, (UINT) b);

    return SetString( pstrValue, s );
}

reterr_t RegKey::GetColorRGBFromString( PCTSTR pstrValue, BYTE& r, BYTE& g, BYTE& b )
{
    BYTE a;
    return GetColorARGBFromString( pstrValue, a, r, g, b );
}

reterr_t RegKey::SetColorARGBAsString( PCTSTR pstrValue, BYTE a, BYTE r, BYTE g, BYTE b )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    REGLOC;

    TCHAR s[80];
    _sprintf_( s, _T("(a%03u,r%03u,g%03u,b%03u)"),
               (UINT) a, (UINT) r, (UINT) g, (UINT) b);

    return SetString( pstrValue, s );
}

reterr_t RegKey::GetColorARGBFromString( PCTSTR pstrValue, BYTE& a, BYTE& r, BYTE& g, BYTE& b )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    TCHAR s[80];
    reterr_t err = GetString( pstrValue, s, _countof(s) );
    if( !err ) return err;

    REGLOC;

    _size_t_ len = _strlen_( s );
    if( len <= 0 ) return syserr (ERROR_INVALID_DATA);

    if( s[0] != _T('(') && s[len - 1] != _T(')') )
        return syserr (ERROR_INVALID_DATA);

    a = r = g = b = 0;
    PCTSTR p;

    p = _strichr_( s, _T('a') ); if( p++ ) { while( _isspace_(*p) ) ++p;  a = (BYTE) _stoi_( p ); }
    p = _strichr_( s, _T('r') ); if( p++ ) { while( _isspace_(*p) ) ++p;  r = (BYTE) _stoi_( p ); }
    p = _strichr_( s, _T('g') ); if( p++ ) { while( _isspace_(*p) ) ++p;  g = (BYTE) _stoi_( p ); }
    p = _strichr_( s, _T('b') ); if( p++ ) { while( _isspace_(*p) ) ++p;  b = (BYTE) _stoi_( p ); }

    return true;
}

reterr_t RegKey::SetPointAsString( PCTSTR pstrValue, POINT pt )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    REGLOC;

    TCHAR s[80];
    _sprintf_( s, _T("(%d,%d)"), (int) pt.x, (int) pt.y );

    return SetString( pstrValue, s );
}

reterr_t RegKey::GetPointFromString( PCTSTR pstrValue, POINT& pt )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    TCHAR s[80];
    reterr_t err = GetString( pstrValue, s, _countof(s) );
    if( !err ) return err;

    REGLOC;

    int x = 0, y = 0;
    if( _stscanf( s, _T("(%d,%d)"), &x, &y ) != 2 )
        return syserr (ERROR_INVALID_DATA);

    pt.x = (LONG) x;
    pt.y = (LONG) y;

    return true;
}

reterr_t RegKey::SetSizeAsString( PCTSTR pstrValue, SIZE size )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    REGLOC;

    TCHAR s[80];
    _sprintf_( s, _T("[%d,%d]"), (int) size.cx, (int) size.cy );

    return SetString( pstrValue, s );
}

reterr_t RegKey::GetSizeFromString( PCTSTR pstrValue, SIZE& size )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    TCHAR s[80];
    reterr_t err = GetString( pstrValue, s, _countof(s) );
    if( !err ) return err;

    REGLOC;

    int cx = 0, cy = 0;
    if( _stscanf( s, _T("[%d,%d]"), &cx, &cy ) != 2 )
        return syserr (ERROR_INVALID_DATA);

    size.cx = (LONG) cx;
    size.cy = (LONG) cy;

    return true;
}

// Sets a range (start point with width) stored as string
reterr_t RegKey::SetRangeAsString( PCTSTR pstrValue, long nStart, long nWidth )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    REGLOC;

    TCHAR s[80];
    _sprintf_( s, _T("(%d)[%d]"), (int) nStart, (int) nWidth );

    return SetString( pstrValue, s );
}

// Gets a range (start point with width) stored as string
// If the method fails, the value will not be changed.
reterr_t RegKey::GetRangeFromString( PCTSTR pstrValue, long& nStart, long& nWidth )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    TCHAR s[80];
    reterr_t err = GetString( pstrValue, s, _countof(s) );
    if( !err ) return err;

    REGLOC;

    int i = 0, w = 0;
    if( _stscanf( s, _T("(%d)[%d]"), &i, &w ) != 2 )
        return syserr (ERROR_INVALID_DATA);

    nStart = (long) i;
    nWidth = (long) w;

    return true;
}

reterr_t RegKey::SetRectAsString( PCTSTR pstrValue, RECT rect )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    REGLOC;

    TCHAR s[80];
    _sprintf_( s, _T("(%d,%d)[%d,%d]"),
               (int) rect.left, (int) rect.top,
               (int) (rect.right - rect.left), (int) (rect.bottom - rect.top) );

    return SetString( pstrValue, s );
}

reterr_t RegKey::GetRectFromString( PCTSTR pstrValue, RECT& rect )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    TCHAR s[80];
    reterr_t err = GetString( pstrValue, s, _countof(s) );
    if( !err ) return err;

    REGLOC;

    int x = 0, y = 0, cx = 0, cy = 0;
    if( _stscanf( s, _T("(%d,%d)[%d,%d]"), &x, &y, &cx, &cy ) != 4 )
        return syserr (ERROR_INVALID_DATA);

    rect.right  = ( rect.left = (LONG) x ) + (LONG) cx;
    rect.bottom = ( rect.top  = (LONG) y ) + (LONG) cy;

    return true;
}

// Sets a list of Rects and stores it as a string
reterr_t RegKey::SetRectListAsString( PCTSTR pstrValue, const RECT *prcRects, _size_t_ cRects )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);
    if( !prcRects && cRects > 0 ) return syserr (ERROR_INVALID_PARAMETER);

    REGLOC;

    String str;
    for( _idx_ i = 0 ; (_size_t_) i < cRects ; ++i )
    {
        if( i > 0 ) str += _T(',');

        TCHAR s[80];
        _sprintf_( s, _T("(%d,%d)[%d,%d]"),
                   (int) prcRects[i].left, (int) prcRects[i].top,
                   (int) (prcRects[i].right - prcRects[i].left), (int) (prcRects[i].bottom - prcRects[i].top) );

        str += s;
    }

    return SetString( pstrValue, str );
}

// Gets a list of Rects from a string
//  : pprcRects = will contain the result.
//              NULL = just 'cRects' shall be retrieved but not the rects themselves.
//  > pprcRects : will be filled with a pointer to an array of RECT structs.
//                To free the memory call #free()#.
//  > cRects : will contain the number of rects of the new 'pprcRects' array.
reterr_t RegKey::GetRectListFromString( PCTSTR pstrValue, PRECT *pprcRects, _size_t_& cRects )
{
    if( !isOpen() ) return interr (errRegDerivedNotOpen);

    _size_t_ cnt;

    String str;
    reterr_t err = GetString( pstrValue, str );
    if( !err ) return err;

    PCTSTR p = str;

    REGLOC;

    for( cnt = 0 ;; ++cnt )
    {
        int x = 0, y = 0, cx = 0, cy = 0;
        if( _stscanf( p, _T("(%d,%d)[%d,%d]"), &x, &y, &cx, &cy ) != 4 )
            break;

        while( *p && *p != _T(']') ) ++p;
        while( *p && *p != _T('(') ) ++p;
    }

    if( cnt == 0 )
    {
        if( pprcRects ) *pprcRects = NULL;
        cRects = 0;
        return true;
    }

    if( !pprcRects )
        { cRects = cnt; return true; }

    p = str;

    PRECT prcRects = (PRECT) malloc( cnt * sizeof(RECT) );
    if( !prcRects )
        return syserr (ERROR_OUTOFMEMORY);

    PRECT prc = prcRects;
    for( _idx_ i = 0 ; (_size_t_) i < cnt ; ++i )
    {
        int x = 0, y = 0, cx = 0, cy = 0;
        if( _stscanf( p, _T("(%d,%d)[%d,%d]"), &x, &y, &cx, &cy ) != 4 )
            break;

        prc->right  = ( prc->left = (LONG) x ) + (LONG) cx;
        prc->bottom = ( prc->top  = (LONG) y ) + (LONG) cy;
        ++prc;

        while( *p && *p != _T(']') ) ++p;
        while( *p && *p != _T('(') ) ++p;
    }

    if( pprcRects )
        *pprcRects = prcRects;

    cRects = cnt;

    return true;
}

#ifndef _WIN32_WCE
reterr_t RegKey::SetWindowPlacement( PCTSTR pstrValue, const WINDOWPLACEMENT& wp )
{
    //wp.length = sizeof(WINDOWPLACEMENT);
    return SetBinary( pstrValue, (const void*) &wp, sizeof(WINDOWPLACEMENT) );
}

reterr_t RegKey::GetWindowPlacement( PCTSTR pstrValue, WINDOWPLACEMENT& wp )
{
    return GetBinary( pstrValue, (void*) &wp, sizeof(WINDOWPLACEMENT) );
    //wp.length = sizeof(WINDOWPLACEMENT);
}

reterr_t RegKey::StoreWindowPlacementFromWnd( PCTSTR pstrValue, HWND hwnd )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue || !hwnd ) return syserr (ERROR_INVALID_PARAMETER);

    WINDOWPLACEMENT wp;
    wp.length = sizeof(wp);
    if( !::GetWindowPlacement( hwnd, &wp ) )
        return syserr (GetLastError());

    return SetWindowPlacement( pstrValue, wp );
}

///<param name="pwp"> if not NULL, will contain the window placement data
///  read from the RegKey storage. </param>
reterr_t RegKey::LoadWindowPlacementIntoWnd( PCTSTR pstrValue, HWND hwnd, WINDOWPLACEMENT *pwp )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue || !hwnd ) return syserr (ERROR_INVALID_PARAMETER);

    WINDOWPLACEMENT wp;
    reterr_t ret = GetWindowPlacement( pstrValue, wp );
    if( !ret ) return ret;

    if( pwp ) *pwp = wp;

    if( !::SetWindowPlacement( hwnd, &wp ) )
        return syserr (GetLastError());

    return true;
}

reterr_t RegKey::StoreWindowposTopleftFromWnd( PCTSTR pstrValue, HWND hwnd )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue || !hwnd ) return syserr (ERROR_INVALID_PARAMETER);

    RECT rect;
    if( !::GetWindowRect( hwnd, &rect ) )
        return syserr (GetLastError());

    return SetPointAsString( pstrValue, *(POINT*) &rect );
}

///<param name="ppt"> if not NULL, will contain the window top-left coordinate
///  read from the RegKey storage. </param>
reterr_t RegKey::LoadWindowposTopleftIntoWnd( PCTSTR pstrValue, HWND hwnd, POINT *ppt )
{
    if( !m_hKey ) return interr (errRegNotOpen);
    if( !pstrValue || !hwnd ) return syserr (ERROR_INVALID_PARAMETER);

    POINT pt;
    reterr_t ret = GetPointFromString( pstrValue, pt );
    if( !ret ) return ret;

    if( ppt ) *ppt = pt;

    if( !::SetWindowPos( hwnd, HWND_TOP, pt.x, pt.y, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE ) )
        return syserr (GetLastError());

    return true;
}
#endif




#ifndef _argon_no_PrivateProfile

reterr_t PrivateProfile::Create( PCTSTR pstrProfileFile, PCTSTR pstrSection )
{
    if( !m_strProfileFile.IsEmpty() ) return interr (errProfileAlreadyOpen);
    if( _isstrempty_(pstrProfileFile) || _isstrempty_(pstrSection) ) return syserr (ERROR_INVALID_PARAMETER);

    if( m_strProfileFile.GetFullpathname( pstrProfileFile ).IsEmpty() )
        { m_strProfileFile.Empty(); return syserr (ERROR_INVALID_DATA); }

    m_strSection = pstrSection;

    return true;
}

reterr_t PrivateProfile::Open( PCTSTR pstrProfileFile, PCTSTR pstrSection )
{
    if( !m_strProfileFile.IsEmpty() ) return interr (errProfileAlreadyOpen);
    if( _isstrempty_(pstrProfileFile) || _isstrempty_(pstrSection) ) return syserr (ERROR_INVALID_PARAMETER);

    if( m_strProfileFile.GetFullpathname( pstrProfileFile ).IsEmpty() )
        { m_strProfileFile.Empty(); return syserr (ERROR_INVALID_DATA); }

    // Since we want to open it, the file must already exist.
    WIN32_FIND_DATA w32fd;
    HANDLE hfind = ::FindFirstFile( m_strProfileFile, &w32fd );

    bool fileexists = hfind != INVALID_HANDLE_VALUE;
    if( hfind != INVALID_HANDLE_VALUE ) ::FindClose( hfind );

    if( !fileexists )
        { m_strProfileFile.Empty(); return syserr (ERROR_FILE_NOT_FOUND); }

    m_strSection = pstrSection;

    return true;
}

reterr_t PrivateProfile::Create( const RegKey& parentkey, PCTSTR pstrSubkey )
{
    if( parentkey.m_Storeloc != StoreLocationInPrivateProfile )
        return syserr (ERROR_INVALID_DATA);

    const PrivateProfile *ptr = (PrivateProfile*) &parentkey;

    return Create( ptr->m_strProfileFile, ptr->m_strSection | pstrSubkey );
}

reterr_t PrivateProfile::Open( const RegKey& parentkey, PCTSTR pstrSubkey )
{
    if( parentkey.m_Storeloc != StoreLocationInPrivateProfile )
        return syserr (ERROR_INVALID_DATA);

    const PrivateProfile *ptr = (PrivateProfile*) &parentkey;

    return Open( ptr->m_strProfileFile, ptr->m_strSection | pstrSubkey );
}

reterr_t PrivateProfile::Close()
{
    m_strProfileFile.Empty();
    m_strSection.Empty();
    return true;
}



bool PrivateProfile::EnumValue( _idx_ nValue, PTSTR pstrValue, _size_t_ zlenValue, PDWORD pdwType )
{
    if( !isValid() ) return false;
    if( !pstrValue || zlenValue <= 0 || nValue < 0 ) return false;

    _size_t_ zlen = 0x10000;
    Array<BYTE> dat;
    dat.SetSize( zlen * sizeof(TCHAR) );

    // Get all the key names in the current section as string list
    GetPrivateProfileString( m_strSection, NULL, _T(""), (PTSTR) dat.Ptr(), zlen, m_strProfileFile );

    PCTSTR pstr = (PCTSTR) dat.Ptr();
    if( !pstr[0] && !pstr[1] ) return false;

    while( pstr && nValue >= 0 )
    {
        if( nValue == 0 ) {
            _strcpy_( pstrValue, pstr, zlenValue );
            if( pdwType )
                *pdwType = REG_SZ;
            return true;
        }

        --nValue;
        pstr = _fwd_to_next_strlst_str_( pstr );
    }

    return false;
}

bool PrivateProfile::EnumValue( _idx_ nValue, String& strValue, PDWORD pdwType )
{
    TCHAR s[REGKEY_MAXNAMELEN];

    if( !EnumValue( nValue, s, _countof(s), pdwType ) )
        return false;

    strValue = s;

    return true;
}



bool PrivateProfile::isValue( PCTSTR pstrValue )
{
    if( !isValid() ) return false;
    if( _isstrempty_(pstrValue) ) return false;

    _size_t_ zlen = 0x10000;
    Array<BYTE> dat;
    dat.SetSize( zlen * sizeof(TCHAR) );

    // Get all the key names in the current section as string list
    GetPrivateProfileString( m_strSection, NULL, _T(""), (PTSTR) dat.Ptr(), zlen, m_strProfileFile );

    PCTSTR pstr = (PCTSTR) dat.Ptr();
    if( !pstr[0] && !pstr[1] ) return false;

    while( pstr )
    {
        if( _striis_( pstr, pstrValue ) )
            return true;

        pstr = _fwd_to_next_strlst_str_( pstr );
    }

    return false;
}



reterr_t PrivateProfile::DeleteValue( PCTSTR pstrValue )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) ) return syserr (ERROR_INVALID_PARAMETER);

    if( !WritePrivateProfileString( m_strSection, pstrValue, NULL, m_strProfileFile ) )
        return syserr (GetLastError());

    return true;
}



#define  PP_STR_UNDEF    "<argonLib___undef>"

reterr_t PrivateProfile::SetDWord( PCTSTR pstrValue, DWORD val )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) ) return syserr (ERROR_INVALID_PARAMETER);

    REGLOC;

    TCHAR s[80];
    _sprintf_( s, _T("%lu"), val );

    if( !WritePrivateProfileString( m_strSection, pstrValue, s, m_strProfileFile ) )
        return syserr (GetLastError());

    return true;
}

reterr_t PrivateProfile::GetDWord( PCTSTR pstrValue, DWORD& val )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) ) return syserr (ERROR_INVALID_PARAMETER);

    TCHAR s[128];
    GetPrivateProfileString( m_strSection, pstrValue, _T(PP_STR_UNDEF), s, _countof(s), m_strProfileFile );

    REGLOC;

    if( _stris_( s, _T(PP_STR_UNDEF) ) )
        return interr (errNoDataAvailable);

    val = (DWORD) _stoI_( s );  // #s# can be a decimal, hexadecimal or octal number

    return true;
}

reterr_t PrivateProfile::SetBinary( PCTSTR pstrValue, const void *pData, _size_t_ cData )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) || !pData || cData <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( !WritePrivateProfileStruct( m_strSection, pstrValue, (void*) pData, cData, m_strProfileFile ) )
        return syserr (GetLastError());

    return true;
}

///<remarks> Like <b>GetPrivateProfileStruct()</b>, <b>GetBinary()</b>
///  will fail if the specifed buffer size is not exactly the size of
///  the binary data in the profile. <br/>
///  HINTS: The size of a binary data can be retrieved using #GetString()#
///  and divide the string-length by 2 (because a hexadecimal value is used
///  for each binary byte) and then subtracted with 1 (because the last
///  2 chars represent a checksum byte). </remarks>
reterr_t PrivateProfile::GetBinary( PCTSTR pstrValue, void *pData, _size_t_ cData )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) || !pData || cData <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( !GetPrivateProfileStruct( m_strSection, pstrValue, pData, cData, m_strProfileFile ) )
        return interr (errNoDataAvailable);

    return true;
}

reterr_t PrivateProfile::GetBinaryPart( PCTSTR pstrValue, void *pData, _size_t_ cData )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) || !pData || cData <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    // Note: There is no explicit way to determine neither the type
    // nor the size of a value in the profile file.
    // But since we know that binary data is stored as a string of
    // alphabetic characters where each pair represents a hexadecimal
    // byte value, we can determine a binary data size from
    // the string length holding in mind that the last 2 chars
    // represent a checksum byte.
    String str;
    reterr_t err = GetString( pstrValue, str );
    if( !err ) return err;

    if( str.length() & 1 ) return syserr (ERROR_INVALID_DATA);
    _size_t_ size = str.length() / 2 - 1;

    if( size > cData )
    {
        // Special handling.

        Array<BYTE> mem;
        PBYTE ptr = mem.AddNew( size );
        if( !ptr )
            return syserr (ERROR_OUTOFMEMORY);

        m_dwLastValueSize = (DWORD) size;

        if( !GetPrivateProfileStruct( m_strSection, pstrValue, ptr, size, m_strProfileFile ) )
            return interr (errNoDataAvailable);

        memcpy( pData, ptr, cData );

        return true;
    }

    // Same handling as in #GetBinary()#.

    if( !GetPrivateProfileStruct( m_strSection, pstrValue, pData, cData, m_strProfileFile ) )
        return interr (errNoDataAvailable);

    return true;
}

reterr_t PrivateProfile::SetString( PCSTR pstrValue, PCSTR pStr )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) || !pStr ) return syserr (ERROR_INVALID_PARAMETER);

    if( !WritePrivateProfileStringA( StringA(m_strSection), pstrValue, pStr, StringA(m_strProfileFile) ) )
        return syserr (GetLastError());

    return true;
}

reterr_t PrivateProfile::SetString( PCWSTR pstrValue, PCWSTR pStr )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) || !pStr ) return syserr (ERROR_INVALID_PARAMETER);

    if( !WritePrivateProfileStringW( StringW(m_strSection), pstrValue, pStr, StringW(m_strProfileFile) ) )
        return syserr (GetLastError());

    return true;
}

reterr_t PrivateProfile::GetString( PCSTR pstrValue, PSTR pStr, _size_t_ zlenStr )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) || !pStr || zlenStr <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    // First we get the first part of the string to see if the
    // string exists. We do this by comparing with default undef string.
    CHAR s[128];
    _size_t_ len = (_size_t_) GetPrivateProfileStringA( StringA(m_strSection), pstrValue, PP_STR_UNDEF, s, _countof(s), StringA(m_strProfileFile) );

    if( _stris_( s, PP_STR_UNDEF ) )
        return interr (errNoDataAvailable);

    // Now we get the complete string.
    if( len < _countof(s) - 1 )
        _strcpy_( pStr, s, zlenStr );
    else
        GetPrivateProfileStringA( StringA(m_strSection), pstrValue, "", pStr, zlenStr, StringA(m_strProfileFile) );

    return true;
}

reterr_t PrivateProfile::GetString( PCWSTR pstrValue, PWSTR pStr, _size_t_ zlenStr )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) || !pStr || zlenStr <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    // First we get the first part of the string to see if the
    // string exists. We do this by comparing with default undef string.
    WCHAR s[128];
    _size_t_ len = (_size_t_) GetPrivateProfileStringW( StringW(m_strSection), pstrValue, _W(PP_STR_UNDEF), s, _countof(s), StringW(m_strProfileFile) );

    if( _stris_( s, _W(PP_STR_UNDEF) ) )
        return interr (errNoDataAvailable);

    // Now we get the complete string.
    if( len < _countof(s) - 1 )
        _strcpy_( pStr, s, zlenStr );
    else
        GetPrivateProfileStringW( StringW(m_strSection), pstrValue, L"", pStr, zlenStr, StringW(m_strProfileFile) );

    return true;
}

reterr_t PrivateProfile::GetString( PCTSTR pstrValue, String& str )
{
    if( !isValid() ) return interr (errProfileNotOpen);
    if( _isstrempty_(pstrValue) ) return syserr (ERROR_INVALID_PARAMETER);

    // Now get the string into a buffer and in a loop, we enlarge
    // the buffer till the whole string is loaded.
    String s;

    _size_t_ zlenBuf = 256;
    if( !s.GetBuffer( zlenBuf ) )
        return false;

    _size_t_ readlen = 0;
    zlenBuf >>= 1;

    do
    {
        if( zlenBuf >= 0x100000 ) zlenBuf += 0x100000;
        else zlenBuf <<= 1;

        if( zlenBuf == 0 )
            { s.ReleaseBuffer(); return syserr (ERROR_INVALID_DATA); }

        PTSTR ptr = s.GetBuffer( zlenBuf - 1 );
        if( !ptr )
            { s.ReleaseBuffer(); return false; }

        readlen = (_size_t_) GetPrivateProfileString( m_strSection, pstrValue, _T(PP_STR_UNDEF), ptr, zlenBuf, m_strProfileFile );

        if( _stris_( ptr, _T(PP_STR_UNDEF) ) )
            { s.ReleaseBuffer(); return interr (errNoDataAvailable); }
    } while( readlen >= zlenBuf - 1 );

    s.ReleaseBuffer();

    str = s;

    return true;
}

#endif




StoreLocation::StoreLocation( storelocation_t storeloc )
{
    m_Storeloc = storeloc;
    m_hRegKey  = NULL;
}

StoreLocation::StoreLocation( PCTSTR pstrFrom )
{
    m_Storeloc = StoreLocationInRegistry;
    m_hRegKey  = NULL;

    fromString( pstrFrom );
}



bool StoreLocation::IsEmpty() const
{
    switch( m_Storeloc )
    {
        case StoreLocationInRegistry :
            return !m_hRegKey || m_strRegSubkey.IsEmpty();

        #ifndef _argon_no_PrivateProfile
        case StoreLocationInPrivateProfile :
            return m_strProfileFile.IsEmpty() || m_strSection.IsEmpty();
        #endif
    }

    return true;
}

StoreLocation& StoreLocation::Empty()
{
    m_hRegKey = NULL;
    m_strRegSubkey.Empty();

    #ifndef _argon_no_PrivateProfile
    m_strProfileFile.Empty();
    m_strSection.Empty();
    #endif

    return *this;
}

StoreLocation& StoreLocation::Copy( const StoreLocation& storeloc )
{
    m_Storeloc          = storeloc.m_Storeloc;

    m_hRegKey           = storeloc.m_hRegKey;
    m_strRegSubkey      = storeloc.m_strRegSubkey;

    #ifndef _argon_no_PrivateProfile
    m_strProfileFile    = storeloc.m_strProfileFile;
    m_strSection        = storeloc.m_strSection;
    #endif

    return *this;
}



StoreLocation& StoreLocation::SetRegistryCreate( HKEY hKey, PCTSTR pstrSubkey )
{
    Empty();

    m_Storeloc          = StoreLocationInRegistry;

    m_hRegKey           = hKey;
    m_strRegSubkey      = pstrSubkey;

    return *this;
}

StoreLocation& StoreLocation::SetRegistryOpen( HKEY hKey, PCTSTR pstrSubkey )
{
    Empty();

    m_Storeloc          = StoreLocationInRegistry;

    m_hRegKey           = hKey;
    m_strRegSubkey      = pstrSubkey;

    return *this;
}

#ifndef _argon_no_PrivateProfile
StoreLocation& StoreLocation::SetPrivateProfileCreate( PCTSTR pstrProfileFile, PCTSTR pstrSection )
{
    Empty();

    m_Storeloc          = StoreLocationInPrivateProfile;

    m_strProfileFile    = pstrProfileFile;
    m_strSection        = pstrSection;

    return *this;
}

StoreLocation& StoreLocation::SetPrivateProfileOpen( PCTSTR pstrProfileFile, PCTSTR pstrSection )
{
    Empty();

    m_Storeloc          = StoreLocationInPrivateProfile;

    m_strProfileFile    = pstrProfileFile;
    m_strSection        = pstrSection;

    return *this;
}
#endif



///<summary>
///  Tries to convert the current store location information into
///  a string. It *tries* means, that it will fail if in case of registry
///  the HKEY is *not* one of the predefined HKEY_* values. <br/>
///  Syntax: <br/>
///      "storeloc:[_type_]:_param1_|_param2_|..." <br/>
///  Registry sample: <br/>
///      "storeloc:[Registry]:HKCU/Software/myName/myApp" <br/>
///  Private Profile sample: <br/>
///      "storeloc:[PrivateProfile]:C:\Program Files\myApp\settings.ini|Properties/BrowserProps/IE-Settings" <br/>
///   In this sample, the param1 is the file and param2 is the section (which may contain path delimiters).
///</summary>
String StoreLocation::toString( DWORD flags )
{
    String s;

    switch( m_Storeloc )
    {
        case StoreLocationInRegistry :
        {
            if( !(flags & storelocf_no_prefixLocation) )
                s = _T("storeloc:[Registry]:");

            PCTSTR pstrHk = GetHkeyPredefName( m_hRegKey );
            if( !pstrHk ) return String();

            s += pstrHk | m_strRegSubkey;
            s.UnixPath();

            break;
        }

        #ifndef _argon_no_PrivateProfile
        case StoreLocationInPrivateProfile :
        {
            if( !(flags & storelocf_no_prefixLocation) )
                s = _T("storeloc:[PrivateProfile]:");

            s += m_strProfileFile;
            s += _T('|');
            s += m_strSection;

            break;
        }
        #endif
    }

    return s;
}

reterr_t StoreLocation::fromString( PCTSTR pstrStoreloc )
{
    static PCTSTR pstrSlReg = _T("storeloc:[Registry]:");
    #ifndef _argon_no_PrivateProfile
    static PCTSTR pstrSlPP  = _T("storeloc:[PrivateProfile]:");
    #endif

    if( _strishere_(pstrStoreloc, pstrSlReg) )
    {
        PCTSTR p = pstrStoreloc + _strlen_(pstrSlReg);
        return fromStringReg( p );
    }

    #ifndef _argon_no_PrivateProfile
    if( _strishere_(pstrStoreloc, pstrSlPP) )
    {
        PCTSTR p = pstrStoreloc + _strlen_(pstrSlPP);
        return fromStringPP( p );
    }
    #endif

    return false;
}

reterr_t StoreLocation::fromStringReg( PCTSTR pstrRegistryLoc )
{
    if( _isstrempty_(pstrRegistryLoc) ) return syserr (ERROR_INVALID_PARAMETER);

    if( _strishere_(pstrRegistryLoc, _T("storeloc:")) )
        return fromString( pstrRegistryLoc );

    String fullpath = pstrRegistryLoc;

    String hk       = fullpath.Findword( 0, findword_SeparatorChars, _T("\\/") );
    String subkey   = fullpath.Inflate( hk.length() + 1, 0 ).Trim();

    if( fullpath.IsEmpty() || hk.IsEmpty() || subkey.IsEmpty() )
        return false;

    HKEY hkey = GetHkeyByPredefName( hk );
    if( !hkey )
        return false;

    Empty();

    m_Storeloc     = StoreLocationInRegistry;
    m_hRegKey      = hkey;
    m_strRegSubkey = subkey;

    return true;
}

#ifndef _argon_no_PrivateProfile
reterr_t StoreLocation::fromStringPP( PCTSTR pstrPrivateProfileLoc )
{
    if( _isstrempty_(pstrPrivateProfileLoc) ) return syserr (ERROR_INVALID_PARAMETER);

    if( _strishere_(pstrPrivateProfileLoc, _T("storeloc:")) )
        return fromString( pstrPrivateProfileLoc );

    String loc      = pstrPrivateProfileLoc;
    String filepath = loc.Findword( 1, findword_SeparatorChars, _T("|") );
    String section  = loc.Findword( 2, findword_SeparatorChars, _T("|") );

    if( filepath.IsEmpty() || section.IsEmpty() )
        return false;

    Empty();

    m_Storeloc       = StoreLocationInPrivateProfile;
    m_strProfileFile = filepath;
    m_strSection     = section;

    return true;
}
#endif



///<summary> Creates a StoreLocation object the indicates a location
///  below the current <b>this</b> location. </summary>
StoreLocation StoreLocation::GetSubLocation( PCTSTR pstrSubkey )
{
    if( _isstrempty_(pstrSubkey) ) return *this;

    StoreLocation loc( m_Storeloc );

    switch( m_Storeloc )
    {
        case StoreLocationInRegistry :
        {
            if( !m_hRegKey || m_strRegSubkey.IsEmpty() )
                return loc;

            loc.m_strRegSubkey  = m_strRegSubkey | pstrSubkey;
            if( loc.m_strRegSubkey.IsEmpty() ) return loc;

            loc.m_hRegKey       = m_hRegKey;

            break;
        }

        #ifndef _argon_no_PrivateProfile
        case StoreLocationInPrivateProfile :
        {
            if( m_strProfileFile.IsEmpty() || m_strSection.IsEmpty() )
                return loc;

            loc.m_strSection     = m_strSection | pstrSubkey;
            if( loc.m_strSection.IsEmpty() ) return loc;

            loc.m_strProfileFile = m_strProfileFile;

            break;
        }
        #endif
    }

    return loc;
}



///<summary> Create the appropriate storage information object
///  according to the current location settings. </summary>
///<remarks> If the current <b>m_Storeloc</b> is <b>StoreLocationInRegistry</b>,
///  a <b>RegKey</b> object will be created. <br/>
///  For <b>StoreLocationInPrivateProfile</b>, a <b>PrivateProfile</b>
///  object will be created. <br/><br/>
///  IMPORTANT: The returned object is occupied by you. So you will
///  have to delete it when you no longer need it. </remarks>
///<param name="bCreate"> if the key (and maybe file) shall be created if it does not exist yet </param>
///<param name="samDesired"> access rights to use (used only for registry locations) </param>
///<returns> A <b>RegKey</b> or derived object,
///  or NULL if the location information is empty or invalid or
///  when an error occured. </returns>
RegKey *StoreLocation::Create_at_CurrentLocation( bool bCreate, REGSAM samDesired )
{
    return Create_at_SubLocation( NULL, bCreate, samDesired );
}



///<summary> Create the appropriate storage information object
///  according to the current location settings and the specified subkey. </summary>
///<remarks> If the current <b>m_Storeloc</b> is <b>StoreLocationInRegistry</b>,
///  a <b>RegKey</b> object will be created. <br/>
///  For <b>StoreLocationInPrivateProfile</b>, a <b>PrivateProfile</b>
///  object will be created. <br/><br/>
///  IMPORTANT: The returned object is occupied by you. So you will
///  have to delete it when you no longer need it. </remarks>
///<param name="pstrSubkeypath"> a Subkey or a path of subkeys. If this is NULL or an empty string,
///  the current location will be used only. <br/>
///  NOTE: In case of a Private Profile, the subkey will be appended to the
///  section (by joining paths like using <b>_joinpaths_()</b>). </param>
///<param name="bCreate"> if the key (and maybe file) shall be created if it does not exist yet </param>
///<param name="samDesired"> access rights to use (used only for registry locations) </param>
///<returns> A <b>RegKey</b> or derived object,
///  or NULL if the location information is empty or invalid or
///  when an error occured. </returns>
RegKey *StoreLocation::Create_at_SubLocation( PCTSTR pstrSubkeypath, bool bCreate, REGSAM samDesired )
{
    if( IsEmpty() ) return NULL;

    switch( m_Storeloc )
    {
        case StoreLocationInRegistry :
        {
            RegKey *ptr = new RegKey;
            if( !ptr ) return NULL;

            if( bCreate ) {
                if( !ptr->Create( m_hRegKey, m_strRegSubkey | pstrSubkeypath, samDesired ) )
                    { delete ptr; return NULL; }
            } else {
                if( !ptr->Open( m_hRegKey, m_strRegSubkey | pstrSubkeypath, samDesired ) )
                    { delete ptr; return NULL; }
            }

            return ptr;
        }

        #ifndef _argon_no_PrivateProfile
        case StoreLocationInPrivateProfile :
        {
            PrivateProfile *ptr = new PrivateProfile;
            if( !ptr ) return NULL;

            if( bCreate ) {
                if( !ptr->Create( m_strProfileFile, m_strSection | pstrSubkeypath ) )
                    { delete ptr; return NULL; }
            } else {
                if( !ptr->Open( m_strProfileFile, m_strSection | pstrSubkeypath ) )
                    { delete ptr; return NULL; }
            }

            return ptr;
        }
        #endif
    }

    return NULL;
}



struct HkeyNameStruct { HKEY m_hkey; PCTSTR m_pstrName; PCTSTR m_pstrShortName; };

HkeyNameStruct g_pHkeyNames[] =
{
    { HKEY_CLASSES_ROOT,           _T("HKEY_CLASSES_ROOT"),            _T("HKCR") },
    { HKEY_CURRENT_USER,           _T("HKEY_CURRENT_USER"),            _T("HKCU") },
    { HKEY_LOCAL_MACHINE,          _T("HKEY_LOCAL_MACHINE"),           _T("HKLM") },
    { HKEY_USERS,                  _T("HKEY_USERS"),                   _T("HKU")  },
    #ifdef HKEY_PERFORMANCE_DATA
    { HKEY_PERFORMANCE_DATA,       _T("HKEY_PERFORMANCE_DATA"),        _T("HKPD") },
    #endif
    #ifdef HKEY_PERFORMANCE_TEXT
    { HKEY_PERFORMANCE_TEXT,       _T("HKEY_PERFORMANCE_TEXT"),        _T("HKPT") },
    #endif
    #ifdef HKEY_PERFORMANCE_NLSTEXT
    { HKEY_PERFORMANCE_NLSTEXT,    _T("HKEY_PERFORMANCE_NLSTEXT"),     _T("HKPN") },
    #endif
    #ifdef HKEY_CURRENT_CONFIG
    { HKEY_CURRENT_CONFIG,         _T("HKEY_CURRENT_CONFIG"),          _T("HKCC") },
    #endif
    #ifdef HKEY_DYN_DATA
    { HKEY_DYN_DATA,               _T("HKEY_DYN_DATA"),                _T("HKDD") },
    #endif
    { NULL, NULL, NULL }
};

PCTSTR StoreLocation::GetHkeyPredefName( HKEY hkey, bool bShortName )
{
    for( int i = 0 ; g_pHkeyNames[i].m_hkey ; ++i )
    {
        if( g_pHkeyNames[i].m_hkey == hkey )
            return bShortName ? g_pHkeyNames[i].m_pstrShortName : g_pHkeyNames[i].m_pstrName;
    }

    return NULL;
}

HKEY StoreLocation::GetHkeyByPredefName( PCTSTR pstrName )
{
    if( _isstrempty_(pstrName) ) return NULL;

    for( int i = 0 ; g_pHkeyNames[i].m_hkey ; ++i )
    {
        if( g_pHkeyNames[i].m_pstrName )
            if( _stris_(pstrName, g_pHkeyNames[i].m_pstrName) )
                return g_pHkeyNames[i].m_hkey;

        if( g_pHkeyNames[i].m_pstrShortName )
            if( _stris_(pstrName, g_pHkeyNames[i].m_pstrShortName) )
                return g_pHkeyNames[i].m_hkey;
    }

    return NULL;
}




}
