/*
    File:               argonRegistry.h
    Description:        Registry, PrivateProfiles.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         13.11.2003
    last updated:       26.10.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonRegistry_h
#define __argonRegistry_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonString.h"
    #include "argonArray.h"
    #include "argonErrors.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




// REGKEY_NAMESIZE is size of the name string of a key or value
// possible at maximum (excluding terminating 0)
#define REGKEY_MAXNAMELEN   _MAX_PATH_

#ifdef _WIN32_WCE
#define _argon_no_PrivateProfile
#endif

enum storelocation_t {
    StoreLocationInRegistry             = 1,
    #ifndef _argon_no_PrivateProfile
    StoreLocationInPrivateProfile,
    #endif
};

class PrivateProfile;

class RegKey
{
    friend class RegKey;
    friend class PrivateProfile;

    protected :
        storelocation_t m_Storeloc;

        HKEY    m_hKey;
        REGSAM  m_samKey;           // the access conditions for #m_hKey# key
        bool    m_bCreated;         // if key has been created (true), or just opened (false) because it did exist already
        DWORD   m_dwLastValueSize;  // Last size of the Value queried by a 'Get..()' method

        bool    m_bUse_C_Locale;    // if true, the "C" locale will be used. if false, the current locale will be used
        // The locale is relevant if you want to store time, float, ... values as strings.

    public :
        RegKey() { m_Storeloc = StoreLocationInRegistry; m_hKey = NULL; m_samKey = 0; m_bCreated = false; m_dwLastValueSize = 0; m_bUse_C_Locale = true; }
        virtual ~RegKey() { Close(); }

        operator HKEY   () const { return this ? m_hKey : NULL; }
        REGSAM   GetSAM () const { return m_samKey; }
        bool     Created() const { return m_bCreated; }

    // Creating, opening, closing
    public :
        reterr_t Create( HKEY hkeyParent, PCTSTR pstrSubkey, REGSAM samDesired = KEY_ALL_ACCESS ) // Parameters like in #RegCreateKeyEx()#
            { return _Create( hkeyParent, pstrSubkey, true, samDesired ); }
        reterr_t Open  ( HKEY hkeyParent, PCTSTR pstrSubkey, REGSAM samDesired = KEY_ALL_ACCESS ) // Parameters like in #RegOpenKeyEx()#
            { return _Create( hkeyParent, pstrSubkey, false, samDesired ); }

        // And overridable creators.
        virtual reterr_t Create( const RegKey& parentkey, PCTSTR pstrSubkey )
            { return _Create( parentkey, pstrSubkey, true, parentkey.GetSAM() ); }
        virtual reterr_t Open  ( const RegKey& parentkey, PCTSTR pstrSubkey )
            { return _Create( parentkey, pstrSubkey, false, parentkey.GetSAM() ); }

        virtual reterr_t Close();
        virtual bool     isOpen() const { return this ? m_hKey != NULL : false; }
    private :
        reterr_t _Create( HKEY hkeyParent, PCTSTR pstrSubkey, bool bCreate, REGSAM samDesired );

    // Enumeration, states, ...
    public :
        virtual bool EnumKey( _idx_ nSubKey, PTSTR pstrSubKey, _size_t_ zlenSubKey );
        virtual bool EnumKey( _idx_ nSubKey, String& strSubKey );
        virtual bool EnumValue( _idx_ nValue, PTSTR pstrValue, _size_t_ zlenValue, PDWORD pdwType = NULL );
        virtual bool EnumValue( _idx_ nValue, String& strValue, PDWORD pdwType = NULL );

        #ifndef _WIN32_WCE
        virtual reterr_t GetValueTypeSize( PCSTR  pstrValue, DWORD& dwType, _size_t_& cSizeBytes );
        virtual reterr_t GetValueTypeSize( PCWSTR pstrValue, DWORD& dwType, _size_t_& cSizeBytes );
        #else
        virtual reterr_t GetValueTypeSize( PCTSTR pstrValue, DWORD& dwType, _size_t_& cSizeBytes );
        #endif
        virtual bool isValue ( PCTSTR pstrValue ); // returns true, if the value exists
        virtual bool isString( PCTSTR pstrValue ); // returns true, if the value is of type REG_SZ or REG_EXPAND_SZ
        virtual bool isDWord ( PCTSTR pstrValue ); // returns true, if the value is of type REG_DWORD
        virtual bool isBinary( PCTSTR pstrValue ); // returns true, if the value is of type REG_BINARY
        virtual bool isKey   ( PCTSTR pstrSubkey );
        static  bool isKey   ( HKEY hkey, PCTSTR pstrKey );
        _size_t_ GetLastSize() const { return (_size_t_) m_dwLastValueSize; }

        virtual reterr_t DeleteSubKey( PCTSTR pstrSubKey );
        static  reterr_t DeleteRootKey( HKEY hRootKey, PCTSTR pstrSubKey );
        virtual reterr_t DeleteValue( PCTSTR pstrValue );
        virtual reterr_t DeleteAllValues();
        virtual reterr_t DeleteSomeValues( PCTSTR pstrMasks );

    // Data access
    public :
        virtual reterr_t SetRegNone( PCTSTR pstrValue );
        virtual reterr_t SetDWord( PCTSTR pstrValue, DWORD  val );
        virtual reterr_t GetDWord( PCTSTR pstrValue, DWORD& val );
        virtual reterr_t SetWord( PCTSTR pstrValue, WORD  val ) { return SetDWord( pstrValue, (DWORD) val ); }
        virtual reterr_t GetWord( PCTSTR pstrValue, WORD& val )
            { DWORD dw = (DWORD) val; reterr_t ret = GetDWord( pstrValue, dw ); if( ret ) val = (WORD) dw; return ret; }
        virtual reterr_t SetByte( PCTSTR pstrValue, BYTE  val ) { return SetDWord( pstrValue, (DWORD) val ); }
        virtual reterr_t GetByte( PCTSTR pstrValue, BYTE& val )
            { DWORD dw = (DWORD) val; reterr_t ret = GetDWord( pstrValue, dw ); if( ret ) val = (BYTE) dw; return ret; }
        virtual reterr_t SetUInt( PCTSTR pstrValue, UINT  val ) { return SetDWord( pstrValue, (DWORD) val ); }
        virtual reterr_t GetUInt( PCTSTR pstrValue, UINT& val )
            { DWORD dw = (DWORD) val; reterr_t ret = GetDWord( pstrValue, dw ); if( ret ) val = (UINT) dw; return ret; }
        virtual reterr_t SetInt( PCTSTR pstrValue, int  val ) { return SetDWord( pstrValue, (DWORD) val ); }
        virtual reterr_t GetInt( PCTSTR pstrValue, int& val )
            { DWORD dw = (DWORD) val; reterr_t ret = GetDWord( pstrValue, dw ); if( ret ) val = (int) dw; return ret; }
        virtual reterr_t SetBool( PCTSTR pstrValue, bool  val ) { return SetDWord( pstrValue, val ); }
        virtual reterr_t GetBool( PCTSTR pstrValue, bool& val )
            { DWORD dw = (DWORD) val; reterr_t ret = GetDWord( pstrValue, dw ); if( ret ) val = dw != 0; return ret; }
        virtual reterr_t SetBoolAsYesNo  ( PCTSTR pstrValue, bool  val ) { return SetString( pstrValue, val ? _T("yes") : _T("no") ); }
        virtual reterr_t GetBoolFromYesNo( PCTSTR pstrValue, bool& val );
        virtual reterr_t SetBinary( PCTSTR pstrValue, const void *pData, _size_t_ cData );
        virtual reterr_t GetBinary( PCTSTR pstrValue, void       *pData, _size_t_ cData );
        virtual reterr_t GetBinaryPart( PCTSTR pstrValue, void   *pData, _size_t_ cData );
        #ifndef _WIN32_WCE
        virtual reterr_t SetString( PCSTR  pstrValue, PCSTR  pStr );
        virtual reterr_t SetString( PCWSTR pstrValue, PCWSTR pStr );
        virtual reterr_t SetExpandString( PCSTR  pstrValue, PCSTR  pStr );
        virtual reterr_t SetExpandString( PCWSTR pstrValue, PCWSTR pStr );
        virtual reterr_t GetString( PCSTR  pstrValue, PSTR  pStr, _size_t_ zlenStr );
        virtual reterr_t GetString( PCWSTR pstrValue, PWSTR pStr, _size_t_ zlenStr );
        #else
        virtual reterr_t SetString( PCTSTR pstrValue, PCTSTR pStr );
        virtual reterr_t SetExpandString( PCTSTR pstrValue, PCTSTR pStr );
        virtual reterr_t GetString( PCTSTR pstrValue, PTSTR pStr, _size_t_ zlenStr );
        #endif
        virtual reterr_t GetString( PCTSTR pstrValue, String& str  );
        virtual reterr_t SetStringList( PCTSTR pstrValue, PCTSTR pStrLst );
        virtual reterr_t GetStringList( PCTSTR pstrValue, PTSTR  pStrLst, _size_t_ zzlenStrLst );
        virtual reterr_t SetFloatAsString   ( PCTSTR pstrValue, float   val );
        virtual reterr_t GetFloatFromString ( PCTSTR pstrValue, float&  val );
        virtual reterr_t SetDoubleAsString  ( PCTSTR pstrValue, double  val );
        virtual reterr_t GetDoubleFromString( PCTSTR pstrValue, double& val );
        virtual reterr_t SetIndexIntoStringList( PCTSTR pstrValue, _idx_  nIndex, PCTSTR pstrStringList );
        virtual reterr_t GetIndexIntoStringList( PCTSTR pstrValue, _idx_& nIndex, PCTSTR pstrStringList );
        virtual reterr_t SetDWordBitsAsStringList( PCTSTR pstrValue, DWORD  dwBits,
                                                   PCTSTR pstrStringList_Names, PCTSTR pstrStringList_BitNrs,
                                                   TCHAR chSep = _T(';') );
        virtual reterr_t GetDWordBitsAsStringList( PCTSTR pstrValue, DWORD& dwBits,
                                                   PCTSTR pstrStringList_Names, PCTSTR pstrStringList_BitNrs,
                                                   TCHAR chSep = _T(';') );
        virtual reterr_t SetColorrefAsString   ( PCTSTR pstrValue, COLORREF  crColor );
        virtual reterr_t GetColorrefFromString ( PCTSTR pstrValue, COLORREF& crColor );
        virtual reterr_t SetColorRGBAsString   ( PCTSTR pstrValue, BYTE  r, BYTE  g, BYTE  b );
        virtual reterr_t GetColorRGBFromString ( PCTSTR pstrValue, BYTE& r, BYTE& g, BYTE& b );
        virtual reterr_t SetColorARGBAsString  ( PCTSTR pstrValue, BYTE  a, BYTE  r, BYTE  g, BYTE  b );
        virtual reterr_t GetColorARGBFromString( PCTSTR pstrValue, BYTE& a, BYTE& r, BYTE& g, BYTE& b );
        virtual reterr_t SetPoint( PCTSTR pstrValue, const POINT *pPoint ) { return SetBinary( pstrValue, pPoint, sizeof(POINT) ); }
        virtual reterr_t GetPoint( PCTSTR pstrValue, POINT       *pPoint ) { return GetBinary( pstrValue, pPoint, sizeof(POINT) ); }
        virtual reterr_t SetPointAsString  ( PCTSTR pstrValue, POINT  pt );
        virtual reterr_t GetPointFromString( PCTSTR pstrValue, POINT& pt );
        virtual reterr_t SetSize( PCTSTR pstrValue, const SIZE *pSize ) { return SetBinary( pstrValue, pSize, sizeof(SIZE) ); }
        virtual reterr_t GetSize( PCTSTR pstrValue, SIZE       *pSize ) { return GetBinary( pstrValue, pSize, sizeof(SIZE) ); }
        virtual reterr_t SetSizeAsString  ( PCTSTR pstrValue, SIZE  size );
        virtual reterr_t GetSizeFromString( PCTSTR pstrValue, SIZE& size );
        virtual reterr_t SetRangeAsString  ( PCTSTR pstrValue, long  nStart, long  nWidth );
        virtual reterr_t GetRangeFromString( PCTSTR pstrValue, long& nStart, long& nWidth );
        virtual reterr_t GetRangeFromString( PCTSTR pstrValue, int&  nStart, int&  nWidth )
            { long s, w; reterr_t ret = GetRangeFromString( pstrValue, s, w ); if( ret ) { nStart = (int) s; nWidth = (int) w; } return ret; }
        virtual reterr_t SetRect( PCTSTR pstrValue, const RECT *pRect ) { return SetBinary( pstrValue, pRect, sizeof(RECT) ); }
        virtual reterr_t GetRect( PCTSTR pstrValue, PRECT       pRect ) { return GetBinary( pstrValue, pRect, sizeof(RECT) ); }
        virtual reterr_t SetRectAsString  ( PCTSTR pstrValue, RECT  rect );
        virtual reterr_t GetRectFromString( PCTSTR pstrValue, RECT& rect );
        virtual reterr_t SetRectListAsString  ( PCTSTR pstrValue, const RECT *prcRects, _size_t_  cRects );
        virtual reterr_t GetRectListFromString( PCTSTR pstrValue, PRECT     *pprcRects, _size_t_& cRects );
        #ifndef _WIN32_WCE
        virtual reterr_t SetWindowPlacement( PCTSTR pstrValue, const WINDOWPLACEMENT& wp );
        virtual reterr_t GetWindowPlacement( PCTSTR pstrValue, WINDOWPLACEMENT& wp );
        virtual reterr_t StoreWindowPlacementFromWnd( PCTSTR pstrValue, HWND hwnd );
        virtual reterr_t LoadWindowPlacementIntoWnd ( PCTSTR pstrValue, HWND hwnd, WINDOWPLACEMENT *pwp = NULL );
        virtual reterr_t StoreWindowposTopleftFromWnd( PCTSTR pstrValue, HWND hwnd );
        virtual reterr_t LoadWindowposTopleftIntoWnd ( PCTSTR pstrValue, HWND hwnd, POINT *ppt = NULL );
        #endif

        template< class T > reterr_t SetArbitraryT( PCTSTR pstrValue, T  t ) { return SetBinary( pstrValue, (void*) &t, sizeof(t) ); }
        template< class T > reterr_t GetArbitraryT( PCTSTR pstrValue, T& t ) { return GetBinary( pstrValue, (void*) &t, sizeof(t) ); }
};


#ifndef _argon_no_PrivateProfile
class PrivateProfile : public RegKey
{
    friend class PrivateProfile;

    protected :
        String m_strProfileFile; // full file that is the ini file
        String m_strSection;     // section name

        bool isValid() const { return !m_strProfileFile.IsEmpty() && !m_strSection.IsEmpty(); }


    public :
        PrivateProfile() { m_Storeloc = StoreLocationInPrivateProfile; }
        virtual ~PrivateProfile() { Close(); }

    // Creating/Opening and closing
    public :
        reterr_t Create( PCTSTR pstrProfileFile, PCTSTR pstrSection );
        reterr_t Open  ( PCTSTR pstrProfileFile, PCTSTR pstrSection );

        // And overridable creators.
        virtual reterr_t Create( const RegKey& parentkey, PCTSTR pstrSubkey );
        virtual reterr_t Open  ( const RegKey& parentkey, PCTSTR pstrSubkey );

        virtual reterr_t Close();
        virtual bool     isOpen() const { return !m_strProfileFile.IsEmpty(); }

    // Enumeration, states, ...
    public :
        virtual bool     EnumValue( _idx_ nValue, PTSTR pstrValue, _size_t_ zlenValue, PDWORD pdwType = NULL );
        virtual bool     EnumValue( _idx_ nValue, String& strValue, PDWORD pdwType = NULL );

        virtual bool     isValue( PCTSTR pstrValue ); // returns true, if the value exists

        virtual reterr_t DeleteValue( PCTSTR pstrValue );

    // Data access
    public :
        virtual reterr_t SetDWord( PCTSTR pstrValue, DWORD  val );
        virtual reterr_t GetDWord( PCTSTR pstrValue, DWORD& val );
        virtual reterr_t SetBinary( PCTSTR pstrValue, const void *pData, _size_t_ cData );
        virtual reterr_t GetBinary( PCTSTR pstrValue, void       *pData, _size_t_ cData );
        virtual reterr_t GetBinaryPart( PCTSTR pstrValue, void   *pData, _size_t_ cData );
        virtual reterr_t SetString( PCSTR  pstrValue, PCSTR  pStr );
        virtual reterr_t SetString( PCWSTR pstrValue, PCWSTR pStr );
        virtual reterr_t SetExpandString( PCSTR  pstrValue, PCSTR  pStr ) { return SetString( pstrValue, pStr ); }
        virtual reterr_t SetExpandString( PCWSTR pstrValue, PCWSTR pStr ) { return SetString( pstrValue, pStr ); }
        virtual reterr_t GetString( PCSTR  pstrValue, PSTR  pStr, _size_t_ zlenStr );
        virtual reterr_t GetString( PCWSTR pstrValue, PWSTR pStr, _size_t_ zlenStr );
        virtual reterr_t GetString( PCTSTR pstrValue, String& str  );
        virtual reterr_t SetStringList( PCTSTR pstrValue, PCTSTR pStrLst ) { return syserr (ERROR_NOT_SUPPORTED); }
        virtual reterr_t GetStringList( PCTSTR pstrValue, PTSTR  pStrLst, _size_t_ zzlenStrLst ) { return syserr (ERROR_NOT_SUPPORTED); }
};
#endif



enum storeloc_flags_t {
    storelocf_no_prefixLocation         = 1,    // do not include "storeloc:[<type>]:"
    storelocf_Default                   = 0
};

///<summary> Use this class as storage location which allows to create a
///  <b>RegKey</b> or derived storage object depending on the location
///  settings in this class. </summary>
class StoreLocation
{
    public :
        storelocation_t     m_Storeloc;

        // when in Registry
        HKEY                m_hRegKey;
        String              m_strRegSubkey;

        #ifndef _argon_no_PrivateProfile
        // when in PrivateProfile
        String              m_strProfileFile;   // full file that is the ini file
        String              m_strSection;       // section name
        #endif

    public :
        StoreLocation( storelocation_t storeloc = StoreLocationInRegistry );
        StoreLocation( PCTSTR pstrFrom );
        StoreLocation( const StoreLocation& storeloc ) { Copy( storeloc ); }
        StoreLocation& operator =( const StoreLocation& storeloc ) { return Copy( storeloc ); }

        bool           IsEmpty() const;
        StoreLocation& Empty();
        StoreLocation& Copy( const StoreLocation& storeloc );

        StoreLocation& SetRegistryCreate( HKEY hKey, PCTSTR pstrSubkey );
        StoreLocation& SetRegistryOpen  ( HKEY hKey, PCTSTR pstrSubkey );
        #ifndef _argon_no_PrivateProfile
        StoreLocation& SetPrivateProfileCreate( PCTSTR pstrProfileFile, PCTSTR pstrSection );
        StoreLocation& SetPrivateProfileOpen  ( PCTSTR pstrProfileFile, PCTSTR pstrSection );
        #endif

        String   toString     ( DWORD flags = storelocf_Default );
        reterr_t fromString   ( PCTSTR pstrStoreloc );
        reterr_t fromStringReg( PCTSTR pstrRegistryLoc );
        #ifndef _argon_no_PrivateProfile
        reterr_t fromStringPP ( PCTSTR pstrPrivateProfileLoc );
        #endif

        StoreLocation GetSubLocation( PCTSTR pstrSubkey );

        RegKey *OpenCurrentLocationForReading  () { return Create_at_CurrentLocation( false, KEY_READ ); }
        RegKey *CreateCurrentLocationForWriting() { return Create_at_CurrentLocation( true,  KEY_READ | KEY_WRITE ); }
        RegKey *Create_at_CurrentLocation( bool bCreate, REGSAM samDesired = KEY_ALL_ACCESS );

        RegKey *OpenSubLocationForReading  ( PCTSTR pstrSubkeypath ) { return Create_at_SubLocation( pstrSubkeypath, false, KEY_READ ); }
        RegKey *CreateSubLocationForWriting( PCTSTR pstrSubkeypath ) { return Create_at_SubLocation( pstrSubkeypath, true,  KEY_READ | KEY_WRITE ); }
        RegKey *Create_at_SubLocation( PCTSTR pstrSubkeypath, bool bCreate, REGSAM samDesired = KEY_ALL_ACCESS );

        static PCTSTR GetHkeyPredefName( HKEY hkey, bool bShortName = true );
        static HKEY   GetHkeyByPredefName( PCTSTR pstrName );
};




}

#endif
