/*
    File:               argonThreadLocalStorage.cpp
    Description:        Handling thread local storage more easily.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         16.04.2004
    last updated:       27.05.2006 (Roman Komary)

    Copyright (c) 2004-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonThreadLocalStorage_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonThreadLocalStorage.h"

namespace ArgonLibrary
{




Thread_Local_Storage::Creator Thread_Local_Storage::s_Creator;
bool Thread_Local_Storage::s_bDeinited = false;
DWORD Thread_Local_Storage::s_dwTlsIdx = TLS_INVALID_INDEX;
Thread_Local_Storage::TlsPtrArray *Thread_Local_Storage::s_paTlsPtrs = NULL;

Thread_Local_Storage::Creator::Creator()
{
    // The creator is used just for the purpose to initialize
    // the thread local storage somewhen while global objects
    // are being constructed.
    // But of course, #InitTls()# may be called even earlier
    // because it knowns if it the object has been created
    // already somewhen before and then will not recreate it.

    Thread_Local_Storage::InitTls();
}

Thread_Local_Storage::Thread_Local_Storage()
                     :m_htHandleMapper( 177, 16 )
{
    m_idThrd = GetCurrentThreadId();

    m_hInternalThrd = NULL;

    // To get a thread handle, we take the pseudo handle
    // from #GetCurrentThread()# and make a copy. The
    // new handle will address the current thread from
    // any thread within the current process.
    DuplicateHandle( GetCurrentProcess(), GetCurrentThread(),
                     GetCurrentProcess(), &m_hInternalThrd,
                     0, FALSE, DUPLICATE_SAME_ACCESS );
}

Thread_Local_Storage::~Thread_Local_Storage()
{
    DeleteThreadLocalAutoObjects();

    if( m_hInternalThrd )
        CloseHandle( m_hInternalThrd );
}



bool Thread_Local_Storage::ThreadHasEnded() const
{
    if( !m_hInternalThrd ) return false; // if NULL handle (which should never occur) return false so that we do not delete #this# object immediately.
    return WaitForSingleObject( m_hInternalThrd, 0 ) != WAIT_TIMEOUT;
}



Thread_Local_Storage *Thread_Local_Storage::GetStorage()
{
    if( s_dwTlsIdx == TLS_INVALID_INDEX )
        if( !InitTls() )
            return NULL;

    Thread_Local_Storage *ptr = (Thread_Local_Storage*) TlsGetValue( s_dwTlsIdx );
    if( ptr ) return ptr;

    if( !s_bDeinited )
    {
        LockModuleWide lmw;

        // First we delete any thread local storage objects
        //  whose threads have ended already.
        // Note: It is sufficient that we do this only whenever
        //  a new thread requests a new storage object, because
        //  we do these deletions only to prevent the #s_paTlsPtrs#
        //  array to grow and grow whenever new threads are created
        //  (and ended again) endlessly.
        //  If we would not delete useless storage objects, we could
        //  run into problems if even one single thread is permanently
        //  created, exited and recreated, and so on.
        //  To prevent this case, we delete useless storage objects
        //  only when the #s_paTlsPtrs# array is about to grow.
        //  But if we want useless storage objects to be deleted as
        //  early as possible, we should call #DeleteEndedThreadStorageObjs()#
        //  directly before the #TlsGetValue()# call above, but
        //  this could increase the processing of a thread (maybe even
        //  dramatically) especially if there are many threads running
        //  at the same time.
        //  So we think it is not really necessary to delete these
        //  objects as early as possible.
        DeleteEndedThreadStorageObjs();

        ptr = new Thread_Local_Storage;

        if( ptr )
        {
            if( !s_paTlsPtrs )
                { s_paTlsPtrs = new TlsPtrArray; s_paTlsPtrs->SetBlockSize( 128 ); }
            s_paTlsPtrs->Add( ptr );

            TlsSetValue( s_dwTlsIdx, (PVOID) ptr );
        }
    }

    return ptr;
}

void Thread_Local_Storage::DeleteEndedThreadStorageObjs()
{
    if( !s_paTlsPtrs ) return;

    Thread_Local_Storage **ptr = s_paTlsPtrs->Ptr();
    _size_t_               cnt = s_paTlsPtrs->GetSize();
    _idx_                  idx = 0;

    while( idx < (_idx_) cnt )
    {
        if( *ptr )
            if( (*ptr)->ThreadHasEnded() )
            {
                // If thread is has ended, we remove and delete
                // its thread storage object.

                delete (*ptr);

                if( s_paTlsPtrs->Remove( idx ) ) {
                    --cnt;
                    ptr = s_paTlsPtrs->Ptr() + idx;
                    continue;
                } else
                    (*ptr) = NULL;
            }

        ++idx;
        ++ptr;
    }
}

bool Thread_Local_Storage::InitTls()
{
    if( s_bDeinited ) return false;

    LockModuleWide lmw;

    if( s_bDeinited ) return false;

    if( s_dwTlsIdx == TLS_INVALID_INDEX )
        s_dwTlsIdx = TlsAlloc();    // #TlsAlloc()# will return #TLS_INVALID_INDEX# on error

    if( s_dwTlsIdx == TLS_INVALID_INDEX )
        return false;

    atexit( DeinitTls );

    return true;
}

void Thread_Local_Storage::DeinitTls()
{
    if( s_dwTlsIdx != TLS_INVALID_INDEX )
    {
        LockModuleWide lmw;

        s_bDeinited = true;

        TlsFree( s_dwTlsIdx );
        s_dwTlsIdx = TLS_INVALID_INDEX;

        if( s_paTlsPtrs )
        {
            for( _idx_ i = (_idx_) s_paTlsPtrs->GetSize() - 1 ; i >= 0 ; --i )
                if( (*s_paTlsPtrs) [i] )
                {
                    Thread_Local_Storage *p = (*s_paTlsPtrs) [i];
                    (*s_paTlsPtrs) [i] = NULL;
                    // During destruction of the #p# object, it
                    //  can be, that a call to #GetStorage()#
                    //  and #DeleteEndedThreadStorageObjs()# is done.
                    // To prevent array data inconsitencies, we first
                    //  set the pointer to NULL in the array
                    //  before deleting the object.
                    // But we must also prevent any further new
                    //  #Thread_Local_Storage# objects to be created
                    //  since we are deleting them currently.
                    //  We do this using the #s_bDeinited# member.
                    delete p;
                }

            delete s_paTlsPtrs;
            s_paTlsPtrs = NULL;
        }
    }
}



_idx_ Thread_Local_Storage::s_ThreadLocalIndexCounter = 0;

_idx_ Thread_Local_Storage::s_ThreadLocalAutoObjectIndexCounter = 0;

void Thread_Local_Storage::DeleteThreadLocalAutoObjects()
{
    // Since the destructor of #this# object can be called by
    // any different thread and therefore also this method,
    // we must lock its access module-wide to be safe.
    LockModuleWide::Enter();

    for( _idx_ i = (_idx_) m_aThreadLocalAutoObjectPointers.GetSize() - 1 ; i >= 0 ; --i )
        if( m_aThreadLocalAutoObjectPointers[i] )
        {
            ThreadLocalAutoBase *p = m_aThreadLocalAutoObjectPointers[i];
            m_aThreadLocalAutoObjectPointers[i] = NULL;
            delete p;
        }

    m_aThreadLocalAutoObjectPointers.Empty();

    LockModuleWide::Leave();
}



INT_PTR Thread_Local_Storage::AllocHandleNameID( PCTSTR pstrHandleName )
{
    if( _isstrempty_( pstrHandleName ) ) return -1;

    _size_t_  count = m_aHandleNames.GetSize();
    String   *p     = m_aHandleNames.Ptr();

    for( _idx_ i = 0 ; i < (_idx_) count ; ++i )
        if( p[i] == pstrHandleName )
            return (INT_PTR) i;

    String s = pstrHandleName;
    m_aHandleNames.Add( s );    // #m_aHandleNames# holds the name for the lifetime of the #Thread_Local_Storage# object

    return (INT_PTR) count;
}




}
