/*
    File:               argonTrayicon.h
    Description:        Class for easy tray icon handling.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         17.11.2003
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonTrayicon_h
#define __argonTrayicon_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include <shellapi.h>       // for Shell_NotifyIcon()
    #include "argonInstance.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4100 )    // unreferenced parameter
#endif

namespace ArgonLibrary
{




#ifndef _countof
#define _countof(array) (sizeof(array) / sizeof((array)[0]))
#endif

enum trayPopupFlags_t {
    traypopupflags_setForeground        = 0x00000001,   // if the notification window shall be brought to foreground
                                                        //  when popupmenu opens (or on left-double-click for default-menuitem).
                                                        //  This flag is recommended when the menuitems in the popupmenu
                                                        //  create new windows or dialogs so that the application has the focus.
    traypopupflags_allowDuringModalDlg  = 0x00000002,   // if popupmenu is allowed when a modal dialog exists
                                                        //  (who's parent is the notification window).
                                                        //  This is done by checking if the notification window
                                                        //  is enabled or not. (it is disabled if it has a modal dlg).
    traypopupflags_setFocusToTrayicon   = 0x00000004,   // if this is specified, then the trayicon in the systray will
                                                        //  receive the focus when the popupmenu has closed.
                                                        //  It is better for user handling to disable this flag if
                                                        //  the menuitems in the popup menu create new windows or
                                                        //  dialogs because it would be better them to have the focus
                                                        //  instead of the systray.
    traypopupflags_Default              = traypopupflags_setForeground
};

class Trayicon
{
    protected :
        UINT    m_idTrayicon;
        bool    m_bTrayiconExists;
        HWND    m_hwndNotify;

        NOTIFYICONDATA m_nid;
        HICON   m_hIcon;
        bool    m_autodelHicon;             // if to auto delete #m_hIcon#

        bool    m_autoPopupMenu;
        PCTSTR  m_pstrMenuname_or_resid;    // the menu to load to get the popup menu
        int     m_nSubmenu;                 // the submenu index to use as popup menu
        UINT    m_idDefaultMenuitem;        // if != 0 then this is the default menu item. left-double clicking the icon will invoke it
        DWORD   m_dwPopupmenuFlags;         // combination of #trayPopupFlags_t# values

    public :
        Trayicon( UINT idTrayicon = 1 );
        ~Trayicon() { DeleteIcon(); DelHicon(); }

        UINT getTrayiconID() const { return m_idTrayicon; }
        bool trayiconExists() const { m_bTrayiconExists; }

    // Icon creation, recreation and modification
        bool CreateIcon( HWND hwndNotify, UINT  idIcon, UINT idMsg, PCTSTR pstrTip );
        bool CreateIcon( HWND hwndNotify, HICON hIcon,  UINT idMsg, PCTSTR pstrTip );
        bool DeleteIcon();
        bool SetIcon( UINT  idIcon, PCTSTR pstrTip = NULL );
        bool SetIcon( HICON hIcon,  PCTSTR pstrTip = NULL );
        bool SetTip( PCTSTR pstrTip );
        bool SetFocus();
        bool RecreateIcon();

    // automatic popup menu
        void SetAutoPopupMenu( PCTSTR pstrMenuname, int nSubmenu, UINT idDefaultMenuitem = 0, DWORD flags = traypopupflags_Default );
        void SetAutoPopupMenu( UINT   idMenu,       int nSubmenu, UINT idDefaultMenuitem = 0, DWORD flags = traypopupflags_Default )
            { SetAutoPopupMenu( MAKEINTRESOURCE(idMenu), nSubmenu, idDefaultMenuitem, flags ); }
        void DisableAutoPopupMenu() { m_autoPopupMenu = false; }

        void openPopupMenu();   // this will be called by #HandleWndMessage()# if an auto-popup-menu is set

    protected :
        bool __Shell_NotifyIcon( DWORD dwMessage, PNOTIFYICONDATA pData );
        void DelHicon();

    // Taskbar restart notification
    protected :
        static UINT s_wmTaskbarRestart;
    public :
        static UINT getTaskbarRestartMessage();

    // window message handling
    public :
            ///<summary> Call this from the notification-window's window proc to automatically
            ///  perform the necessary operations for the Trayicon. </summary>
            ///<param name="lr"> If true is returned, <b>lr</b> will contain the result value to return to the system. </param>
            ///<returns> true if the message has been handled in here and you should not pass it on
            ///  to the other handling. false means that the message has been handled or not, but you
            ///  should pass it on to further window message handling. </returns>
        bool HandleWndMessage( LRESULT& lr, HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );
};




}

#endif
